/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include "struct.h"
#include "struct/grdraw.h"
#include "global.h"
#include "misc.h"
#include "calc.h"
#include "linestyle.h"
#include "layer.h"
#include "obj_line.h"
#include "obj_circle.h"
#include "obj_arc.h"
#include "obj_polyline.h"
#include "obj_point.h"
#include "obj_rline.h"
#include "obj_cline.h"
#include "obj_text.h"
#include "obj_ellipse.h"
#include "obj_elarc.h"
#include "obj_spline.h"
#include "obj_insert.h"
#include "redraw.h"
#include "xor.h"
#include "select.h"

/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */


/* --  F U N C T I O N S  -- */
static GList *duplicate_attr(OBJECT *object)
     /* Copy the attribute list for blocks */
{
  GList *list=NULL;
  GList *dup_list;

  dup_list=g_list_first(object->data.block.attr);

  while(dup_list!=NULL)
    {
      list=g_list_append(list, g_strdup(dup_list->data));
      dup_list=g_list_next(dup_list);
    }

  return list;
}

OBJECT *copy_polyline(DOCUMENT *doc, LAYER *layer, OBJECT *pline)
     /* Copy the polyline pline to the document doc and layer layer.
	Returns the newly created polyline. */
{
  OBJECT *new_pline;
  GList *node_list;

  node_list=g_list_first(pline->data.polyline.node_list);

  new_pline=create_new_polyline(doc, layer, pline->name, pline->type_colour, pline->colour,
				pline->type_linestyle, pline->linestyle, pline->type_linewidth,
				pline->linewidth, pline->data.polyline.closed);

  while(node_list!=NULL)
    {
      add_node(new_pline, ((NODE *)node_list->data)->type, ((NODE *)node_list->data)->x,((NODE *)node_list->data)->y,
	       ((NODE *)node_list->data)->sw,((NODE *)node_list->data)->ew,
	       ((NODE *)node_list->data)->weight);

      node_list=g_list_next(node_list);
    }

  obj_bound_polyline(new_pline);
  layer->object_list=g_list_append(layer->object_list,(gpointer)new_pline);

  return new_pline;

}

OBJECT *copy_spline(DOCUMENT *doc, LAYER *layer, OBJECT *spline)
     /* Copy the spline spline to the document doc and layer layer.
	Returns the newly created spline. */
{
  OBJECT *new_spline;
  GList *node_list;

  node_list=g_list_first(spline->data.spline.node_list);

  new_spline=create_new_spline(doc, layer, spline->name, spline->type_colour, spline->colour,
			       spline->type_linestyle, spline->linestyle, spline->type_linewidth,
			       spline->linewidth, spline->data.spline.closed);

  while(node_list!=NULL)
    {
      spline_add_node(new_spline,((NODE *)node_list->data)->x,
		      ((NODE *)node_list->data)->y, ((NODE *)node_list->data)->weight);

      node_list=g_list_next(node_list);
    }

  obj_bound_spline(new_spline);
  layer->object_list=g_list_append(layer->object_list,(gpointer)new_spline);

  return new_spline;

}

void copy_selected_object(DOCUMENT *src_doc, DOCUMENT *dest_doc, TRANSFORM *transform)
     /* Copy selected objects from a document to another and
	appling the transformation on the copied objects */
{
  GList *select_list;
  LAYER *src_layer;
  LAYER *dest_layer;
  LINESTYLE *dest_linestyle=NULL;
  OBJECT *dest_object;
  OBJECT *src_object;

  remove_old();

  if(src_doc==NULL || src_doc->selection_list==NULL)
    return;

  if(dest_doc==NULL)
    dest_doc=src_doc;

  select_list=g_list_first(src_doc->selection_list);

  while(select_list!=NULL)
    {
      src_object=(OBJECT *)select_list->data;
      src_layer=src_object->layer;

      /* Linestyle is only valid for none-dimension-objects */
      if(src_object->type!=TYPE_DIMENSION)
	{
	  /* Check line style */
	  dest_linestyle=linestyle_find(dest_doc, src_object->linestyle->name);
	  
	  /* It was not found, create it! */
	  if(dest_linestyle==NULL)
	    {
	      gdouble *pattern=NULL;
	      gint i;
	      
	      if(src_object->linestyle->type!=LINE_SOLID)
		{
		  pattern=g_new(gdouble, src_object->linestyle->dash_no);
		  
		  for(i=0;i< src_object->linestyle->dash_no; i++)
		    pattern[i]=src_object->linestyle->dash_dlist[i];
		}
	      
	      dest_linestyle=linestyle_new(dest_doc, -1, src_object->linestyle->name, 
					   src_object->linestyle->type, src_object->linestyle->dash_length, 
					   src_object->linestyle->dash_no, pattern);		      
	    }
	}
      
      /* Check layer */
      /* Check this layers name. If it is "0" use the dest_doc current layer */
      if(!strcmp("0", src_layer->name))
	{
	  dest_layer=dest_doc->current_layer;
	  if(src_object->type_linestyle==BYLAYER)
	    dest_linestyle=dest_layer->linestyle;
	}
      else /* Create a layer with this name in the dest_co */
	{
	  /* Is this layer also available in dest_doc? */	    
	  dest_layer=layer_find(dest_doc, src_layer->name);
	  
	  /* If not found create such a layer */
	  if(dest_layer==NULL)
	    {
	      LINESTYLE *layer_linestyle;
	      layer_linestyle=linestyle_find(dest_doc, src_layer->linestyle->name);
	      
	      if(layer_linestyle==NULL)
		{
		  gdouble *pattern=NULL;
		  gint i;
		  
		  if(src_layer->linestyle->type!=LINE_SOLID)
		    {
		      pattern=g_new(gdouble, src_layer->linestyle->dash_no);
		      
		      for(i=0;i<src_layer->linestyle->dash_no; i++)
			pattern[i]=src_layer->linestyle->dash_dlist[i];
		    }
		  
		  layer_linestyle=linestyle_new(dest_doc, -1, src_layer->linestyle->name, 
						src_layer->linestyle->type, src_layer->linestyle->dash_length, 
						src_layer->linestyle->dash_no, pattern);
		}
	      
	      dest_layer=layer_new(dest_doc, -1, layer_linestyle, src_layer->linewidth, 
				   src_layer->colour->c[0], src_layer->colour->c[1], src_layer->colour->c[2],
				   src_layer->name, src_layer->visible, src_layer->editable);
	    }
	}
      
      switch(src_object->type)
	{
	case TYPE_LINE:
	  dest_object=obj_add_line(dest_doc, dest_layer,src_object->name,src_object->type_colour, src_object->colour,
				   src_object->type_linestyle, dest_linestyle,
				   src_object->type_linewidth, src_object->linewidth,
				   src_object->data.line.p1.x, src_object->data.line.p1.y,
				   src_object->data.line.p2.x, src_object->data.line.p2.y);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_line(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_RLINE:
	  dest_object=obj_add_rline(dest_doc, dest_layer,src_object->name,src_object->type_colour, src_object->colour,
				    src_object->type_linestyle, dest_linestyle,
				    src_object->type_linewidth, src_object->linewidth,
				    src_object->data.line.p1.x, src_object->data.line.p1.y,
				    src_object->data.line.p2.x, src_object->data.line.p2.y);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_rline(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_CLINE:
	  dest_object=obj_add_cline(dest_doc, dest_layer,src_object->name,src_object->type_colour, src_object->colour,
				    src_object->type_linestyle, dest_linestyle,
				    src_object->type_linewidth, src_object->linewidth,
				    src_object->data.line.p1.x, src_object->data.line.p1.y,
				    src_object->data.line.p2.x, src_object->data.line.p2.y);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_cline(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_CIRCLE:
	  dest_object=obj_add_circle(dest_doc, dest_layer, src_object->name,src_object->type_colour, src_object->colour,
				     src_object->type_linestyle, dest_linestyle,
				     src_object->type_linewidth, src_object->linewidth,
				     src_object->data.circle.centrum.x,
				     src_object->data.circle.centrum.y,
				     src_object->data.circle.radius);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_circle(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_ARC:
	  dest_object=obj_add_arc(dest_doc, dest_layer,src_object->name,src_object->type_colour, src_object->colour,
				  src_object->type_linestyle, dest_linestyle,
				  src_object->type_linewidth, src_object->linewidth,
				  src_object->data.arc.centrum.x,
				  src_object->data.arc.centrum.y,
				  src_object->data.arc.radius,
				  src_object->data.arc.start_angle,
				  src_object->data.arc.end_angle);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_arc(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_POLYLINE:
	  dest_object=copy_polyline(dest_doc,dest_layer,src_object);
	  dest_object->linestyle=dest_linestyle;
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_polyline(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_POINT:
	  dest_object=obj_add_point(dest_doc, dest_layer, src_object->name, src_object->type_colour, src_object->colour, 
				    src_object->type_linestyle, dest_linestyle,
				    src_object->type_linewidth, src_object->linewidth,
				    src_object->data.point.point_style, src_object->data.point.point_type,
				    src_object->data.point.point_size, 
				    src_object->data.point.centrum.x, src_object->data.point.centrum.y);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_point(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_ELLIPSE:
	  dest_object=obj_add_ellipse(dest_doc, dest_layer, src_object->name, src_object->type_colour, src_object->colour, 
				      src_object->type_linestyle, dest_linestyle,
				      src_object->type_linewidth, src_object->linewidth,
				      src_object->data.ellipse.centrum.x, 
				      src_object->data.ellipse.centrum.y,
				      src_object->data.ellipse.major,  src_object->data.ellipse.minor,
				      src_object->data.ellipse.angle);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_ellipse(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_ELARC:
	  dest_object=obj_add_elarc(dest_doc, dest_layer, src_object->name, src_object->type_colour, src_object->colour, 
				    src_object->type_linestyle, dest_linestyle,
				    src_object->type_linewidth, src_object->linewidth,
				    src_object->data.elarc.centrum.x, 
				    src_object->data.elarc.centrum.y,
				    src_object->data.elarc.major, src_object->data.elarc.minor,
				    src_object->data.elarc.angle, src_object->data.elarc.start_angle,
				    src_object->data.elarc.end_angle);	       				   
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_elarc(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_TEXT:
	  dest_object=obj_add_text(dest_doc, dest_layer, src_object->name, src_object->type_colour, src_object->colour, 
				   src_object->type_linestyle, dest_linestyle,
				   src_object->type_linewidth, src_object->linewidth,
				   src_object->data.text.style, src_object->data.text.justification,
				   src_object->data.text.angle, src_object->data.text.height,
				   src_object->data.text.insert.x,src_object->data.text.insert.y,
				   src_object->data.text.alignment.x,src_object->data.text.alignment.y,
				   src_object->data.text.text);
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_text(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_SPLINE:
	  dest_object=copy_spline(dest_doc, dest_layer, src_object);
	  dest_object->linestyle=dest_linestyle;
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_spline(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	case TYPE_BLOCK:
	  dest_object=obj_add_insert(dest_doc, dest_layer, src_object->name, src_object->type_colour, src_object->colour, 
				     src_object->type_linestyle, dest_linestyle,
				     src_object->type_linewidth, src_object->linewidth,
				     src_object->data.block.insert_point.x, src_object->data.block.insert_point.y,
				     src_object->data.block.scale_x, src_object->data.block.scale_y,
				     src_object->data.block.rotation,
				     duplicate_attr(src_object), src_object->data.block.block);
	  
	  make_object_selected(dest_object, OBJECT_SEL, FALSE);
	  obj_transform_insert(dest_object, transform);
	  make_object_selected(dest_object, OBJECT_NOT_SEL, FALSE);
	  break;
	default:
	  dest_object=NULL;
		  break;
	}
      
      redraw_object(dest_doc, dest_object);
      
      select_list=g_list_next(select_list);
    }
}














