/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Most of this code is converted code from the ExprParser, 
   written by Giovanni Dicanio and is included in QCad 1.4.4.
 */

/* -- I N C L U D E -- */
#include <gtk/gtk.h>
#include <string.h>                    /* standard C strings managment */
#include <stdlib.h>                    /* memory managment, atof(), etc. */
#include <ctype.h>                     /* isdigit(), isalpha(), etc. */
#include <math.h>                      /* The standard math library. */
#include "struct.h"
#include "global.h"
#include "variable.h"

/* Error codes */
#define EXPRPARSER_OK                         0   /* all right (no error) */
#define EXPRPARSER_DIVIDE_BY_ZERO             1   /* division by 0 */
#define EXPRPARSER_PARENTHESES_NOT_BALANCED   2   /* parentheses non balanced */
#define EXPRPARSER_OPEN_PARENTHESIS_EXPECTED  3   /* a '(' parenthesis was expected */
#define EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED 4   /* '(' expected */                
#define EXPRPARSER_COMMA_EXPECTED             5   /* a ',' was expected */
#define EXPRPARSER_UNKNOWN_VARIABLE           6   /* unknown variable */
#define EXPRPARSER_VARIABLE_MEMORY_LEAK       7   /* the variable array is full! */
#define EXPRPARSER_BAD_IDENTIFIER             8   /* a bad identifier encountered */
#define EXPRPARSER_DOMAIN_ERROR               9   /* value not in function domain
						   * (e.g.: log(x) is defined for x > 0;
						   * so log(-3) causes a DOMAIN_ERROR.) */

/* Pointer to current work document */ 
static DOCUMENT *work_doc;

/* Units for angles */
enum angleUnit{
  RADIANS,                             /* radians */
  DEGREES                              /* degrees */
};

/* Default angle unit. */
#define EXPRPARSER_DEFAULT_ANGLE_UNIT   DEGREES

/* -- L O C A L   V A R I A B E L S -- */
/* This static character is pointed by m_currChar to specify
   that the parsing process must be interrupted, beacuse an
   error occurred (see set_error() implementation for more info). */
static gchar stop_parsing_char = '\0';


static gchar *m_expr;      /* Holds a copy of the string to parse */
static gchar *m_curr_char; /* Holds the current character */
static gint m_error;       /* Holds the error number */
static gint m_error_pos;     /* Holds the position where error occurred */
static gint m_angle_mode;  /* Holds the current angle mode */

/* -- P R O T O T Y P E S -- */
static gdouble do_expression(void);
static gdouble do_function_sin(void);
static gdouble do_function_cos(void);
static gdouble do_function_tan(void);
static gdouble do_function_cot(void);
static gdouble do_function_asin(void);
static gdouble do_function_acos(void);
static gdouble do_function_atan(void);
static gdouble do_function_sqrt(void);
static gdouble do_function_sqr(void);
static gdouble do_function_pow(void);
static gdouble do_function_exp(void);
static gdouble do_function_ln(void);

/* -- FUNCTIONS -- */

static gchar *string_create_from(const gchar *source)
     /* If source is NULL, allocate a '\0' string.
	If source is != NULL, allocate memory and copy source into it. */
{
  gchar *s;
  gint len;

  if (source) 
    {
      /* Get the number of characters in source string. */
      len = strlen(source);

      /* Try allocating string memory. */
      s=g_new(gchar, len+1);
      
      /* Copy characters from source string. */
      strcpy(s, source);
      
      /* Return this new string. */
      return s;
    }
  else 
    {
      /* Create a '\0' string. */ 
      s=g_new(gchar,1);
                     
      /* set to empty string */
      s[0]='\0';                       
                
      return s;
    }
}

static void go_ahead()
{
  /* Go ahead only if there are no errors! */
  if (m_error == EXPRPARSER_OK)
    m_curr_char++;           /* go to next char */
}

static gboolean set_error(gint err)
{
  /* Get the error position in the expression string. */
  m_error_pos = (gint)(m_curr_char - m_expr);
        
  /* Set error status. */
  m_error = err;
  
  /* Stop parsing! */
  m_curr_char = &stop_parsing_char;
  
  /* Always return false, as an error occurred. */
  return FALSE;
}

static void skip_blanks()
{
  while (*m_curr_char)
    if ((*m_curr_char != ' ') && (*m_curr_char != '\t'))
      break;
    else
      m_curr_char++;
}

static gdouble do_number()
{
  gdouble result;
  gboolean stop_parsing_number; 
  const gint buf_chars = 80;
  gchar buf[buf_chars+1];
  gint i;

  stop_parsing_number = TRUE;
  
  for (i = 0; isdigit(*m_curr_char); i++)
    {
      buf[i] = *m_curr_char;
      go_ahead();
    }

  /* We encountered a non digit char. 
     If this char is '.' or 'E' or 'e', we must go on reading;
     else, it means we've done our job. */

  if (*m_curr_char == '.')
    {
      /* continue reading */
      buf[i] = *m_curr_char;
      ++m_curr_char;
      ++i;
      
      for( ; isdigit(*m_curr_char); i++)
	{
	  buf[i] = *m_curr_char;
	  go_ahead();
	}
    }
  
  if (*m_curr_char == 'E' || *m_curr_char == 'e')
    {
      // continue reading
      buf[i] = *m_curr_char;
      ++m_curr_char;
      ++i;
      
      for( ; isdigit(*m_curr_char); i++)
	{
	  buf[i] = *m_curr_char;
	  go_ahead();
	}
    }
  else
    stop_parsing_number = TRUE;
  
  if ( !(stop_parsing_number) && (*m_curr_char == '+' || *m_curr_char == '-'))
    {
      /* continue reading */
      buf[i] = *m_curr_char;
      ++m_curr_char;
      ++i;
      
      for( ; isdigit(*m_curr_char); i++)
	{
	  buf[i] = *m_curr_char;
	  go_ahead();
	}
    }
  
  skip_blanks();
  buf[i] = '\0';
  result = atof(buf);
  
  return result;
}

/* It returns a static string! */
static gchar *get_identifier()
{
  static gchar identifier[32];
  gint i;
  
  /* m_curr_char must point to the first identifier character.
     Identifiers must start with a letter of '_'.
     Other characters can be a-z, A-Z and digits 0-9. */
  
  if ((*m_curr_char != '_') && (!isalpha(*m_curr_char)))
    {
      /* Error: bad identifier */
      set_error(EXPRPARSER_BAD_IDENTIFIER);
      identifier[0] = '\0';           /* empty string */
      return identifier;
    }


  for (i = 0; i < 32; i++)
    {
      if ((isalpha(*m_curr_char)) || (*m_curr_char == '_') || (isdigit(*m_curr_char)))
	{
	  /* This is a valid character.
	     Store it into the identifier string. */
	  identifier[i] = *m_curr_char;
	  
	  ++m_curr_char;           /* go to next character */
	}
      else
	break;  /* end of identifier */
    }
  
  identifier[i] = '\0';   /* don't forget to set the End of String character! */
  
  return identifier;
}

static gboolean get_variable(gchar *name, gdouble *value)
{
  if (name == NULL)
    return set_error(EXPRPARSER_UNKNOWN_VARIABLE);

  if (*name == '\0')
    return set_error(EXPRPARSER_UNKNOWN_VARIABLE);
  
  /* Scan the variable array to find the variable */  
  if(var_get_variable_float(work_doc, name, value))
    return TRUE;
  else
    return set_error(EXPRPARSER_UNKNOWN_VARIABLE);
}

static gdouble do_function_variable()
{
  /* Do both functions and variables stuff!
     (variables and functions are processed in similar ways,
     so I put all the code in this function.) */
  
  gchar *identifier;
  gdouble result;

  /* Get the identifier (maybe a function or variable identifier). */
  identifier = get_identifier();
  if (identifier == NULL)
    {
      /* Error! */
      set_error(EXPRPARSER_BAD_IDENTIFIER);
      return 0;                   /* dummy value */
    }
  
  
  /* Check for the functions we implement; if there is no match,
     it means that the identifier may be a variable identifier. */
  if (strcmp(identifier, "sin") == 0)
    result = do_function_sin();
  else if (strcmp(identifier, "cos") == 0)
    result = do_function_cos();
  else if (strcmp(identifier, "tan") == 0)
    result = do_function_tan();
  else if (strcmp(identifier, "cot") == 0)
    result = do_function_cot();
  else if (strcmp(identifier, "asin") == 0)
    result = do_function_asin();
  else if (strcmp(identifier, "acos") == 0)
    result = do_function_acos();
  else if (strcmp(identifier, "atan") == 0)
    result = do_function_atan();
  else if (strcmp(identifier, "sqrt") == 0)
    result = do_function_sqrt();
  else if (strcmp(identifier, "sqr") == 0)
    result = do_function_sqr();
  else if (strcmp(identifier, "pow") == 0)
    result = do_function_pow();
  else if (strcmp(identifier, "exp") == 0)
    result = do_function_exp();
  else if (strcmp(identifier, "ln") == 0)
    result = do_function_ln();
  /*
    ...
    ... add more functions here ...
    ... 
  */
        
  /* some useful constants ('pi' and 'e') */
  else if (strcmp(identifier, "pi") == 0)
    result = M_PI;
  else if (strcmp(identifier, "e") == 0)
    result = M_E;  
  else
      get_variable(identifier, &result);
  /* If there will be a variable error (e.g.: variable 
     identifier unknown), the get_variable() function
     will call set_error(), stopping the parsing process.*/
  
  /*
    NOTE:
    
    The way I process functions in the above "if...else if..."
    statement is very trivial! I absolutely need to find
    a better solution (e.g.: use hash tables!). */
  
  return result;
}

static gdouble do_number_function()
{
  gdouble result;

  result = 0;

  if (*m_curr_char == '(') 
    {
      go_ahead();
      skip_blanks();
      result = do_expression();

      if (*m_curr_char != ')')
	{
	  set_error(EXPRPARSER_PARENTHESES_NOT_BALANCED);
	  
	  /* NOTE:
	     This error should be discovered during the startup process
             of parsing (there is a function called parenthesesBalanced()
             to do this!). */
	  return 0;       /* beware... */
	}

      go_ahead();
      skip_blanks();      
    }
  else if (isdigit(*m_curr_char))
    result = do_number();            
  else if (isalpha(*m_curr_char))
    result = do_function_variable();
  
  return result;
}

static gdouble do_mul_div()
{
  gboolean stop;
  gdouble result;

  result = do_number_function();

  stop = FALSE;
  while (!stop)
    {
      if (*m_curr_char == '*')
	{
	  go_ahead();
	  skip_blanks();
	  result *= do_number_function();
	  stop = FALSE;
	}
      else if (*m_curr_char == '/') 
	{
	  gdouble denominator;

	  go_ahead();
	  skip_blanks();
	  
	  denominator = do_number_function();
	  if (denominator == 0.0)                 /* better check would be: fabs(denominator) < eps */
	    set_error(EXPRPARSER_DIVIDE_BY_ZERO);
	  else
	    result /= denominator;
	  
	  stop = FALSE;
	}
      else
	stop = 1;

    }

  return result;
}


static gboolean parentheses_balanced()
{
  gint num_open = 0;
  gint num_close = 0;
        
  gchar *ch = m_expr;

  while (*ch != '\0')
    {
      if (*ch == '(')
	num_open++;
      else if (*ch == ')')
	num_close++;
      
      ch++;
    }
  
  return (num_open == num_close);
}

static gboolean start_parsing()
{
  /* Start with first character in the expression string. */
  m_curr_char = m_expr;
  
  /* No error occurred. */
  m_error = EXPRPARSER_OK;

  /* Error position not defined (-1). */
  m_error_pos = -1;
  
  /* Before doing parsing, check parenthese balancing. */
  if (!parentheses_balanced()) 
    {
      set_error(EXPRPARSER_PARENTHESES_NOT_BALANCED);
      return FALSE;
    }
  else
    return TRUE;
}

static gdouble do_expression()
{
  gboolean stop;
  gdouble result;

  skip_blanks();
  result = do_mul_div();

  stop = FALSE;
  while(!stop)
    {
      if(*m_curr_char == '+')
	{       
	  go_ahead();
	  skip_blanks();
	  result += do_mul_div();
	  stop = FALSE;
	} 
      else if(*m_curr_char == '-') 
	{	  
	  go_ahead();
	  skip_blanks();
	  result -= do_mul_div();
	  stop = FALSE;
	}
      else
	stop = TRUE;
    }
  
  return result;
}


static gboolean do_parsing(gdouble *answer)
{
  if (!start_parsing())     /* startup the parsing process */
    return FALSE;
  
  *answer=do_expression();  /* start parsing the expression */
  
  return (m_error == EXPRPARSER_OK);   /* Has the parsing been successful? */ 
}

static gboolean init_parser(const gchar *expr)
     /* Init current parser object. */
{
  m_expr = string_create_from(expr);
  if (m_expr == NULL)
    return FALSE;         /* error: memory allocation failed! */
  
  m_error = EXPRPARSER_OK;           /* no error occurred */

  m_angle_mode = EXPRPARSER_DEFAULT_ANGLE_UNIT; /* set default angle unit */
  
  return TRUE;            /* initialisation has been successful */
}


gint exprparser(DOCUMENT *doc, gchar *buffer, gdouble *answer)
     /* This function is the main function.
	It takes the input string as the first parameter
	and returns the parsed value in the second parameter
	It returns an error code */
{
  work_doc=doc;
  init_parser(buffer);
  do_parsing(answer);

  return m_error;
}


/* -- Angle convertion functions -- */
static gdouble convert_angle_in_radians(gdouble angle)
{
  /* This function converts angles in radians.
     It is useful for trig functions.
     In fact, standard C math library trig functions assume
     that angles are in radians; but this parser object can
     manage also angles expressed in other units (e.g.: 
     the more human readable degrees ^_^).
     So, trig functions implementations will always call
     this function to be sure that their input parameter
     angle is in radians. 
  */
  gdouble rad_angle;  /* angle in radians */

  if (m_angle_mode == DEGREES)
    rad_angle = angle * (gdouble)((gdouble)M_PI / (gdouble)180.0f);        /* convert degrees to radians */
  else    /* m_angle_mode == RADIANS */
    rad_angle = angle;                        /* angle already in radians */

  return rad_angle;                           /* return angle expressed in radians */
}

static gdouble convert_angle_in_curr_mode(gdouble angle_rad)
{
  /* This function convert an angle expressed in radians to
     the same angle expressed in the current angle unit.
     
     This function is useful for return values of inverse
     trig functions (e.g.: asin, acosi, atan, ...).
     In fact, these functions implementations in the standard
     C math library return angles expressed in radians.
     So we need to convert these angles in the current angle mode.
  */
  gdouble angle;                      /* angle expressed in current angle mode */
  
  if (m_angle_mode == DEGREES)
    angle = angle_rad * (gdouble)(180.0f / (gdouble)M_PI); /* get angle in degrees */
  else /* m_angle_mode == RADIANS */
    angle = angle_rad;                 /* angle already in radians */

  return angle;
}

/* -- Implementations of specific functions (e.g.: sin, cos, etc.) parsing. -- */


static gdouble do_function_sin()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
  
  go_ahead();
  param = do_expression();
  
  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  result = sin(convert_angle_in_radians(param));
  return result;
}

static gdouble do_function_cos()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
        
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  result = cos(convert_angle_in_radians(param));

  return result;
}

static gdouble do_function_tan()
{
  gdouble result;
  gdouble param;
  
  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
        
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();
  
  /* tan = sin / cos; beware of cos !! */
  param = convert_angle_in_radians(param);   /* get angle in radians */
  if (cos(param) == 0.0)
    {
      set_error(EXPRPARSER_DOMAIN_ERROR);
      return 0;
    }

  result = tan(param);
  
  return result;
}

static gdouble do_function_cot()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
        
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  /* cot = cos / sin; beware of sin !! */
  param = convert_angle_in_radians(param);   /* get angle in radians */
  if (sin(param) == 0.0)
    {
      set_error(EXPRPARSER_DOMAIN_ERROR);
      return 0;
    }
  
  result = 1.0 / tan(param);
  
  return result;
}

static gdouble do_function_asin()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
        
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();
  
  if (fabs(param) > 1.0)  /* asin(x), |x| <= 1 */
    {
      set_error(EXPRPARSER_DOMAIN_ERROR);
      return 0;
    }
  
  result = convert_angle_in_curr_mode(asin(param));

  return result;
}

static gdouble do_function_acos()
{ 
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
        
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  if (fabs(param) > 1.0)  /* acos(x), |x| <= 1 */
    {
      set_error(EXPRPARSER_DOMAIN_ERROR);
      return 0;
    }

  result = convert_angle_in_curr_mode(acos(param));
  
        return result;
}

static gdouble do_function_atan()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
  
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();
  
  result = convert_angle_in_curr_mode(atan(param));
  
  return result;
}

static gdouble do_function_sqrt()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
  
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();
  
  if (param < 0.0)                                /* sqrt(x), x >= 0 */
    {
      set_error(EXPRPARSER_DOMAIN_ERROR);
      return 0;
    }
  
  result = sqrt(param);
  
  return result;
}

static gdouble do_function_sqr()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
  
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  result = param*param;

  return result;
}

static gdouble do_function_pow()
{
  gdouble result;
  gdouble param1, param2;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
  
  go_ahead();
  param1 = do_expression();
  if (*m_curr_char != ',')
    {
      set_error(EXPRPARSER_COMMA_EXPECTED);
      return 0;
    }
  
  go_ahead();
  skip_blanks();
  param2 = do_expression();
  
  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  result = pow(param1, param2);

  return result;
}

static gdouble do_function_exp()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       // dummy value
    }
  
  go_ahead();
  param = do_expression();

  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  result = exp(param);
  
  return result;
}

static gdouble do_function_ln()
{
  gdouble result;
  gdouble param;

  skip_blanks();
  if (*m_curr_char != '(')
    {
      set_error(EXPRPARSER_OPEN_PARENTHESIS_EXPECTED);
      return 0;       /* dummy value */
    }
  
  go_ahead();
  param = do_expression();
	
  if (*m_curr_char != ')')
    {
      set_error(EXPRPARSER_CLOSE_PARENTHESIS_EXPECTED);
      return 0;
    }
  go_ahead();
  skip_blanks();

  if (param <= 0)         /* log(x); x > 0 ! */
    {
      set_error(EXPRPARSER_DOMAIN_ERROR);
      return 0;
    }
        
  result = log(param);
  
  return result;
}














