/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This file was mainly written by Miloslav Cempirek. 
   It was added to the oCADis-project 2000-10-28. */

/* More functions by Miloslav Cempirek was added 
   2001-01-26. */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include "fcmp.h"

#include "struct.h"
#include "misc.h"
#include "obj_elarc.h"

/* --  F U N C T I O N S  -- */

/*-----------------------------------------------*/
/*  Sorting of characters in string (bubble sort) */

static void sort_string(gchar *str)
{
  gint  i,j;
  gint  n=strlen(str);
  gchar ch;

  for(i=0;i<=n-1;i++)
   for(j=n-1;j>i;j--)
    if(str[j-1]>str[j]) {ch=str[j];
                         str[j]=str[j-1];
                         str[j-1]=ch;}
}


/*--------------------------------------*/
/* return -1 if sign is negativ, otherwise return +1; */

static gdouble sgn(double x) {if (x<0) return -1.; else return +1;}

/*--------------------------------------*/
/* solution of quadratic equation       */  
/*--------------------------------------*/

static void quadratic_equation(gdouble A,    /*  coefficients     */
			       gdouble B,
			       gdouble C,
			       gdouble *x1,  /*  solutions        */
			       gdouble *x2,
			       gint    *ns)  /*  number of solutions */
{
  gdouble dis,q;

  if (fcmp(A,0,FLT_EPSILON)==0) 
  {  
    if (fcmp(B,0,FLT_EPSILON)==0) {(*ns)=0; return;}  
    (*x1)=-C/B; 
    (*ns)=1;
    return;
  }
  
  dis=B*B-4.*A*C;
  
  if (fcmp(dis,0,FLT_EPSILON)==0) 
  {
    (*x1)=-B/2./A;
    (*ns)=1;
    return;
  } 
  
  if (dis>0) 
  {
    q = -(B+sgn(B)*sqrt(dis))/2.;
    (*x1)=q/A; (*x2)=C/q;
    (*ns)=2;
    return;    
  }
  else
  {
    (*ns)=0;
  }  
}

/*--------------------------------------*/
/* distance between two point in 2D     */  
/*--------------------------------------*/

static gdouble point_point_dist2D(gdouble  x1,
				  gdouble  y1,
				  gdouble  x2,
				  gdouble  y2)
{
   gdouble  dx,dy;
   dx=x2-x1;
   dy=y2-y1;

   return  sqrt(dx*dx+dy*dy);
}

/*-------------------------------------------------------------*/ 
/*  distance between point [x3,y3] and line from point 1 to 2  */
/*  dist>0  - point lies right                                 */
/*  dist=0  - point lies on line                               */
/*  dist<0  - point lies left                                  */
/*-------------------------------------------------------------*/
/*  comp.graphics.algorithms Frequently Asked Questions        */

static gdouble line_point_dist2D(gdouble  x1,    /* line */
				 gdouble  y1,
				 gdouble  x2,
				 gdouble  y2,
				 gdouble  x3,    /* point */
				 gdouble  y3
				 )
{ 
  gdouble l=point_point_dist2D(x1,y1,x2,y2);
  
  return ((y1-y3)*(x2-x1)-(x1-x3)*(y2-y1))/l;
}
  
/* -------------------------------------------------------------*/
/*      discard circles that lie on given circle                */
/* -------------------------------------------------------------*/

static void discard_the_same(gdouble xc,
			     gdouble yc,
			     gdouble rc,
			     gdouble *x,     /* field of circles  */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)   /* number of solutions */
{
  gint i=0;
  gint j;
  gdouble diff=0.0001;
  while (i<(*cnt))
  {
    if ((fcmp(xc,x[i],diff) == 0)  && 
        (fcmp(yc,y[i],diff) == 0)  &&
        (fcmp(rc,r[i],diff) == 0))
    {
       for (j=i+1;j<(*cnt);j++)
       {
         x[j-1]=x[j]; y[j-1]=y[j]; r[j-1]=r[j];
       }
       (*cnt)--;
    }            
    else
    { 
      i++;
    }
  } 
}

/* -------------------------------------------------------------*/
/*          circle through 3 points                            */
/* -------------------------------------------------------------*/

static void apolloniusPPP(gdouble x1,      /* point  */
			  gdouble y1,
			  gdouble x2,      /* point  */   
			  gdouble y2,
			  gdouble x3,      /* point  */   
			  gdouble y3,
			  gdouble *x,      /* field of circles */
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)    /* number of solutions */
{
  gdouble  a,b,d;
  gdouble  a1,b1,d1;
  gdouble  K0;

  a=2*(x1-x2);
  b=2*(y1-y2);
  d=(x1*x1+y1*y1)-(x2*x2+y2*y2);
  
  a1=2*(x1-x3);
  b1=2*(y1-y3);
  d1=(x1*x1+y1*y1)-(x3*x3+y3*y3);
  
  K0=(a*b1-b*a1);
  if (fcmp(K0,0,FLT_EPSILON)==0) {(*cnt)=0; return;}   /* K0==0 */
	  
  x[0]=(b1*d-b*d1)/K0;
  y[0]=(-a1*d+a*d1)/K0;
  r[0]=point_point_dist2D(x[0],y[0],x1,y1);

  (*cnt)=1;  
}

/* -------------------------------------------------------------*/
/* circle through 2 point and tangent to line                  */
/* -------------------------------------------------------------*/

static void apolloniusLPP(gdouble x1,        /* point */
			  gdouble y1,
			  gdouble x2,        /* point */   
			  gdouble y2,
			  gdouble xl1,       /* line  */
			  gdouble yl1,
			  gdouble xl2,
			  gdouble yl2,
			  gdouble *x,        /* field of circles     */  
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)      /* number of solutions */
{
  gdouble  A,B,C,na;
  gdouble  a,b,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a;
  gdouble  T,U,V;
  gdouble  zn1;
  gint     i,h,m;
  gdouble  rr[2];
  gint     ns;
  
  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1)-(x2*x2+y2*y2);
  
  A=yl1-yl2;
  B=xl2-xl1;
  C=-yl1*B-xl1*A;

  na=sqrt(A*A+B*B);
  a1=A;
  b1=B;
  d1=-C;
    
  K0=(a*b1-b*a1);

  if (fcmp(K0,0,FLT_EPSILON)==0) {(*cnt)=0; return;}   /* K0==0  */
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  (*cnt)=0;
  h=0;
  for (i=0;i<2;i++)
  {
    if (i==0) zn1=-1; else zn1=+1;
    c1=zn1*na;
    
    K2= b*c1/K0;
    K4=-a*c1/K0;
    T=(K2*K2+K4*K4-1);
    U=2.*(K1a*K2+K3a*K4);
    V=K1a*K1a+K3a*K3a;

    quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

    for (m=0;m<ns;m++)
    {
      if (rr[m]>0)
      { 
        r[h]  = rr[m]; 
        x[h]  = K1 + K2*rr[m]; 
        y[h]  = K3 + K4*rr[m];
        h++;
      }  
    }
      
  }
  (*cnt)=h;
}


/* -------------------------------------------------------------*/
/* circle through 2 point and tangent to circle                */
/* -------------------------------------------------------------*/

static void apolloniusCPP(gdouble x1,    /* point */
			  gdouble y1,
			  gdouble x2,    /* point */  
			  gdouble y2,
			  gdouble x3,    /* circle */
			  gdouble y3,
			  gdouble r3,
			  gdouble *x,    /* field of circles */  
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)  /* number of solutions */
{
  gdouble  a,b,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a;
  gdouble  T,U,V;
  gdouble  zn3;
  gint     i,h,m;
  gdouble  rr[2];
  gint     ns;

  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1)-(x2*x2+y2*y2);
  
  a1=2.*(x1-x3);
  b1=2.*(y1-y3);
  d1=(x1*x1+y1*y1)-(x3*x3+y3*y3-r3*r3);
  
  K0=(a*b1-b*a1);
  if (fcmp(K0,0,FLT_EPSILON)==0) {(*cnt)=0; return;}   /* K0==0  */
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  h=0;
  for (i=0;i<2;i++)
  {
    if (i==0) zn3=-1; else zn3=+1;
    c1=2.*zn3*r3;
    
    K2= b*c1/K0;
    K4=-a*c1/K0;
    T=(K2*K2+K4*K4-1);
    U=2.*(K1a*K2+K3a*K4);
    V=K1a*K1a+K3a*K3a;

    quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

    for (m=0;m<ns;m++)
    {
      if (rr[m]>0)
      { 
        r[h]  = rr[m]; 
        x[h]  = K1 + K2*rr[m]; 
        y[h]  = K3 + K4*rr[m];
        h++;
      }  
    }
    (*cnt)=h;
    discard_the_same(x3,y3,r3,x,y,r,cnt);
  }
}


/* -------------------------------------------------------------*/
/* circle through point, tangent to line and circle            */
/* -------------------------------------------------------------*/

static void apolloniusCLP(gdouble x1,      /* point */
			  gdouble y1,
			  gdouble x2a,     /* line  */
			  gdouble y2a,
			  gdouble x2b,
			  gdouble y2b,
			  gdouble x3,      /* circle */   
			  gdouble y3,
			  gdouble r3,
			  gdouble *x,      /* field of circles */    
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt) /* number of solutions */
{
  gdouble  A,B,C,nas;
  gdouble  a,b,c,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a,T,U,V;
  gdouble  zn2,zn3;
  gint     j,k,h,m;  
  gdouble  rr[2];  
  gint     ns;

  a=2.*(x1-x3);
  b=2.*(y1-y3);
  d=(x1*x1+y1*y1)-(x3*x3+y3*y3-r3*r3);

  A=y2a-y2b;
  B=x2b-x2a;
  C=-y2a*B-x2a*A;

  nas=sqrt(A*A+B*B);
  a1=A;
  b1=B;
  d1=-C;

  K0=(a*b1-b*a1);
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  h=0;
  for (j=0;j<2;j++)
  {
    if (j==0) zn2=-1; else zn2=+1;
    c=2*zn2*r3;
      
    for (k=0;k<2;k++)
    {    
      if (k==0) zn3=-1; else zn3=+1;
      c1=zn3*nas;     
             
      K2=(-b1*c+b*c1)/K0;
      K4=( a1*c-a*c1)/K0;
      T=(K2*K2+K4*K4-1);
      U=2.*(K1a*K2+K3a*K4);
      V=K1a*K1a+K3a*K3a;

      quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

      for (m=0;m<ns;m++)
      {
        if (rr[m]>0)
        { 
          r[h]  = rr[m]; 
          x[h]  = K1 + K2*rr[m]; 
          y[h]  = K3 + K4*rr[m];
          h++;
        }  
      }

    } 
  }
  (*cnt)=h;
  discard_the_same(x3,y3,r3,x,y,r,cnt);
}


/* -------------------------------------------------------------*/
/* circle through point and tangent to two circles             */
/* -------------------------------------------------------------*/

static void apolloniusCCP(gdouble x1,     /* point */     
			  gdouble y1,
			  gdouble x2,     /* circle */     
			  gdouble y2,
			  gdouble r2,
			  gdouble x3,     /* circle */
			  gdouble y3,
			  gdouble r3,
			  gdouble *x,     /* field of circles */    
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)   /* number of solutions */  
{
  gdouble  a,b,c,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a,T,U,V;
  gint     j,k,h,m;
  gdouble  zn2,zn3;
  gdouble  rr[2];  
  gint     ns;

  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1)-(x2*x2+y2*y2-r2*r2);

  a1=2.*(x1-x3);
  b1=2.*(y1-y3);
  d1=(x1*x1+y1*y1)-(x3*x3+y3*y3-r3*r3);

  K0=(a*b1-b*a1);
  if (fcmp(K0,0,FLT_EPSILON)==0) {(*cnt)=0; return;}   /* K0==0 */
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  h=0;
  
  for (j=0;j<2;j++)
  {
    if (j==0) zn2=-1; else zn2=+1;
    c=2*zn2*r2;
    for (k=0;k<2;k++)
    {    
      if (k==0) zn3=-1; else zn3=+1;
      c1=2*zn3*r3;     
           
      K2=(-b1*c+b*c1)/K0;
      K4=( a1*c-a*c1)/K0;
      T=(K2*K2+K4*K4-1);
      U=2.*(K1a*K2+K3a*K4);
      V=K1a*K1a+K3a*K3a;

      quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

      for (m=0;m<ns;m++)
      {
        if (rr[m]>0)
        { 
          r[h]  = rr[m]; 
          x[h]  = K1 + K2*rr[m]; 
          y[h]  = K3 + K4*rr[m];
          h++;
        }  
      }
   
    }
  }
  (*cnt)=h;  

  discard_the_same(x2,y2,r2,x,y,r,cnt);
  discard_the_same(x3,y3,r3,x,y,r,cnt);
}

/* -------------------------------------------------------------*/
/* circle tangent to line and two circles                       */
/* -------------------------------------------------------------*/

static void apolloniusCCL(gdouble x1,    /* circle */
			  gdouble y1,
			  gdouble r1,
			  gdouble x2,    /* circle */
			  gdouble y2,
			  gdouble r2,
			  gdouble x3a,   /* line   */   
			  gdouble y3a,
			  gdouble x3b,
			  gdouble y3b,
			  gdouble *x,    /* field of circles */    
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)  /* number of solutions */
{
  gdouble  A,B,C,nas;
  gdouble  a,b,c,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a,T,U,V;
  gdouble  zn1,zn2,zn3;
  gint     i,j,k,h,m;
  gdouble  rr[2];  
  gint     ns;

  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1-r1*r1)-(x2*x2+y2*y2-r2*r2);

  A=y3a-y3b;
  B=x3b-x3a;
  C=-y3a*B-x3a*A;

  nas=sqrt(A*A+B*B);
  a1=A;
  b1=B;
  d1=-C;

  K0=(a*b1-b*a1);
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  h=0;
  for (i=0;i<2;i++)
  {
    if (i==0) zn1=-1; else zn1=+1;
    for (j=0;j<2;j++)
    {
      if (j==0) zn2=-1; else zn2=+1;
      c=2*(zn1*r1-zn2*r2);
      
      for (k=0;k<2;k++)
      {    
        if (k==0) zn3=-1; else zn3=+1;
        c1=zn3*nas;     
             
        K2=(-b1*c+b*c1)/K0;
        K4=( a1*c-a*c1)/K0;
        T=(K2*K2+K4*K4-1);
        U=2.*(K1a*K2+K3a*K4-zn1*r1);
        V=K1a*K1a+K3a*K3a-r1*r1;

        quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

        for (m=0;m<ns;m++)
        {
          if (rr[m]>0)
          { 
            r[h]  = rr[m]; 
            x[h]  = K1 + K2*rr[m]; 
            y[h]  = K3 + K4*rr[m];
            h++;
          }  
        }
     
      }
    }
  }
  (*cnt)=h;

  discard_the_same(x1,y1,r1,x,y,r,cnt);
  discard_the_same(x2,y2,r2,x,y,r,cnt);
}



/* -------------------------------------------------------------*/
/* circle tangent to circle and two lines                       */
/* -------------------------------------------------------------*/

static void apolloniusCLL(gdouble x1,     /* circle  */
			  gdouble y1,
			  gdouble r1,
			  gdouble x2a,    /* line    */
			  gdouble y2a,
			  gdouble x2b,
			  gdouble y2b,
			  gdouble x3a,    /* line   */   
			  gdouble y3a,
			  gdouble x3b,
			  gdouble y3b,
			  gdouble *x,     /* field of circles */  
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)   /* number of solutions */
{
  gdouble  na,nb;
  gdouble  a,b,c,d;
  
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a,T,U,V;
  gdouble  zn1,zn2,zn3;
  gint     i,j,k,h,m;
  gdouble  rr[2],xx[2],yy[2],rhlp;  
  gint     ns;
  gdouble  A1,B1,C1,A2,B2,C2;
  gdouble  v1,v2;               /* distance between centres */ 

  h=0;

  a=y2a-y2b;
  b=x2b-x2a;
  na=sqrt(a*a+b*b);
  d=y2a*b+x2a*a;

  a1=y3a-y3b;
  b1=x3b-x3a;
  nb=sqrt(a1*a1+b1*b1);
  d1=y3a*b1+x3a*a1;
  

  /*----------------*/ 
  K0=(a*b1-b*a1);
  if (fcmp(K0,0,FLT_EPSILON)==0)      /* lines are parallel */
  {
    A1=a;  B1=b;  C1=-d; 
    A2=a1; B2=b1; C2=-d1;             /* raduis = 1/2 of distance */    
    rhlp=fabs(C1/sqrt(A1*A1+B1*B1)-C2/sqrt(A2*A2+B2*B2))/2.;

    /* --- check if circle lies out of lines   ----*/
    v1=fabs(A1*x1+B1*y1+C1)/sqrt(A1*A1+B1*B1);
    v2=fabs(A2*x1+B2*y1+C2)/sqrt(A2*A2+B2*B2);
    if (((v1>2.*rhlp) || (v2>2.*rhlp)) && (v1>r1) && (v2>r1))
       {(*cnt)=0;  return;}


    if (fcmp(B1,0,FLT_EPSILON)==0)    /* zero dividing when B1=0 */ 
    {
      T=1;
      U=-2*y1;
      for (i=0;i<2;i++)
      {
        if (i==0) zn1=-1; else zn1=+1;       
        K1=(zn1*rhlp*sqrt(A1*A1+B1*B1)-C1)/A1;
        K1a=K1-x1;
        for (j=0;j<2;j++)
        {
          if (j==0) zn2=-1; else zn2=+1;       
          {
             K3=rhlp+zn2*r1;
             V=y1*y1-K3*K3+K1a*K1a;           
             
             quadratic_equation(T,U,V,&yy[0],&yy[1],&ns);
           
             for (m=0;m<ns;m++)
             {
               r[h]  = rhlp;
               x[h] = -(C1/A1+C2/A2)/2; 
               y[h]  = yy[m];
               if (fcmp(y[h],y1,FLT_EPSILON) !=0 )  h++; /* dont get y[h]=y1 */
             }
          }
        }     
      }
      (*cnt)=h;  return;
    }
    else      /* parallel, but not paralell with x or y  */
    {
       K2=-A1/B1;
       for (i=0;i<2;i++)
       {
         if (i==0) zn1=-1; else zn1=+1;       
         K1=(zn1*rhlp*sqrt(A1*A1+B1*B1)-C1)/B1;
         K1a=K1-y1;
         T=K2*K2+1;
         U=2.*(K1a*K2-x1);
         for (j=0;j<2;j++)
         {
           if (j==0) zn2=-1; else zn2=+1;       
           K3=rhlp+zn2*r1;
           V=x1*x1+K1a*K1a-K3*K3;
           quadratic_equation(T,U,V,&xx[0],&xx[1],&ns);
           
           for (m=0;m<ns;m++)
           {
             r[h]  = rhlp; 
             x[h]  = xx[m];
             y[h]  = K1 + K2*xx[m];   /* y[h] = -(C1/B1+C2/B2)/2; */

             /* coordinates have to satisfy second equation */
             /* dont get   x[h],y[h]=x1,y1 */
            
             v2=fabs(A2*x[h]+B2*y[h]+C2)/sqrt(A2*A2+B2*B2);  /* distance   */ 
             if (   (fcmp(v2,rhlp,FLT_EPSILON)==0)  
                 && ((fcmp(y[h],y1,FLT_EPSILON) !=0 ) || (fcmp(x[h],x1,FLT_EPSILON) !=0))   
                )   h++; 
           }
         }
       }
       (*cnt)=h;  return;
    }
    
  }   /* K0=0 */
  /*----------------*/


  /*----- not parallel -------*/                   
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  
  for (i=0;i<2;i++)
  {
    if (i==0) zn1=-1; else zn1=+1;
    c=zn1*na;
    for (j=0;j<2;j++)
    {
      if (j==0) zn2=-1; else zn2=+1;
      c1=zn2*nb;  
      for (k=0;k<2;k++)
      { 
        if (k==0) zn3=-1; else zn3=+1;
           
        K2=(-b1*c+b*c1)/K0;
        K4=( a1*c-a*c1)/K0;
        T=(K2*K2+K4*K4-1);
        U=2.*(K1a*K2+K3a*K4-zn3*r1);
        V=K1a*K1a+K3a*K3a-r1*r1;

        quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

        for (m=0;m<ns;m++)
        {
          if (rr[m]>0)
          { 
            r[h]  = rr[m]; 
            x[h]  = K1 + K2*rr[m]; 
            y[h]  = K3 + K4*rr[m];
            h++;
          }  
        }

      }
    }
  }
  (*cnt)=h;
  discard_the_same(x1,y1,r1,x,y,r,cnt);
}


/*-------------------------------------------------------------*/
/* circle tangent to thre lines                                */
/*-------------------------------------------------------------*/

static void apolloniusLLL(gdouble l1x1,    /* line */
			  gdouble l1y1,
			  gdouble l1x2,
			  gdouble l1y2,
			  gdouble l2x1,    /* line */
			  gdouble l2y1,
			  gdouble l2x2,
			  gdouble l2y2,
			  gdouble l3x1,    /* line */   
			  gdouble l3y1,
			  gdouble l3x2,
			  gdouble l3y2,
			  gdouble *x,      /* field of circles */  
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)    /* number of solutions */
{                   
   gdouble  A1,B1,C1,A2,B2,C2,A3,B3,C3;
   gdouble  del,U1,V1,W1,U2,V2,W2,U3,V3,W3;
   gdouble  m1,n1,o1,m2,n2,o2;
   gint     h;

   A1=l1y1-l1y2;
   B1=l1x2-l1x1;
   C1=-l1y1*B1-l1x1*A1;

   A2=l2y1-l2y2;
   B2=l2x2-l2x1;
   C2=-l2y1*B2-l2x1*A2;

   A3=l3y1-l3y2;
   B3=l3x2-l3x1;
   C3=-l3y1*B3-l3x1*A3;

   del=sqrt(A1*A1+B1*B1);
   U1=A1/del;
   V1=B1/del;
   W1=C1/del;
   
   del=sqrt(A2*A2+B2*B2);
   U2=A2/del;
   V2=B2/del;
   W2=C2/del;
   
   del=sqrt(A3*A3+B3*B3);
   U3=A3/del;
   V3=B3/del;
   W3=C3/del;
   
   
   /* 1. case   + + | + + */
   
   h=0;
   m1=U1-U2;   m2=U1-U3;
   n1=V1-V2;   n2=V1-V3;
   o1=W2-W1;   o2=W3-W1;
   
   del=m1*n2-m2*n1;
   
   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0 */
   {
     x[h]=(o1*n2-o2*n1)/del;
     y[h]=(m1*o2-m2*o1)/del;
     r[h]=fabs(line_point_dist2D(l1x1,l1y1,l1x2,l1y2,x[h],y[h]));
     h++;
   }
   
   /* 2. case   + - | + + */
   
   m1=U1+U2; 
   n1=V1+V2; 
   o1=-W2-W1; 
   
   del=m1*n2-m2*n1;
   
   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0  */
   {
     x[h]=(o1*n2-o2*n1)/del;
     y[h]=(m1*o2-m2*o1)/del;
     r[h]=fabs(line_point_dist2D(l1x1,l1y1,l1x2,l1y2,x[h],y[h]));
     h++;
   }
   
   /* 3. case   + - | + - */
   
   m2=U1+U3;
   n2=V1+V3;
   o2=-W3-W1;
   
   del=m1*n2-m2*n1;

   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0 */
   {
     x[h]=(o1*n2-o2*n1)/del;
     y[h]=(m1*o2-m2*o1)/del;
     r[h]=fabs(line_point_dist2D(l1x1,l1y1,l1x2,l1y2,x[h],y[h]));
     h++;
   }
      
   /* 4. case   + + | + -  */
   
   m1=U1-U2; 
   n1=V1-V2; 
   o1=W2-W1; 
   
   del=m1*n2-m2*n1;

   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0  */
   {
     x[h]=(o1*n2-o2*n1)/del;
     y[h]=(m1*o2-m2*o1)/del;
     r[h]=fabs(line_point_dist2D(l1x1,l1y1,l1x2,l1y2,x[h],y[h]));
     h++;
   }
   (*cnt)=h;
}


/* -------------------------------------------------------------*/
/* circle tangent to thre circles                               */
/* -------------------------------------------------------------*/

static void apolloniusCCC(gdouble x1,    /* circle */
			  gdouble y1,
			  gdouble r1,
			  gdouble x2,    /* circle */   
			  gdouble y2,
			  gdouble r2,
			  gdouble x3,    /* circle */   
			  gdouble y3,
			  gdouble r3,
			  gdouble *x,    /* field of circles */
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)
{
  gdouble  a,b,c,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a,T,U,V;
  gdouble  zn1,zn2,zn3;
  gint     i,j,k,h,m;
  gdouble  rr[2];  
  gint     ns;

  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1-r1*r1)-(x2*x2+y2*y2-r2*r2);

  a1=2.*(x1-x3);
  b1=2.*(y1-y3);
  d1=(x1*x1+y1*y1-r1*r1)-(x3*x3+y3*y3-r3*r3);

  K0=(a*b1-b*a1);
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;

  h=0;
  for (i=0;i<2;i++)
  {
    if (i==0) zn1=-1; else zn1=+1;
    for (j=0;j<2;j++)
    {
      if (j==0) zn2=-1; else zn2=+1;
      c=2*(zn1*r1-zn2*r2);
      for (k=0;k<2;k++)
      {    
        if (k==0) zn3=-1; else zn3=+1;
        c1=2*(zn1*r1-zn3*r3);     
             
        K2=(-b1*c+b*c1)/K0;
        K4=( a1*c-a*c1)/K0;
        T=(K2*K2+K4*K4-1);
        U=2.*(K1a*K2+K3a*K4-zn1*r1);
        V=K1a*K1a+K3a*K3a-r1*r1;

        quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);

        for (m=0;m<ns;m++)
        {
          if (rr[m]>0)
          { 
            r[h]  = rr[m]; 
            x[h]  = K1 + K2*rr[m]; 
            y[h]  = K3 + K4*rr[m];
            h++;
          }  
        }

      }
    }
  }
  (*cnt) = h;
  discard_the_same(x1,y1,r1,x,y,r,cnt);
  discard_the_same(x2,y2,r2,x,y,r,cnt);
  discard_the_same(x3,y3,r3,x,y,r,cnt);
}

/* -------------------------------------------------------------*/
/* circle tangent to point and two lines                        */
/* -------------------------------------------------------------*/

static void apolloniusLLP(gdouble x1,         /* point */
			  gdouble y1,
			  gdouble x2a,        /* line  */
			  gdouble y2a,       
			  gdouble x2b,
			  gdouble y2b,
			  gdouble x3a,        /* line  */
			  gdouble y3a,
			  gdouble x3b,
			  gdouble y3b,
			  gdouble *x,         /* field of circles  */
			  gdouble *y,
			  gdouble *r,
			  gint    *cnt)       /* number of solutions */ 

{
  gdouble  na,nb;
  gdouble  a,b,c,d;
  gdouble  a1,b1,c1,d1;
  gdouble  K0,K1,K2,K3,K4,K1a,K3a,T,U,V;
  gdouble  zn1,zn2,zn3;
  gint     i,j,k,h,m;
  gdouble  rr[2],xx[2],yy[2],rhlp;  
  gint     ns;
  gdouble  A1,B1,C1,A2,B2,C2;
  gdouble  v1,v2;        /* distances between point and lines  */ 

  h=0;

  a=y2a-y2b;
  b=x2b-x2a;
  na=sqrt(a*a+b*b);
  d=y2a*b+x2a*a;

  a1=y3a-y3b;
  b1=x3b-x3a;
  nb=sqrt(a1*a1+b1*b1);
  d1=y3a*b1+x3a*a1;
  
  /*----------------*/ 
  K0=(a*b1-b*a1);
  if (fcmp(K0,0,FLT_EPSILON)==0) /* lines are parallel */
  {
    A1=a;  B1=b;  C1=-d; 
    A2=a1; B2=b1; C2=-d1;             /* radius = 1/2 distance */    
    rhlp=fabs(C1/sqrt(A1*A1+B1*B1)-C2/sqrt(A2*A2+B2*B2))/2.;

    /* ---check if point lies between parallel lines */
    v1=fabs(A1*x1+B1*y1+C1)/sqrt(A1*A1+B1*B1);
    v2=fabs(A2*x1+B2*y1+C2)/sqrt(A2*A2+B2*B2);
    if ((v1>2.*rhlp) || (v2>2.*rhlp))  {(*cnt)=0;  return;}

    K3=rhlp;
    if (fcmp(B1,0,FLT_EPSILON) == 0)    /* zero dividing */
    {
      T=1;
      U=-2*y1;
      for (i=0;i<2;i++)
      {
        if (i==0) zn1=-1; else zn1=+1;            
        K1=(zn1*rhlp*sqrt(A1*A1+B1*B1)-C1)/A1;
        K1a=K1-x1;
        V=y1*y1-K3*K3+K1a*K1a;           
        
        quadratic_equation(T,U,V,&yy[0],&yy[1],&ns);
           
        for (m=0;m<ns;m++)
        {
          r[h]  = rhlp;
          x[h]  = K1;
          y[h]  = yy[m];
          h++;
        }
      }
    (*cnt)=h;  return;
    }
    else
    {
       K2=-A1/B1;
       for (i=0;i<2;i++)
       {
         if (i==0) zn1=-1; else zn1=+1;       
         K1=(zn1*rhlp*sqrt(A1*A1+B1*B1)-C1)/B1;
         K1a=K1-y1;

         T=K2*K2+1;
         U=2.*(K1a*K2-x1);
         V=x1*x1+K1a*K1a-K3*K3;
            
         quadratic_equation(T,U,V,&xx[0],&xx[1],&ns);
           
         for (m=0;m<ns;m++)
         {
            r[h]  = rhlp;
            x[h]  = xx[m]; 
            y[h]  = K1 + K2*xx[m];
            h++;
         }
       }
       (*cnt)=h;  return;
    
    }
    
  }   /* K0==0 */
  /*----------------*/
       
  K1=(b1*d-b*d1)/K0;
  K3=(-a1*d+a*d1)/K0;

  K1a=K1-x1;
  K3a=K3-y1;


  for (j=0;j<2;j++)
  {
    if (j==0) zn2=-1; else zn2=+1;    
    c=zn2*na;  
    for (k=0;k<2;k++)
    { 
      if (k==0) zn3=-1; else zn3=+1;
      c1=zn3*nb;
           
      K2=(-b1*c+b*c1)/K0;
      K4=( a1*c-a*c1)/K0;
      T=(K2*K2+K4*K4-1);
      U=2.*(K1a*K2+K3a*K4);
      V=K1a*K1a+K3a*K3a;
  
      quadratic_equation(T,U,V,&rr[0],&rr[1],&ns);
           
      for (m=0;m<ns;m++)
      {
        if (rr[m]>0)
        { 
          r[h]  = rr[m]; 
          x[h]  = K1 + K2*rr[m]; 
          y[h]  = K3 + K4*rr[m];
          h++;
        }  
      }
  
    }
  }
  (*cnt)=h;
}


/*--------------------------------------------*/

void math_find_nearest_circle(gdouble   *xc,    /* field of found circles */
			      gdouble   *yc,
			      gdouble   *rc,
			      gint      nsol,   /* number of found circles */
			      gdouble   x,      /* cursor coordinates */
			      gdouble   y,
			      gdouble   *x1c,     /* circle to draw */
			      gdouble   *y1c,
			      gdouble   *r1,
			      gint      *pos)   /* position in 0-nsol */
{
  gint     i,j;
  gdouble  dis,dis1;

  if (nsol==0) {(*pos)=-1; return;}
  if (nsol==1) {(*pos)=0; 
                (*x1c)=xc[0]; 
                (*y1c)=yc[0]; 
                (*r1)=rc[0]; 
                return;
               }
  
  dis=point_point_dist2D(x,y,xc[0],yc[0]);
  j=0;

  for (i=1;i<nsol;i++)
  {
     dis1=point_point_dist2D(x,y,xc[i],yc[i]);
     if (dis1<dis) {dis=dis1;j=i;}
  }
  (*x1c)=xc[j]; 
  (*y1c)=yc[j]; 
  (*r1)=rc[j];
  (*pos)=j;
}
                  
/*---------------------------------------------*/


void math_apollonius(OBJECT **obj,   /* selected objects */
		     gdouble *xo,     /* field of found circles */
		     gdouble *yo,
		     gdouble *ro,
		     gint    *nsol    /* number of solutions */
		     )
{
  gint       i;
  gdouble    xpi[3],ypi[3];          /* field for input values (point) */
  gdouble    xli[6],yli[6];          /* field for input values (line) */ 
  gdouble    xci[6],yci[6],rci[3];   /* field for input values (circle,arc)*/
  gint       up=0;                   /* actual position in fields */
  gint       ul=0;
  gint       uc=0;
  gint       ur=0;
  gchar      task[4];
  gint       ns;

  
  for (i=0;i<3;i++)
  {
     switch (obj[i]->type)
     {
     case  TYPE_POINT:
        xpi[up]=obj[i]->data.point.centrum.x;
        ypi[up]=obj[i]->data.point.centrum.y;
        up++;
        task[i]='P';
        break;
     case  TYPE_LINE: 
        xli[ul]=obj[i]->data.line.p1.x;;
        yli[ul]=obj[i]->data.line.p1.y;;
        ul++;
        xli[ul]=obj[i]->data.line.p2.x;;
        yli[ul]=obj[i]->data.line.p2.y;;
        ul++;
        task[i]='L';
        break;
     case  TYPE_CIRCLE: 
        xci[uc]=obj[i]->data.circle.centrum.x;
        yci[uc]=obj[i]->data.circle.centrum.y;
        uc++;
        rci[ur]=obj[i]->data.circle.radius;
        ur++;
        task[i]='C';
        break;
     case  TYPE_ARC:
        xci[uc]=obj[i]->data.arc.centrum.x;
        yci[uc]=obj[i]->data.arc.centrum.y;
        uc++;
        rci[ur]=obj[i]->data.arc.radius;
        ur++;
        task[i]='C';
        break;
     }
  }

  task[3]='\0';
  sort_string(task);     /*  alphabetical sorting   */

  if (strcmp(task,"PPP")==0) 
  {  
     apolloniusPPP(xpi[0],ypi[0],   /* first point  */
                   xpi[1],ypi[1],   /* second point */
                   xpi[2],ypi[2],   /* third point  */
                   xo,yo,ro,&ns);   /* number of solutions */
  }
  else if (strcmp(task,"LPP")==0)
  {
     apolloniusLPP(xpi[0],ypi[0],   /* first point */
                   xpi[1],ypi[1],   /* second point*/
                   xli[0],yli[0],   /* line */
                   xli[1],yli[1],
                   xo,yo,ro,&ns);   /* number of solution */
  } 
  else if (strcmp(task,"CPP")==0)
  {
     apolloniusCPP(xpi[0],ypi[0],   /* first point */
                   xpi[1],ypi[1],   /* second point */
                   xci[0],yci[0],   /* circle  */
                   rci[0],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"CLP")==0)
  {
     apolloniusCLP(xpi[0],ypi[0],   /* first point */
                   xli[0],yli[0],   /* line        */
                   xli[1],yli[1],
                   xci[0],yci[0],   /* circle  */
                   rci[0],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"CCP")==0)
  {
     apolloniusCCP(xpi[0],ypi[0],   /* first point*/
                   xci[0],yci[0],   /* circle  */
                   rci[0],
                   xci[1],yci[1],   /* circle  */
                   rci[1],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"CCL")==0)
  {
     apolloniusCCL(xci[0],yci[0],   /* circle  */
                   rci[0],
                   xci[1],yci[1],   /* circle  */
                   rci[1],
                   xli[0],yli[0],   /* line     */
                   xli[1],yli[1],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"CLL")==0)
  {
     apolloniusCLL(xci[0],yci[0],   /* circle   */
                   rci[0],
                   xli[0],yli[0],   /* line     */
                   xli[1],yli[1],
                   xli[2],yli[2],   /* line     */
                   xli[3],yli[3],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"LLL")==0)
  {
     apolloniusLLL(xli[0],yli[0],   /* line     */
                   xli[1],yli[1],
                   xli[2],yli[2],   /* line     */
                   xli[3],yli[3],
                   xli[4],yli[4],   /* line     */
                   xli[5],yli[5],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"CCC")==0)
  {
     apolloniusCCC(xci[0],yci[0],   /* circle  */
                   rci[0],
                   xci[1],yci[1],   /* circle  */
                   rci[1],
                   xci[2],yci[2],   /* circle  */
                   rci[2],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"LLP")==0)
  {
     apolloniusLLP(xpi[0],ypi[0],   /* point */
                   xli[0],yli[0],   /* line  */
                   xli[1],yli[1],
                   xli[2],yli[2],   /* line  */
                   xli[3],yli[3],
                   xo,yo,ro,&ns);   /* number of solutions */
  } 
  (*nsol)=ns;
}


/*===================================================================*/


/* -------------------------------------------------------------*/
/*      tangent circle with radius r0 to two circles            */
/* -------------------------------------------------------------*/

static void circle_tangentCC(gdouble r0,      /* radius of circle */
			     gdouble x1,      /* circle */
			     gdouble y1,
			     gdouble r1,
			     gdouble x2,      /* circle */
			     gdouble y2,
			     gdouble r2,
			     gdouble *x,      /* field of circles */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)    /* number of solutions */
{
  gdouble  a,b,c,d;
  gdouble  K1,K2,K2a,K3,T,U,V;
  gdouble  zn1,zn2;
  gdouble  ss[2];
  gint     i,j,h,m;
  gint     ns;

  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1-r1*r1)-(x2*x2+y2*y2-r2*r2);

  h=0;
  if (fcmp(b,0,FLT_EPSILON) != 0)         /* b=0 => dividing by zero */
  {
    K1=-a/b;
    T=1+K1*K1;

    for (i=0;i<2;i++)
    {
      if (i==0) zn1=-1; else zn1=+1;
      K3=r0-zn1*r1;        
      
      for (j=0;j<2;j++)
      {
        if (j==0) zn2=-1; else zn2=+1;
        c=2*r0*(zn1*r1-zn2*r2)+d;
        K2=c/b;
        K2a=K2-y1;
        U=2*(K1*K2a-x1);
        V=x1*x1+K2a*K2a-K3*K3;

        quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

        for (m=0;m<ns;m++)
        {
          r[h]  = r0; 
          x[h]  = ss[m];
          y[h]  = K2 + K1*ss[m];
          h++;
        }

      }
    }
  }
  else
  {
    K1=-b/a;
    T=1+K1*K1;

    for (i=0;i<2;i++)
    {
      if (i==0) zn1=-1; else zn1=+1;
      K3=r0-zn1*r1;        
      
      for (j=0;j<2;j++)
      {
        if (j==0) zn2=-1; else zn2=+1;
        c=2*r0*(zn1*r1-zn2*r2)+d;
        K2=c/a;
        K2a=K2-x1;
        U=2*(K1*K2a-y1);
        V=y1*y1+K2a*K2a-K3*K3;

        quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

        for (m=0;m<ns;m++)
        {
          r[h]  = r0; 
          x[h]  = K2 + K1*ss[m];
          y[h]  = ss[m];
          h++;
        }

      }
    }
  }
  (*cnt) = h;
  
  discard_the_same(x1,y1,r1,x,y,r,cnt);
  discard_the_same(x2,y2,r2,x,y,r,cnt);
}

/* ------------------------------------------------------------- */
/* tangent circle with radius r0 to circle and line              */
/* ------------------------------------------------------------- */


static void circle_tangentCL(gdouble r0,      /* radius of circle */
			     gdouble x1,      /* circle */
			     gdouble y1,
			     gdouble r1,
			     gdouble l1x1,    /* line */
			     gdouble l1y1,
			     gdouble l1x2,
			     gdouble l1y2,
			     gdouble *x,      /* field of circles */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)    /* number of solutions */
{
  gdouble  A,B,C,del;
  gdouble  K1,K2,K2a,K3,T,U,V;
  gdouble  zn1,zn2;
  gdouble  ss[2];
  gint     i,j,h,m;
  gint     ns;

  A=l1y1-l1y2;
  B=l1x2-l1x1;
  C=-l1y1*B-l1x1*A;
  del=sqrt(A*A+B*B);

  h=0;
  if (fcmp(B,0,FLT_EPSILON) != 0)         /* b=0 => dividing by zero */
  {
    K1=-A/B;
    T=1+K1*K1;
    for(i=0;i<2;i++)
    {
      if (i==0) zn2=-1; else zn2=+1;  
      K2=(zn2*r0*del-C)/B;
      K2a=K2-y1;
      U=2*(K1*K2a-x1);
      
      for (j=0;j<2;j++)
      {
        if (j==0) zn1=-1; else zn1=+1;
        K3=r0-zn1*r1;
        V=x1*x1+K2a*K2a-K3*K3;
      
        quadratic_equation(T,U,V,&ss[0],&ss[1],&ns); 

        for (m=0;m<ns;m++)
        {
          r[h]  = r0; 
          x[h]  = ss[m];
          y[h]  = K2 + K1*ss[m];
          h++;
        }

      }
    }
  }
  else
  {
    K1=-B/A;
    T=1+K1*K1;
    for(i=0;i<2;i++)
    {
      if (i==0) zn2=-1; else zn2=+1;  
      K2=(zn2*r0*del-C)/A;
      K2a=K2-x1;
      U=2*(K1*K2a-y1);
      
      for (j=0;j<2;j++)
      {
        if (j==0) zn1=-1; else zn1=+1;
        K3=r0-zn1*r1;
        V=y1*y1+K2a*K2a-K3*K3;
      
        quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

        for (m=0;m<ns;m++)
        {
          r[h]  = r0; 
          x[h]  = K2 + K1*ss[m];
          y[h]  = ss[m];
          h++;
        }

      }
    }
  }
  *cnt = h;

  discard_the_same(x1,y1,r1,x,y,r,cnt);
}


/* ------------------------------------------------------------- */
/* tangent circle with radius r0 to  circle and point            */
/* ------------------------------------------------------------- */

static void circle_tangentCP(gdouble r0,      /* radius of circle */
			     gdouble x1,      /* circle */
			     gdouble y1,
			     gdouble r1,
			     gdouble x2,      /* point */
			     gdouble y2,
			     gdouble *x,      /* field of circles */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)    /* number of solutions */
{
  gdouble  a,b,c,d;
  gdouble  K1,K2,K2a,K3,T,U,V;
  gdouble  zn1;
  gdouble  ss[2];
  gint     i,h,m;
  gint     ns;


  a=2.*(x1-x2);
  b=2.*(y1-y2);
  d=(x1*x1+y1*y1-r1*r1)-(x2*x2+y2*y2);

  h=0;
  if (fcmp(b,0,FLT_EPSILON) != 0)         /* b=0 => dividing buy zero  */
  {
    K1=-a/b;
    T=1+K1*K1;

    for (i=0;i<2;i++)
    {
      if (i==0) zn1=-1; else zn1=+1;
      K3=r0-zn1*r1;        
      
      c=2*r0*zn1*r1+d;
      K2=c/b;
      K2a=K2-y1;
      U=2*(K1*K2a-x1);
      V=x1*x1+K2a*K2a-K3*K3;

      quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

      for (m=0;m<ns;m++)
      {
        r[h]  = r0; 
        x[h]  = ss[m];
        y[h]  = K2 + K1*ss[m];
        h++;
      }
    }
  }
  else
  {
    K1=-b/a;
    T=1+K1*K1;

    for (i=0;i<2;i++)
    {
      if (i==0) zn1=-1; else zn1=+1;
      K3=r0-zn1*r1;        
     
      c=2*r0*zn1*r1+d;
      K2=c/a;
      K2a=K2-x1;
      U=2*(K1*K2a-y1);
      V=y1*y1+K2a*K2a-K3*K3;

      quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

      for (m=0;m<ns;m++)
      {
        r[h]  = r0; 
        x[h]  = K2 + K1*ss[m];
        y[h]  = ss[m];
        h++;
      }

    }
  }
  *cnt = h;

  discard_the_same(x1,y1,r1,x,y,r,cnt);
}

/* ------------------------------------------------------------- */
/* tangent circle with radius r0 to two lines                    */
/* ------------------------------------------------------------- */

static void circle_tangentLL(gdouble r0,      /* radius of circle */
			     gdouble l1x1,    /* line */
			     gdouble l1y1,
			     gdouble l1x2,
			     gdouble l1y2,
			     gdouble l2x1,    /* line */
			     gdouble l2y1,
			     gdouble l2x2,
			     gdouble l2y2,
			     gdouble *x,      /* field of circles */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)    /* number of solutions */
{
   gdouble  A1,B1,C1,A2,B2,C2;
   gdouble  del,del1,del2,T1,T2;
   gint     h;

   A1=l1y1-l1y2;
   B1=l1x2-l1x1;
   C1=-l1y1*B1-l1x1*A1;
   del1=sqrt(A1*A1+B1*B1);

   A2=l2y1-l2y2;
   B2=l2x2-l2x1;
   C2=-l2y1*B2-l2x1*A2;
   del2=sqrt(A2*A2+B2*B2);

   h=0;
     
   /* 1. case  + | +  */
 
   T1=r0*del1-C1;   T2=r0*del2-C2;
   del=A1*B2-A2*B1;
   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0 */
   {
     x[h]=(B2*T1-B1*T2)/del;
     y[h]=(A1*T2-A2*T1)/del;
     r[h]=r0;
     h++;
   }
   
   /* 2. pripad   + | - */

   T1=r0*del1-C1;   T2=-r0*del2-C2;
   del=A1*B2-A2*B1;
   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0 */
   {
     x[h]=(B2*T1-B1*T2)/del;
     y[h]=(A1*T2-A2*T1)/del;
     r[h]=r0;
     h++;
   }
   
   /* 3. pripad   - | + */

   T1=-r0*del1-C1;   T2=r0*del2-C2;
   del=A1*B2-A2*B1;
   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0 */
   {
     x[h]=(B2*T1-B1*T2)/del;
     y[h]=(A1*T2-A2*T1)/del;
     r[h]=r0;
     h++;
   }
   
   /* 4. pripad   - | - */
   
   T1=-r0*del1-C1;   T2=-r0*del2-C2;
   del=A1*B2-A2*B1;
   if (fcmp(del,0,FLT_EPSILON) != 0)  /* del != 0 */
   {
     x[h]=(B2*T1-B1*T2)/del;
     y[h]=(A1*T2-A2*T1)/del;
     r[h]=r0;
     h++;
   }

   (*cnt)=h;
}


/* ------------------------------------------------------------- */
/* tangent circle with radius r0 to line and point               */
/* ------------------------------------------------------------- */

static void circle_tangentLP(gdouble r0,      /* radius of circle */
			     gdouble l1x1,    /* line */
			     gdouble l1y1,
			     gdouble l1x2,
			     gdouble l1y2,
			     gdouble x1,      /* point */
			     gdouble y1, 
			     gdouble *x,      /* field of circles */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)    /* number of solutions */
{
  gdouble  A,B,C,del;
  gdouble  K1,K2,K2a,T,U,V;
  gdouble  zn1;
  gdouble  ss[2];
  gint     i,h,m;
  gint     ns;

  A=l1y1-l1y2;
  B=l1x2-l1x1;
  C=-l1y1*B-l1x1*A;
  del=sqrt(A*A+B*B);

  h=0;
  if (fcmp(B,0,FLT_EPSILON) != 0)         /* b=0 => dividing by zero */
  {
    K1=-A/B;
    T=1+K1*K1;
    for(i=0;i<2;i++)
    {
      if (i==0) zn1=-1; else zn1=+1;  
      K2=(zn1*r0*del-C)/B;
      K2a=K2-y1;
      U=2*(K1*K2a-x1);
      V=x1*x1+K2a*K2a-r0*r0;
      
      quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

      for (m=0;m<ns;m++)
      {   
          r[h]  = r0; 
          x[h]  = ss[m];
          y[h]  = K2 + K1*ss[m];
          h++;
      }
    }
  }
  else
  {
    K1=-B/A;
    T=1+K1*K1;
    for(i=0;i<2;i++)
    {
      if (i==0) zn1=-1; else zn1=+1;  
      K2=(zn1*r0*del-C)/A;
      K2a=K2-x1;
      U=2*(K1*K2a-y1);
      V=y1*y1+K2a*K2a-r0*r0;
      
      quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

      for (m=0;m<ns;m++)
      {
        r[h]  = r0; 
        x[h]  = K2 + K1*ss[m];
        y[h]  = ss[m];
        h++;
      }
    }
  }
  (*cnt) = h;
}


/* ------------------------------------------------------------- */
/* tangent circle with radius r0 to two points                   */
/* ------------------------------------------------------------- */

static void circle_tangentPP(gdouble r0,      /* radius of circle */
			     gdouble x1,      /* point */
			     gdouble y1,
			     gdouble x2,      /* point */
			     gdouble y2,
			     gdouble *x,      /* field of circles */
			     gdouble *y,
			     gdouble *r,
			     gint    *cnt)    /* number of solutions */
{
  gdouble  a,b,c;
  gdouble  K1,K2,K2a,T,U,V;
  gdouble  ss[2];
  gint     h,m;
  gint     ns;

  a=2.*(x1-x2);
  b=2.*(y1-y2);
  c=(x1*x1+y1*y1)-(x2*x2+y2*y2);

  h=0;
  if (fcmp(b,0,FLT_EPSILON) != 0)         /* b=0 => dividing by zero */
  {
    K1=-a/b;
    K2=c/b;
    K2a=K2-y1;

    T=1+K1*K1;
    U=2*(K1*K2a-x1);
    V=x1*x1+K2a*K2a-r0*r0;

    quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

    for (m=0;m<ns;m++)
    {
      r[h]  = r0; 
      x[h]  = ss[m]; 
      y[h]  = K2 + K1*ss[m];
      h++;
    }
  }
  else
  {
    K1=-b/a;
    K2=c/a;
    K2a=K2-x1;

    T=1+K1*K1;
    U=2*(K1*K2a-y1);
    V=y1*y1+K2a*K2a-r0*r0;

    quadratic_equation(T,U,V,&ss[0],&ss[1],&ns);

    for (m=0;m<ns;m++)
    {
      r[h]  = r0; 
      x[h]  = K2 + K1*ss[m];
      y[h]  = ss[m];
      h++;
    }

  }
  (*cnt) = h;
}


/*---------------------------------------------*/


void math_circle_tangent(gdouble r0,      /* radius of circle */
			 OBJECT **obj,   /* selected objects */
			 gdouble *xo,     /* field of found circles */
			 gdouble *yo,
			 gdouble *ro,
			 gint    *nsol    /* number of solutions */
			 )
{
  gint       i;
  gdouble    xpi[2],ypi[2];          /* field for input values (point) */
  gdouble    xli[4],yli[4];          /* field for input values (line) */ 
  gdouble    xci[4],yci[4],rci[2];   /* field for input values (circle,arc)*/
  gint       up=0;                   /* actual position in fields */
  gint       ul=0;
  gint       uc=0;
  gint       ur=0;
  gchar      task[3];
  gint       ns;

  
  for (i=0;i<2;i++)
  {
     switch (obj[i]->type)
     {
     case  TYPE_POINT:
        xpi[up]=obj[i]->data.point.centrum.x;
        ypi[up]=obj[i]->data.point.centrum.y;
        up++;
        task[i]='P';
        break;
     case  TYPE_LINE: 
        xli[ul]=obj[i]->data.line.p1.x;;
        yli[ul]=obj[i]->data.line.p1.y;;
        ul++;
        xli[ul]=obj[i]->data.line.p2.x;;
        yli[ul]=obj[i]->data.line.p2.y;;
        ul++;
        task[i]='L';
        break;
     case  TYPE_CIRCLE: 
        xci[uc]=obj[i]->data.circle.centrum.x;
        yci[uc]=obj[i]->data.circle.centrum.y;
        uc++;
        rci[ur]=obj[i]->data.circle.radius;
        ur++;
        task[i]='C';
        break;
     case  TYPE_ARC:
        xci[uc]=obj[i]->data.arc.centrum.x;
        yci[uc]=obj[i]->data.arc.centrum.y;
        uc++;
        rci[ur]=obj[i]->data.arc.radius;
        ur++;
        task[i]='C';
        break;
     }
  }

  task[2]='\0';
  sort_string(task);     /*  alphabetical sorting   */

  if (strcmp(task,"CC")==0) 
  {  
     circle_tangentCC(r0,              /* radius of circle */
                      xci[0],yci[0],   /* circle  */
                      rci[0],
                      xci[1],yci[1],   /* circle  */
                      rci[1],
                      xo,yo,ro,&ns);   /* number of solutions */
  }
  else if (strcmp(task,"CL")==0)
  {
     circle_tangentCL(r0,              /* radius of circle */
                      xci[0],yci[0],   /* circle  */
                      rci[0],
                      xli[0],yli[0],   /* line    */
                      xli[1],yli[1],
                      xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"CP")==0)
  {
     circle_tangentCP(r0,              /* radius of circle */
                      xci[0],yci[0],   /* circle  */
                      rci[0],
                      xpi[0],ypi[0],   /* point */
                      xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"LL")==0)
  {
     circle_tangentLL(r0,              /* radius of circle */
                      xli[0],yli[0],   /* line     */
                      xli[1],yli[1],
                      xli[2],yli[2],   /* line     */
                      xli[3],yli[3],
                      xo,yo,ro,&ns);   /* number of solutions */

  } 
  else if (strcmp(task,"LP")==0)
  {
     circle_tangentLP(r0,              /* radius of circle */
                      xli[0],yli[0],   /* line     */
                      xli[1],yli[1],
                      xpi[0],ypi[0],   /* point    */
                      xo,yo,ro,&ns);   /* number of solutions */
  } 
  else if (strcmp(task,"PP")==0)
  {
     circle_tangentPP(r0,              /* radius of circle */
                      xpi[0],ypi[0],   /* point  */
                      xpi[1],ypi[1],   /* point  */
                      xo,yo,ro,&ns);   /* number of solutions */
  } 
  (*nsol)=ns;
}

/* Functions by Miloslav Cempirek added 2001-01-26 */

/*------------------------------------------
   Check if point lies out of  ellipse
--------------------------------------------*/

static gboolean out_of_ellipse(OBJECT_ELLIPSE  *e,     
			       gdouble         x,     
			       gdouble         y)     
{
  gdouble  dist;
  gdouble  xa,ya;
  gdouble  an,t;
  gdouble  dx1,dy1;
  gdouble  d1;

  dist=point_point_dist2D(e->centrum.x,e->centrum.y, x,y); 
 
  xa=e->centrum.x+e->major*cos(e->angle);  
  ya=e->centrum.y+e->major*sin(e->angle); 
 
  /* angle of cursor from major  */
  an=get_angle(e->centrum.x,e->centrum.y,x,y)
     -get_angle(e->centrum.x,e->centrum.y,xa,ya);
 
  t=ellipse_alfa2t(an,e->major,e->minor);   /* convert an -> t */

  dx1 = e->major*cos(t*M_PI/180.0);
  dy1 = e->minor*sin(t*M_PI/180.0);
  d1  = sqrt(dx1*dx1+dy1*dy1);       /* point on ellipse */					 

  return(dist>d1);
}

/*------------------------------------------
  Line from  point p and tangent to circle 
  Return TRUE if line was found.
-------------------------------------------*/

gboolean math_tangent_CP_2D(OBJECT_CIRCLE  *c,  
			    POINT   *p,
			    POINT   *u1,
			    POINT   *u2)
{
  gdouble  A,B;
  gdouble  len;
  gdouble  cos_alfa,sin_alfa;
  gdouble  x0t,y0t,r_2;
  gdouble  x1t,x2t,y1t,y2t;  
   
  /* Check so that p is outside circle */
  if( dist(c->centrum.x, c->centrum.y, p->x, p->y)<c->radius)
    return FALSE;

  r_2=c->radius*c->radius; 

  len=point_point_dist2D(c->centrum.x,c->centrum.y,p->x,p->y);
  cos_alfa=(p->x-c->centrum.x)/len;
  sin_alfa=(p->y-c->centrum.y)/len;

  /* coordinate transformation to circle centrum, 
     x-axis goes thorough point p  => y0t=0 m=0 n=0 */
     
  x0t =  (p->x-c->centrum.x)*cos_alfa+(p->y-c->centrum.y)*sin_alfa;  
  y0t =  0;
  
  A=r_2/x0t;
  B=r_2-A*A;

  y1t=+sqrt(B);
  y2t=-sqrt(B);  

  x1t = A;
  x2t = A;

  /* transformation back */

  u1->x=x1t*cos_alfa-y1t*sin_alfa + c->centrum.x;
  u1->y=x1t*sin_alfa+y1t*cos_alfa + c->centrum.y;

  u2->x=x2t*cos_alfa-y2t*sin_alfa + c->centrum.x;
  u2->y=x2t*sin_alfa+y2t*cos_alfa + c->centrum.y;

  return TRUE;
}


/*----------------------------------------
 Line from point p and tangent to ellipse
------------------------------------------*/
  
gboolean math_tangent_EP_2D(OBJECT_ELLIPSE  *e,   
			    POINT    *p,
			    POINT    *u1,
			    POINT    *u2)
{ 
  gdouble  A,B,C,D,E,G,H,I;
  gdouble  x0t,y0t;
  gdouble  a_2,b_2;   
  gdouble  x1t,x2t,y1t,y2t;
  gdouble  cos_angle, sin_angle;
  gint     nsol;

  /* Check so point is outside ellipse */
  if(!out_of_ellipse(e, p->x, p->y) )
    return FALSE;

  x0t=p->x; y0t=p->y;

  cos_angle=cos(e->angle);    /* coordinate transformation to ellipse centrum */
  sin_angle=sin(e->angle);    /* x-axist in major direction => m=0,n=0 */

  x0t =  (p->x-e->centrum.x)*cos_angle+(p->y-e->centrum.y)*sin_angle;
  y0t = -(p->x-e->centrum.x)*sin_angle+(p->y-e->centrum.y)*cos_angle;
 
  a_2 = e->major*e->major; 
  b_2 = e->minor*e->minor;
 
  A = b_2*x0t;
  B = a_2*y0t;
  C = -a_2*b_2;
  D = -B/A;
  E = -C/A;
  G = b_2*D*D + a_2;
  H = 2*b_2*D*E;
  I = b_2*E*E  - a_2*b_2;

  quadratic_equation(G,H,I,&y1t,&y2t,&nsol); 
  /* point should be out of ellipse  => two solutions */

  x1t=D*y1t+E;
  x2t=D*y2t+E;

  /* transformation back  */

  u1->x=x1t*cos_angle-y1t*sin_angle + e->centrum.x;
  u1->y=x1t*sin_angle+y1t*cos_angle + e->centrum.y;

  u2->x=x2t*cos_angle-y2t*sin_angle + e->centrum.x;
  u2->y=x2t*sin_angle+y2t*cos_angle + e->centrum.y;
  
  return TRUE;
}

