/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>
#include "struct.h"
#include "global.h"
#include "calc.h"
#include "draw.h"
#include "dia_comm.h"
#include "dia_message.h"
#include "dia_coords.h"
#include "obj_ellipse.h"
#include "xor.h"
#include "misc.h"
#include "parse.h"
#include "select.h"
#include "translation.h"


/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static gdouble major,minor, grot;
static gdouble sangle;

/* --  P R O T O T Y P E S  -- */
void obj_draw_elarc(DOCUMENT *doc, OBJECT *elarc);

/* --  F U N C T I O N S  -- */
void obj_set_prompt_elarc(void)
{
  switch(gl_stage)
    {
    case 0:
      set_prompt_text( _("Center/<Axis endpoint 1>: "));
      break;
    case 1:
      set_prompt_text( _("Axis endpoint 2: "));
      break;
    case 2:
    case 5:
      set_prompt_text( _("Rotation/<Other axis distance>: "));
      break;
    case 3:
      set_prompt_text( _("Center of ellipse: "));
      break;
    case 4:
      set_prompt_text( _("Axis endpoint: "));
      break;
    case 6:
      set_prompt_text( _("Rotation around major axis: "));
      break;
    case 7:
      set_prompt_text( _("Start angle: "));
      break;
    case 8:
      set_prompt_text( _("Include angle/<End angle>: "));
      break;
    case 9:
      set_prompt_text( _("Include angle: "));
      break;
    }
}

static void calc_bound_elarc(OBJECT *obj, gboolean first, gdouble start_angle, gdouble end_angle)
{
  gint i;
  gint num;
  gdouble sa,ea,an, mi, ma;
  gdouble sina,cosa,dt;
  POINT *list;

  ma=obj->data.elarc.major;
  mi=obj->data.elarc.minor;

  num=MAX(65, ma*mi/1000);
  num=MIN(num, 700);

  sa=start_angle*M_PI/180.0;
  ea=end_angle*M_PI/180.0;

  if (ABS(sa-ea)<GFLOAT_SENSITIVE) return; /* Nothing to draw */

  list=g_new(POINT,num+5);

  dt=ABS(sa-ea)/(gdouble)num;

  for(i=0, an=sa;i<=num;i++, an+=dt)
  {
    list[i].x=((gdouble)ma)*cos(2*M_PI-an);
    list[i].y=((gdouble)mi)*sin(2*M_PI-an);
  }

  sina=sin(-obj->data.elarc.angle*M_PI/180);
  cosa=cos(-obj->data.elarc.angle*M_PI/180);
  
  if(first)
    {
      obj->bounding_box[0]=obj->data.elarc.centrum.x+(list[i].x*cosa-list[i].y*sina);
      obj->bounding_box[1]=obj->data.elarc.centrum.y+(list[i].x*sina+list[i].y*cosa);
    }

  for(i=0;i<=num;i++)
    {
      obj->bounding_box[0]=MIN(obj->data.elarc.centrum.x+(list[i].x*cosa-list[i].y*sina), obj->bounding_box[0]);
      obj->bounding_box[1]=MIN(obj->data.elarc.centrum.y+(list[i].x*sina+list[i].y*cosa), obj->bounding_box[1]);
      obj->bounding_box[2]=MAX(obj->data.elarc.centrum.x+(list[i].x*cosa-list[i].y*sina), obj->bounding_box[2]);
      obj->bounding_box[3]=MAX(obj->data.elarc.centrum.y+(list[i].x*sina+list[i].y*cosa), obj->bounding_box[3]);
    }

  g_free(list);
}

void obj_bound_elarc(OBJECT *obj)
{
  if(obj->data.elarc.start_angle>obj->data.elarc.end_angle)
    {
      calc_bound_elarc(obj, TRUE, obj->data.elarc.start_angle, 360);
      calc_bound_elarc(obj, FALSE, 0, obj->data.elarc.end_angle);
    }
  else
    calc_bound_elarc(obj, TRUE, obj->data.elarc.start_angle, obj->data.elarc.end_angle);    
}


OBJECT *obj_add_elarc(DOCUMENT *doc, LAYER *layer, gchar *name, gint type_colour, COLOUR *colour, 
		      gint type_linestyle, LINESTYLE *linestyle, gint type_linewidth, 
		      gdouble linewidth,
		      gdouble xcen, gdouble ycen, gdouble major, gdouble minor, 
		      gdouble angle, gdouble start_angle, gdouble end_angle)
{
  OBJECT *obj;
  
  /* Make sure we have valid parameters */
  if(doc==NULL || layer==NULL)
    {
      g_warning( _("Unvalid parameters for function %s in file %s"), __FUNCTION__, __FILE__);
      return NULL;
    }

  /* Create object */
  obj=g_new(OBJECT, 1);
  
  /* Type of object */
  obj->type=TYPE_ELARC;
  
  /* Pointer to layer */
  obj->layer=layer;

  /* Set up attributes for this object */
  obj_set_attributes(doc, obj, name, type_colour, colour, type_linestyle, 
		     linestyle, type_linewidth, linewidth); 
  
  /* Set object specific attributes */
  obj->data.elarc.centrum.x=xcen;
  obj->data.elarc.centrum.y=ycen;
  obj->data.elarc.major=major;
  obj->data.elarc.minor=minor;
  obj->data.elarc.angle=angle;
  obj->data.elarc.start_angle=start_angle;
  obj->data.elarc.end_angle=end_angle;

  /* Add object to layer list */
  layer->object_list=g_list_append(layer->object_list,obj);

  /* Make bounding box for object */
  obj_bound_elarc(obj);

  doc->saved=FALSE;
  make_window_title(doc);

  /* Return new object */
  return obj;
}

void obj_set_tool_elarc(GtkWidget *widget, gpointer data)
{
  remove_old();

  if(!command_from_keyboard)
    print( _("elarc"));

  if(work_document!=NULL)
    deselect_all_objects(work_document);

  if(misc_object_get_document()==NULL)
    return;
  
  gl_tool=TOOL_ELARC;
  gl_stage=0;
  work_document=NULL;
  set_prompt();
  set_cursor(NULL);
  UpdateCoordsLabel();
}

gdouble ellipse_alfa2t(gdouble alfa,gdouble major,gdouble minor)
{
  gdouble t0;
  
  if (fabs(alfa-90)<GFLOAT_SENSITIVE || fabs(alfa-270)<GFLOAT_SENSITIVE)
    return alfa;
  
  t0=fabs(atan(tan(alfa*M_PI/180)*major/minor)*180/M_PI);
  
  if (alfa>270)
    return 360 - t0;
  else if (alfa>180)
    return 180 + t0; 
  else if (alfa>90)
    return 180 - t0;
  else 
    return t0;
}

void obj_button_elarc(DOCUMENT *doc, gint button, gdouble x, gdouble y)
{
  gdouble nx,ny,gr1,gr2;
  gdouble eangle;

  /* Right mousebutton => cancel/break */
  if(button==3)
    {
      remove_old();
      print( _("*Cancel*"));
      tool_idle(doc);
      return;
    }

  switch(gl_stage)
    {
    case 0:   /* Axis endpoint 1*/
      if(doc->gl_snap)
	make_snap(doc,&x,&y);
      gl_oldmx=x;
      gl_oldmy=y;
      gl_stage++;
      set_prompt();
      break;
    case 1:   /* Axis endpoint 2 */
      remove_old();
      make_ortho_snap(doc,gl_oldmx, gl_oldmy,  &x, &y);
      xc=x;
      yc=y;
      gl_stage++;
      set_prompt();
      break;
    case 3:   /* Center point of elarc */
      if(doc->gl_snap)
	make_snap(doc,&x,&y);
      xc=gl_oldmx=x;
      yc=gl_oldmy=y;
      gl_stage++;
      set_prompt();
      break;
    case 4:
      remove_old();
      make_ortho_snap(doc,gl_oldmx, gl_oldmy,  &x, &y);
      gl_oldmx=x;
      gl_oldmy=y;
      gl_stage++;
      set_prompt();
      break;
    case 2:   /* Axis distance */
    case 5:
      remove_old();
      if(doc->gl_snap)
	make_snap(doc,&x,&y);

      calc_ellipse(x,y,&grot,&nx,&ny);

      if( dist(xc,yc,nx,ny)<dist(xc,yc,gl_oldmx,gl_oldmy) )
	{
	  minor=dist(xc,yc,nx,ny);
	  major=dist(xc,yc,gl_oldmx,gl_oldmy);
	  grot=get_angle(xc,yc, gl_oldmx, gl_oldmy);
	}
      else
	{
	  major=dist(xc,yc,nx,ny);
	  minor=dist(xc,yc,gl_oldmx,gl_oldmy);
	  grot=get_angle(xc,yc, nx, ny);
	}

      gl_stage=7;
      UpdateCoordsLabel();
      set_prompt();
      break;
    case 6:  /* Rotation */
      remove_old();
      if(doc->gl_snap)
	make_snap(doc,&x,&y);

      calc_ellipse(x,y,&grot,&nx,&ny);

      gr1=dist(xc,yc,gl_oldmx,gl_oldmy);
      gr2=dist(xc,yc,nx,ny);

      if(gr2>gr1)
	gr2=gr1;
	
      if( gr2<gr1 )
	{
	  minor=gr2;
	  major=gr1;
	  grot=get_angle(xc,yc, gl_oldmx, gl_oldmy);
	}
      else
	{
	  major=gr2;
	  minor=gr1;
	  grot=get_angle(xc,yc, nx, ny);
	}

      gl_stage=7;
      UpdateCoordsLabel();
      set_prompt();
      break;
    case 7:
      sangle=get_angle(xc,yc,x,y)-grot;

      while(sangle<0) sangle+=360;
      while(sangle>360) sangle-=360;

      sangle=ellipse_alfa2t(sangle,major,minor);

      gl_stage=8;
      set_prompt();
      break;
    case 8:
      remove_old();

      eangle=get_angle(xc,yc,x,y)-grot;

      while(eangle<0) eangle+=360;
      while(eangle>360) eangle-=360;

      eangle=ellipse_alfa2t(eangle,major,minor);

      obj_draw_elarc(doc, obj_add_elarc(doc, doc->current_layer, NULL, 
					doc->type_colour, doc->current_colour, 
					doc->type_linestyle, doc->current_linestyle,
					doc->type_linewidth, doc->current_linewidth,
					xc,yc, major,minor,grot,sangle,eangle));
      tool_idle(doc);
      break;
    }
}

void obj_prompt_elarc(DOCUMENT *doc, gchar *buffer)
{
  gdouble tx, ty, value;
  gint mn;

  switch(gl_stage)
    {
    case 0:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2|PROMPT_MENU,buffer,
			  FALSE, _("C/"),0,0,&tx,&ty,NULL,&mn)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_elarc(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	gl_stage=3;	    
	set_prompt();
	break;
      }
      break;
    case 1:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  TRUE,NULL,gl_oldmx,gl_oldmy,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_elarc(doc,1,tx,ty);
	break;
      }
      break;
    case 3:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  FALSE,NULL,0,0,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_elarc(doc,1,tx,ty);
	break;
      }
      break;
    case 2:
    case 5:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1|PROMPT_MENU,buffer,
			  FALSE, _("R/"),0,0,NULL,NULL,&value,&mn)){
      case PROMPT_POINT1:
	remove_old();
	
	if( value<dist(xc,yc,gl_oldmx,gl_oldmy) )
	{
	  minor=value;
	  major=dist(xc,yc,gl_oldmx,gl_oldmy);
	  grot=get_angle(xc,yc, gl_oldmx, gl_oldmy);
	}
	else
	  {
	    major=value;
	    minor=dist(xc,yc,gl_oldmx,gl_oldmy);
	    grot=get_angle(xc,yc, gl_oldmx, gl_oldmy)+90;
	  }

	gl_stage=7;
	set_prompt();
	break;
      case PROMPT_MENU:
	gl_stage=6;
	set_prompt();
	break;
      }
      break;
    case 4:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  TRUE,NULL,gl_oldmx,gl_oldmy,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_elarc(doc,1,tx,ty);
	break;
      }
      break;
    case 6:  /* Rotation */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1,buffer,
			  FALSE, NULL,0,0,NULL,NULL,&value,NULL)){
      case PROMPT_POINT1:
	remove_old();

	minor=dist(xc,yc,gl_oldmx,gl_oldmy)*cos(value*M_PI/180.0);		   
	major=dist(xc,yc,gl_oldmx,gl_oldmy);
	
	grot=get_angle(xc,yc, gl_oldmx, gl_oldmy);
      	
	gl_stage=7;
	UpdateCoordsLabel();
	set_prompt();
	break;
      }
      break;
    case 7:  /* Start angle */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1,buffer,
			  FALSE, NULL,0,0,NULL,NULL,&value,NULL)){
      case PROMPT_POINT1:
	remove_old();

	while(value<0) value+=360;
	while(value>360) value-=360;
	
	sangle=ellipse_alfa2t(value,major,minor);

	gl_stage=8;
	UpdateCoordsLabel();
	set_prompt();
	break;
      }
      break;
    case 8:  /* Include/<End angle> */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1|PROMPT_MENU,buffer,
			  FALSE, _("I/"),0,0,NULL,NULL,&value,&mn)){
      case PROMPT_MENU:
	gl_stage=9;
	set_prompt();
	break;
      case PROMPT_POINT1:
	remove_old();
	
	while(value<0) value+=360;
	while(value>360) value-=360;
	
	obj_draw_elarc(doc, obj_add_elarc(doc, doc->current_layer, NULL, 
					  doc->type_colour, doc->current_colour, 
					  doc->type_linestyle, doc->current_linestyle,
					  doc->type_linewidth, doc->current_linewidth,
					  xc,yc, major,minor,grot,sangle,ellipse_alfa2t(value,major,minor)));
	tool_idle(doc);
	break;
      }
      break;
    }
}

void obj_xor_elarc_start_angle(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble a;

  a=get_angle(xc,yc,x,y)-grot;
  
  while(a<0) a+=360;
  while(a>360) a-=360;
  
  UpdateSingelCoord(2,a);
  UpdateSingelCoord(-3,0);

  misc_setup_black_line(doc);
  
  gdk_draw_line(doc->widgets.draw->window, 
		doc->xor_gc,
		WLDtoSCR_x(doc,x),WLDtoSCR_y(doc,y),
		WLDtoSCR_x(doc,xc),WLDtoSCR_y(doc,yc));
  
  draw_ellipse(doc,doc->widgets.draw->window, 
	       misc_setup_gc(doc),
	       WLDtoSCR_x(doc,xc), WLDtoSCR_y(doc,yc), 
	       WLDtoSCR(doc,major),WLDtoSCR(doc, minor),grot);
}

void obj_xor_elarc_end_angle(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble eangle;

  misc_setup_black_line(doc);
  
  gdk_draw_line(doc->widgets.draw->window, 
		doc->xor_gc,
		WLDtoSCR_x(doc,x),WLDtoSCR_y(doc,y),
		WLDtoSCR_x(doc,xc),WLDtoSCR_y(doc,yc));

  eangle=get_angle(xc,yc,x,y)-grot;

  while(eangle<0) eangle+=360;
  while(eangle>360) eangle-=360;

  eangle=ellipse_alfa2t(eangle,major,minor);

  draw_elarc(doc,doc->widgets.draw->window, 
	     misc_setup_gc(doc),
	     WLDtoSCR_x(doc,xc), WLDtoSCR_y(doc,yc), 
	     WLDtoSCR(doc,major),WLDtoSCR(doc, minor),
	     grot, sangle,eangle);

  UpdateSingelCoord(2,sangle);
  if(gl_stage==8)
    UpdateSingelCoord(3,eangle);
  else 
    {
      /* This does not give the correct angles */
      if(eangle>sangle)
	UpdateSingelCoord(3,eangle-sangle);
      else
	UpdateSingelCoord(3,eangle+sangle);
    }
}

void obj_transform_elarc(OBJECT *elarc, TRANSFORM *trans)
{
 gdouble cx,cy;
  gdouble tx,ty;
  gdouble major,minor;

  misc_transform_point(trans,elarc->data.elarc.centrum.x,elarc->data.elarc.centrum.y,&cx,&cy);
  elarc->data.elarc.centrum.x=cx;
  elarc->data.elarc.centrum.y=cy;
  
  major=elarc->data.elarc.major;
  minor=elarc->data.elarc.minor;

  if(trans!=NULL && (trans->type & OBJ_SCALE))
    {
      elarc->data.elarc.major=major=ABS(major*trans->scale_x);
      elarc->data.elarc.minor=minor=ABS(minor*trans->scale_y);
    }

  if(trans!=NULL && (trans->type & OBJ_ROTATE))
    {
      tx=elarc->data.elarc.centrum.x+major*cos(elarc->data.elarc.start_angle*M_PI/180);
      ty=elarc->data.elarc.centrum.y-minor*sin(elarc->data.elarc.start_angle*M_PI/180);
      misc_rotate_point(trans->rot_angle,trans->cen_x,trans->cen_y, &tx,&ty);
      elarc->data.elarc.start_angle=get_angle(cx,cy,tx,ty);

      tx=elarc->data.elarc.centrum.x+major*cos(elarc->data.elarc.end_angle*M_PI/180);
      ty=elarc->data.elarc.centrum.y-minor*sin(elarc->data.elarc.end_angle*M_PI/180);
      misc_rotate_point(trans->rot_angle,trans->cen_x,trans->cen_y, &tx,&ty);
      elarc->data.elarc.end_angle=get_angle(cx,cy,tx,ty);

      elarc->data.elarc.angle+=trans->rot_angle;
    }


  obj_bound_elarc(elarc);
}

void obj_draw_elarc_block(DOCUMENT *doc, GdkGC *gc, OBJECT *elarc, TRANSFORM *trans)
{
  gdouble sa,ea,rot;
  gdouble Ax,Ay,Bx,By,major,minor;
 
  misc_transform_point(trans,elarc->data.elarc.centrum.x,elarc->data.elarc.centrum.y,&Ax,&Ay);

  major=elarc->data.elarc.major;
  minor=elarc->data.elarc.minor;
  rot=elarc->data.elarc.angle;
  sa=elarc->data.elarc.start_angle;
  ea=elarc->data.elarc.end_angle;

  if(trans!=NULL && (trans->type & OBJ_SCALE))
    {
      major=ABS(major*trans->scale_x);
      minor=ABS(minor*trans->scale_y);
    }

  if(trans!=NULL && (trans->type & OBJ_ROTATE))
    {
      Bx=elarc->data.elarc.centrum.x+major*cos(elarc->data.elarc.start_angle*M_PI/180);
      By=elarc->data.elarc.centrum.y-minor*sin(elarc->data.elarc.start_angle*M_PI/180);
      misc_rotate_point(trans->rot_angle,trans->cen_x,trans->cen_y, &Bx,&By);
      sa=get_angle(Ax,Ay,Bx,By);

      Bx=elarc->data.elarc.centrum.x+major*cos(elarc->data.elarc.end_angle*M_PI/180);
      By=elarc->data.elarc.centrum.y-minor*sin(elarc->data.elarc.end_angle*M_PI/180);
      misc_rotate_point(trans->rot_angle,trans->cen_x,trans->cen_y, &Bx,&By);
      ea=get_angle(Ax,Ay,Bx,By);

      rot+=trans->rot_angle;
    }

  /* Draw bounding box */
  /*  gdk_draw_line(doc->widgets.draw->window, gc, 
		WLDtoSCR_x(doc, elarc->bounding_box[0]),WLDtoSCR_y(doc, elarc->bounding_box[1]),
		WLDtoSCR_x(doc, elarc->bounding_box[2]),WLDtoSCR_y(doc, elarc->bounding_box[1]));
  gdk_draw_line(doc->widgets.draw->window, gc, 
		WLDtoSCR_x(doc, elarc->bounding_box[2]),WLDtoSCR_y(doc, elarc->bounding_box[1]),
		WLDtoSCR_x(doc, elarc->bounding_box[2]),WLDtoSCR_y(doc, elarc->bounding_box[3]));
  gdk_draw_line(doc->widgets.draw->window, gc, 
		WLDtoSCR_x(doc, elarc->bounding_box[2]),WLDtoSCR_y(doc, elarc->bounding_box[3]),
		WLDtoSCR_x(doc, elarc->bounding_box[0]),WLDtoSCR_y(doc, elarc->bounding_box[3]));
  gdk_draw_line(doc->widgets.draw->window, gc, 
		WLDtoSCR_x(doc, elarc->bounding_box[0]),WLDtoSCR_y(doc, elarc->bounding_box[1]),
		WLDtoSCR_x(doc, elarc->bounding_box[0]),WLDtoSCR_y(doc, elarc->bounding_box[3]));
  */

  draw_elarc(doc,doc->widgets.draw->window, gc,
	     WLDtoSCR_x(doc,Ax), WLDtoSCR_y(doc,Ay), 
	     WLDtoSCR(doc,major),WLDtoSCR(doc,minor),
	     rot, sa, ea);	 
}

void obj_draw_elarc(DOCUMENT *doc, OBJECT *elarc)
{
  obj_draw_elarc_block(doc,misc_setup_gc_object(doc, elarc),elarc,NULL);
}


gint obj_select_elarc(OBJECT *elarc, gdouble x, gdouble y, gdouble d)
{
  gdouble c;
  gdouble rx,ry;
  gdouble tx,ty;
  gdouble tangle;

  x-=elarc->data.elarc.centrum.x;
  y-=elarc->data.elarc.centrum.y;

  rx=elarc->data.elarc.minor;
  ry=elarc->data.elarc.major;

  c=elarc->data.elarc.angle*M_PI/180;

  tx=x*cos(-c)+y*sin(-c);
  ty=-x*sin(-c)+y*cos(-c);

  if( !((tx*tx)/((ry+d)*(ry+d))+(ty*ty)/((rx+d)*(rx+d))<1   && 
      (tx*tx)/((ry-d)*(ry-d))+(ty*ty)/((rx-d)*(rx-d))>1)  )
    return FALSE;

  tangle=get_angle(0,0,tx,ty);
  
  while(tangle<0) tangle+=360;
  while(tangle>360) tangle-=360;
  
  tangle=ellipse_alfa2t(tangle, elarc->data.elarc.major, elarc->data.elarc.minor);

  if(elarc->data.elarc.end_angle>elarc->data.elarc.start_angle)
    return( (tangle>elarc->data.elarc.start_angle) && (tangle<elarc->data.elarc.end_angle));
  else
    return( (tangle>elarc->data.elarc.start_angle) || (tangle<elarc->data.elarc.end_angle));
 
  return FALSE;
}








