/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>
#include "struct.h"
#include "global.h"
#include "calc.h"
#include "layer.h"
#include "linestyle.h"
#include "dia_comm.h"
#include "dia_message.h"
#include "dia_coords.h"
#include "xor.h"
#include "obj_line.h"
#include "obj_arc.h"
#include "obj_circle.h"
#include "obj_text.h"
#include "obj_attribute.h"
#include "misc.h"
#include "parse.h"
#include "select.h"
#include "tool_block.h"
#include "translation.h"

/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static BLOCK *curr_block;

static gdouble ix,iy;
static gdouble scale_x=1, scale_y=1;
static gdouble rot=0;

static gchar *attr_prompt, *attr_value;
static GList *attr_list;
static GList *attr_value_list;

static gchar *previous_block_name=NULL;

/* --  P R O T O T Y P E S  -- */
void obj_draw_insert(DOCUMENT *doc, OBJECT *insert);


/* --  F U N C T I O N S  -- */
void obj_set_prompt_insert(void)
{
  switch(gl_stage)
    {
    case 0:
      if(previous_block_name==NULL)
	set_prompt_text( _("Select block to insert (or ?): "));
      else
	set_prompt_text( _("Select block to insert (or ?) <%s>: "), previous_block_name);
      break;
    case 1:
      set_prompt_text( _("Insertion base point: "));
      break;
    case 2:
      set_prompt_text( _("Corner/<X scale factor=%f>: "), scale_x);
      break;
    case 3:
      set_prompt_text( _("Y scale factor <%f>: "), scale_x);
      break;
    case 4:
      set_prompt_text( _("Rotation angle <%f>: "), rot);    
      break;
    case 5:
      set_prompt_text( _("%s <%s>: "), attr_prompt, attr_value);    
      break;
    }
}

static gboolean get_first_attribute()
{
  BLOCK_OBJECT *block_object;

  /* Make a new list of attributes value */
  /* This is a memory leak if it isnt release correctly */
  attr_value_list=NULL;

  /* Get first object in block */
  attr_list=g_list_first(curr_block->object_list);
  
  while(attr_list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)attr_list->data;

      if(block_object->type==TYPE_ATTRIBUTE)
	{
	  attr_prompt=block_object->data.attribute.prompt;
	  attr_value=block_object->data.attribute.text.text;
	  return TRUE;
	}

      attr_list=g_list_next(attr_list);
    }

  return FALSE;
}

static gboolean get_next_attribute()
{
  BLOCK_OBJECT *block_object;

  attr_list=g_list_next(attr_list);

  while(attr_list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)attr_list->data;

      if(block_object->type==TYPE_ATTRIBUTE)
	{
	  attr_prompt=block_object->data.attribute.prompt;
	  attr_value=block_object->data.attribute.text.text;
	  return TRUE;
	}

      attr_list=g_list_next(attr_list);
    }

  return FALSE;
}

void obj_bound_insert(OBJECT *insert)
{
  gboolean first=TRUE;
  DOCUMENT *doc;
  TRANSFORM insert_transform;
  LAYER *layer;
  OBJECT object;
  BLOCK_OBJECT *block_object;
  GList *list;
  GList *attr_list;

  /* Get document for object */
  doc=insert->layer->doc;

  /* Make tranform matrix */
  insert_transform.type=OBJ_TRANSLATE;
  insert_transform.trans_x=insert->data.block.insert_point.x;
  insert_transform.trans_y=insert->data.block.insert_point.y;
  insert_transform.cen_x=0;
  insert_transform.cen_y=0;
  
  /* Check if need to scale block */
  if(ABS(insert->data.block.scale_x-1)>GFLOAT_SENSITIVE ||
     ABS(insert->data.block.scale_y-1)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_SCALE;
      insert_transform.scale_x=insert->data.block.scale_x;
      insert_transform.scale_y=insert->data.block.scale_y;
    }

  /* Check if need to rotate block */
  if(ABS(insert->data.block.rotation)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_ROTATE;
      insert_transform.rot_angle=-insert->data.block.rotation*M_PI/180.0;
    }

  /* Get list of attributes in this block */
  attr_list=g_list_first(insert->data.block.attr);

  /* Get first object in block */
  list=g_list_first(insert->data.block.block->object_list);

  while(list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)list->data;

      /* Get layer for this object */

      /* Check if we should use the layer of the block */
      if(!strcmp(block_object->name_of_layer,"0"))
	layer=insert->layer;
      else
	{
	  layer=layer_find(doc, block_object->name_of_layer);
	  if(layer==NULL)
	    {
	      g_warning( _("Layer %s not found. Creating it."), block_object->name_of_layer );
	      layer=layer_new(doc, -1, doc->current_linestyle, 0.0 , 0.0, 0.0, 0.0,  
			  block_object->name_of_layer, TRUE, TRUE);
	    }
	}

      /* Draw only if layer is visible */      
      if(layer->visible)
	{
	  /* Make nice object to draw */
	  object.type=block_object->type;
	  object.layer=layer;
	  
	  /* Make linewidth */
	  object.type_linewidth=block_object->type_linewidth;
	  switch(object.type_linewidth)
	    {
	    case BYLAYER:
	      object.linewidth=layer->linewidth;
	      break;
	    case BYBLOCK:
	      object.linewidth=insert->linewidth;
	      break;
	    case USERDEFINED:
	      object.linewidth=block_object->linewidth;
	      break;
	    }
	  
	  memcpy(&object.data, &block_object->data, sizeof(OBJECT_UNION));
	  
	  make_object_selected(&object, OBJECT_SEL, FALSE);
	  object.selected=insert->selected;
	  
	  switch(object.type)
	    {
	    case TYPE_LINE:
	      obj_transform_line( &object, &insert_transform);
	      obj_bound_line(&object);
	      break;
	    case TYPE_CIRCLE:
	      obj_transform_circle( &object, &insert_transform);
	      obj_bound_circle(&object);
	      break;
	    case TYPE_ARC:
	      obj_transform_arc( &object, &insert_transform);
	      obj_bound_arc(&object);
	      break;
	    case TYPE_TEXT:
	      obj_transform_text( &object, &insert_transform);
	      obj_bound_text(&object);
	      break;
	    case TYPE_ATTRIBUTE:
	      /* Fix text of attribute */
	      if(attr_list!=NULL)
		{
		  object.data.attribute.text.text=(gchar *)attr_list->data;
		  attr_list=g_list_next(attr_list);
		  obj_transform_attribute( &object, &insert_transform);
		  obj_bound_attribute(&object);
		}
	      break;
	    }

	  if(first)
	    {
	      insert->bounding_box[0]=MIN(object.bounding_box[0], object.bounding_box[2]);
	      insert->bounding_box[1]=MIN(object.bounding_box[1], object.bounding_box[3]);
	      insert->bounding_box[2]=MAX(object.bounding_box[0], object.bounding_box[2]);
	      insert->bounding_box[3]=MAX(object.bounding_box[1], object.bounding_box[3]);
	      first=FALSE; 
	    }
	  else
	    {
	      insert->bounding_box[0]=MIN(insert->bounding_box[0], object.bounding_box[0]);
	      insert->bounding_box[0]=MIN(insert->bounding_box[0], object.bounding_box[2]);
	      insert->bounding_box[1]=MIN(insert->bounding_box[1], object.bounding_box[1]);
	      insert->bounding_box[1]=MIN(insert->bounding_box[1], object.bounding_box[3]);
	      insert->bounding_box[2]=MAX(insert->bounding_box[2], object.bounding_box[0]);
	      insert->bounding_box[2]=MAX(insert->bounding_box[2], object.bounding_box[2]);
	      insert->bounding_box[3]=MAX(insert->bounding_box[3], object.bounding_box[1]);
	      insert->bounding_box[3]=MAX(insert->bounding_box[3], object.bounding_box[3]);
	    }

	}

      list=g_list_next(list);
    }
  
}

OBJECT *obj_add_insert(DOCUMENT *doc, LAYER *layer, gchar *name, gint type_colour, COLOUR *colour, 
		       gint type_linestyle, LINESTYLE *linestyle, gint type_linewidth, gdouble linewidth, 
		       gdouble ins_x, gdouble ins_y, gdouble scale_x, gdouble scale_y, gdouble rot, GList *attr,
		       BLOCK *block)
{
  OBJECT *obj;
  
  /* Make sure we have valid parameters */
  if(doc==NULL || layer==NULL)
    {
      g_warning( _("Unvalid parameters for function %s in file %s"), __FUNCTION__, __FILE__);
      return NULL;
    }
  
  /* Create object */
  obj=g_new(OBJECT, 1);
  
  /* Type of object */
  obj->type=TYPE_BLOCK;
  
  /* Pointer to layer */
  obj->layer=layer;

  /* Set up attributes for this object */
  obj_set_attributes(doc, obj, name, type_colour, colour, type_linestyle, 
		     linestyle, type_linewidth, linewidth); 
  
  /* Set object specific attributes */  
  obj->data.block.insert_point.x=ins_x;
  obj->data.block.insert_point.y=ins_y;
  obj->data.block.scale_x=scale_x;
  obj->data.block.scale_y=scale_y;
  obj->data.block.rotation=rot;
  obj->data.block.block=block;
  obj->data.block.attr=attr;

  /* Add object to layer list */
  layer->object_list=g_list_append(layer->object_list,obj);

  /* Make bounding box for object */
  obj_bound_insert(obj);

  doc->saved=FALSE;
  make_window_title(doc);

  /* Return new object */
  return obj;
}

void obj_set_tool_insert(GtkWidget *widget, gpointer data)
{
  remove_old();

  if(!command_from_keyboard)
    print("insert");

  if(work_document!=NULL)
    deselect_all_objects(work_document);

  if(misc_object_get_document()==NULL)
    return;
  
  gl_tool=TOOL_INSERT;
  gl_stage=0;
  work_document=NULL;
  set_prompt();
  set_cursor(NULL);
  UpdateCoordsLabel();
}

void obj_button_insert(DOCUMENT *doc, gint button, gdouble x, gdouble y)
{
  /* Right mousebutton => cancel/break */
  if(button==3)
    {
      remove_old();
      print("*Cancel*");
      tool_idle(doc);
      return;
    }

  switch(gl_stage)
    {
    case 1:  /* Insert */
      if(doc->gl_snap)
	make_snap(doc, &x, &y);

      remove_old();

      gl_oldmx=ix=x;
      gl_oldmy=iy=y;

      gl_stage++;
      set_prompt();     
      break;
    case 2:  /* Scale X */
      make_ortho_snap(doc, gl_oldmx, gl_oldmy, &x, &y);
      remove_old();
      scale_y=scale_x=dist(ix, iy, x, y);
      gl_stage++;
      set_prompt();
      break;
    case 3:  /* Scale Y */
      make_ortho_snap(doc, gl_oldmx, gl_oldmy, &x, &y);
      remove_old();
      scale_y=dist(ix, iy, x, y);
      gl_stage++;
      gl_oldmx=ix;
      gl_oldmy=iy;
      set_prompt();
      break;
    case 4:  /* Rotation */
      remove_old();

      make_ortho_snap(doc, gl_oldmx, gl_oldmy, &x, &y);

      rot=get_angle(ix, iy, x, y);

      if(!get_first_attribute())
	{
	  if(previous_block_name!=NULL)
	    g_free(previous_block_name);
	  previous_block_name=g_strdup(curr_block->name);

	  obj_draw_insert(doc, obj_add_insert(doc,doc->current_layer, NULL, doc->type_colour, doc->current_colour,
					      doc->type_linestyle, doc->current_linestyle, 
					      doc->type_linewidth, doc->current_linewidth,
					      ix, iy, scale_x, scale_y, rot, NULL, curr_block));
	  tool_idle(doc);
	}
     else
       {
	 gl_stage++;
	 set_prompt();
       }
      break;
    }
}

void obj_prompt_insert(DOCUMENT *doc, gchar *buffer)
{
  gdouble x,y,e;

  switch(gl_stage)
    {
    case 0:  /* Enter name */
      if(!strcmp(buffer,"?"))
	{
	  tool_block_list(NULL, 0);
	  return;
	}
      else
	{
	  if(!strcmp(buffer,""))	   
	    curr_block=find_block_name(doc, previous_block_name);
	  else
	    curr_block=find_block_name(doc,buffer);
	  
	  if(curr_block!=NULL)
	    {
	      gl_stage++;
	      work_document=doc;
	    }
	  else
	    print( _("\n  This block is not defined."));
	}
      set_prompt();
      break;
    case 1:  /* Set insertion point */
      switch(parse_prompt((PROMPT_OSNAP|PROMPT_POINT2),
			  buffer, FALSE, NULL,
			  0,0,&x,&y,NULL, NULL))
	{
	case PROMPT_POINT2:
	  work_document=doc;
	  obj_button_insert(doc,1,x,y);
	  break;
	}
      break;
    case 2:  /* Set scale x */
      switch(parse_prompt(PROMPT_POINT1|PROMPT_RETURN,
			  buffer, FALSE, NULL,
			  0,0, NULL, NULL, &e, NULL))
	{
	case PROMPT_POINT1:
	  scale_x=scale_y=e;
	case PROMPT_RETURN:
	  remove_old();
	  work_document=doc;
	  gl_stage++;      
	  set_prompt();
	  break;
	}
      break;
    case 3:  /* Set scale y */
      switch(parse_prompt(PROMPT_POINT1|PROMPT_RETURN,
			  buffer, FALSE, NULL,
			  0,0, NULL, NULL, &e, NULL))
	{
	case PROMPT_POINT1:
	  scale_y=e;
	case PROMPT_RETURN:
	  remove_old();
	  work_document=doc;
	  gl_stage++;      
	  set_prompt();
	  break;
	}
      break;
    case 4:  /* Set rotation */
      switch(parse_prompt(PROMPT_POINT1|PROMPT_RETURN,
			  buffer, FALSE, NULL,
			  0,0, NULL, NULL, &e, NULL))
	{
	case PROMPT_POINT1:
	  rot=e;
	case PROMPT_RETURN:
	  remove_old();
	  if(!get_first_attribute())
	    {
	      if(previous_block_name!=NULL)
		g_free(previous_block_name);
	      previous_block_name=g_strdup(curr_block->name);

	      obj_draw_insert(doc, obj_add_insert(doc,doc->current_layer, NULL, doc->type_colour, doc->current_colour,
						  doc->type_linestyle, doc->current_linestyle, 
						  doc->type_linewidth, doc->current_linewidth,
						  ix, iy, scale_x, scale_y, rot, NULL, curr_block));
	      tool_idle(doc);
	    }
	  else
	    {
	      gl_stage++;
	      set_prompt();
	    }
	  break;
	}
      break;
    case 5:  /* Get attributes values */
      if(!strcmp("",buffer))
	attr_value_list=g_list_append(attr_value_list, attr_value);
      else
	attr_value_list=g_list_append(attr_value_list, g_strdup(buffer));
      if(!get_next_attribute())
	{
	  if(previous_block_name!=NULL)
	    g_free(previous_block_name);
	  previous_block_name=g_strdup(curr_block->name);

	  obj_draw_insert(doc, obj_add_insert(doc,doc->current_layer, NULL, doc->type_colour, doc->current_colour,
					      doc->type_linestyle, doc->current_linestyle, 
					      doc->type_linewidth, doc->current_linewidth,
					      ix, iy, scale_x, scale_y, rot, attr_value_list, curr_block));
	  tool_idle(doc);
	}
      else
	set_prompt();
      break;
  }
}

void draw_block_xor(DOCUMENT *doc, GdkGC *gc, BLOCK *block, gdouble ins_x, gdouble ins_y, gdouble scale_x, gdouble scale_y,
		    gdouble rot)
{
  OBJECT object;
  TRANSFORM transform;
  BLOCK_OBJECT *block_object;
  GList *list;

  /* Make tranform matrix */
  transform.type=OBJ_TRANSLATE;
  transform.trans_x=ins_x;
  transform.trans_y=ins_y;
  transform.cen_x=0;
  transform.cen_y=0;

  /* Check if need to scale block */
  if(ABS(scale_x-1)>GFLOAT_SENSITIVE ||
     ABS(scale_y-1)>GFLOAT_SENSITIVE)
    {
      transform.type|=OBJ_SCALE;
      transform.scale_x=scale_x;
      transform.scale_y=scale_y;
    }

  /* Check if need to rotate block */
  if(ABS(rot)>GFLOAT_SENSITIVE)
    {
      transform.type|=OBJ_ROTATE;
      transform.rot_angle=-rot*M_PI/180.0;
    }

  /* Get first object in block */
  list=g_list_first(block->object_list);

  while(list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)list->data;

      /* Make nice object to draw */
      object.type=block_object->type;
      object.selected=OBJECT_NOT_SEL;

      memcpy(&object.data, &block_object->data, sizeof(OBJECT_UNION));

      /* Make sure all nodes are selected */
      make_object_selected(&object, OBJECT_SEL, FALSE);      
  
      switch(object.type)
	{
	case TYPE_LINE:
	  obj_draw_line_block(doc, gc, &object, &transform);
	  break;
	case TYPE_CIRCLE:
	  obj_draw_circle_block(doc, gc, &object, &transform);
	  break;
	case TYPE_ARC:
	  obj_draw_arc_block(doc, gc, &object, &transform);
	  break;
	case TYPE_TEXT:
	  obj_draw_text_block(doc, gc, &object, &transform);
	  break;
	case TYPE_ATTRIBUTE:
	  obj_draw_attribute_block(doc, gc, &object, &transform);
	  break;
	}
      list=g_list_next(list);
    }
}

void obj_xor_insert_ins(DOCUMENT *doc,gdouble x, gdouble y)
{
  draw_block_xor(doc, misc_setup_gc(doc), curr_block, x, y, 1, 1, 0);
}

void obj_xor_insert_scale_x(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble scale;

  misc_xor_line(doc, x, y);

  scale=dist(ix, iy, x, y);

  draw_block_xor(doc, misc_setup_gc(doc), curr_block, ix, iy, scale, 1, 0);
}

void obj_xor_insert_scale_y(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble scale;

  misc_xor_line(doc, x, y);

  scale=dist(ix, iy, x, y);

  draw_block_xor(doc, misc_setup_gc(doc), curr_block, ix, iy, scale_x, scale, 0);
}

void obj_xor_insert_rot(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble angle;

  misc_xor_line(doc, x, y);

  angle=get_angle(ix, iy, x, y);

  draw_block_xor(doc, misc_setup_gc(doc), curr_block, ix, iy, scale_x, scale_y, angle);
}

void obj_transform_insert(OBJECT *insert, TRANSFORM *trans)
{
  gdouble tx,ty;

  misc_transform_point_select(trans,&insert->data.block.insert_point,&tx,&ty);
  insert->data.block.insert_point.x=tx;
  insert->data.block.insert_point.y=ty;

  if(trans!=NULL && trans->type & OBJ_ROTATE)
    insert->data.block.rotation-=trans->rot_angle*180/M_PI;

  if(trans!=NULL && trans->type & OBJ_SCALE)
    {
      insert->data.block.scale_x*=trans->scale_x;
      insert->data.block.scale_y*=trans->scale_y;
    }
  
  obj_bound_insert(insert);
}

void obj_draw_insert_block(DOCUMENT *doc, GdkGC *gc, OBJECT *insert, TRANSFORM *trans)
{
  GdkGC *use_gc;

  TRANSFORM insert_transform;
  LINESTYLE *linestyle;
  LAYER *layer;
  OBJECT object;
  BLOCK_OBJECT *block_object;
  GList *list;
  GList *attr_list;

  /* Make tranform matrix */
  insert_transform.type=OBJ_TRANSLATE;
  insert_transform.trans_x=insert->data.block.insert_point.x;
  insert_transform.trans_y=insert->data.block.insert_point.y;
  insert_transform.cen_x=0;
  insert_transform.cen_y=0;
  
  /* Check if need to scale block */
  if(ABS(insert->data.block.scale_x-1)>GFLOAT_SENSITIVE ||
     ABS(insert->data.block.scale_y-1)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_SCALE;
      insert_transform.scale_x=insert->data.block.scale_x;
      insert_transform.scale_y=insert->data.block.scale_y;
    }

  /* Check if need to rotate block */
  if(ABS(insert->data.block.rotation)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_ROTATE;
      insert_transform.rot_angle=-insert->data.block.rotation*M_PI/180.0;
    }

  /* Get list of attributes in this block */
  attr_list=g_list_first(insert->data.block.attr);

  /* Get first object in block */
  list=g_list_first(insert->data.block.block->object_list);

  while(list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)list->data;

      /* Get layer for this object */

      /* Check if we should use the layer of the block */
      if(!strcmp(block_object->name_of_layer,"0"))
	layer=insert->layer;
      else
	{
	  layer=layer_find(doc, block_object->name_of_layer);
	  if(layer==NULL)
	    {
	      g_warning( _("Layer %s not found. Creating it."), block_object->name_of_layer );
	      layer=layer_new(doc, -1, doc->current_linestyle, 0.0 , 0.0, 0.0, 0.0,  
			  block_object->name_of_layer, TRUE, TRUE);
	    }
	}

      /* Draw only if layer is visible */      
      if(layer->visible)
	{
	  /* Make nice object to draw */
	  object.type=block_object->type;
	  object.layer=layer;
	  
	  /* Make colour */
	  object.type_colour=block_object->type_colour;
	  switch(object.type_colour)
	    {
	    case BYLAYER:	
	      object.colour=layer->colour;
	      break;
	    case BYBLOCK:
	      object.colour=insert->colour;
	      break;
	    case PREDEFINED:
	    case USERDEFINED:
	      object.colour=block_object->colour;
	      break;
	    }
	  
	  /* Make linestyle */
	  object.type_linestyle=block_object->type_linestyle;
	  switch(object.type_linestyle)
	    {
	    case BYLAYER:
	      object.linestyle=layer->linestyle;
	      break;
	    case BYBLOCK:
	      object.linestyle=insert->linestyle;
	      break;
	    case USERDEFINED:
	      linestyle=linestyle_find(doc, block_object->name_of_linestyle);
	      
	      if(linestyle==NULL)
		{
		  g_warning( _("Linestyle %s not found. Creating."), block_object->name_of_linestyle);
		  linestyle=linestyle_new(doc, -1, block_object->name_of_linestyle, LINE_SOLID, 1, 0,NULL);
		}
	      
	      object.linestyle=linestyle;
	      break;
	    }
	  
	  /* Make linewidth */
	  object.type_linewidth=block_object->type_linewidth;
	  switch(object.type_linewidth)
	    {
	    case BYLAYER:
	      object.linewidth=layer->linewidth;
	      break;
	    case BYBLOCK:
	      object.linewidth=insert->linewidth;
	      break;
	    case USERDEFINED:
	      object.linewidth=block_object->linewidth;
	      break;
	    }
	  
	  memcpy(&object.data, &block_object->data, sizeof(OBJECT_UNION));
	  
	  make_object_selected(&object, OBJECT_SEL, FALSE);
	  object.selected=insert->selected;
	  
	  if(trans==NULL)
	    use_gc=misc_setup_gc_object(doc, &object);
	  else
	    use_gc=gc;

	  switch(object.type)
	    {
	    case TYPE_LINE:
	      obj_transform_line( &object, &insert_transform);
	      obj_draw_line_block(doc, use_gc, &object, trans);
	      break;
	    case TYPE_CIRCLE:
	      obj_transform_circle( &object, &insert_transform);
	      obj_draw_circle_block(doc, use_gc, &object, trans);
	      break;
	    case TYPE_ARC:
	      obj_transform_arc( &object, &insert_transform);
	      obj_draw_arc_block(doc, use_gc, &object, trans);
	      break;
	    case TYPE_TEXT:
	      obj_transform_text( &object, &insert_transform);
	      obj_draw_text_block(doc, use_gc, &object, trans);
	      break;
	    case TYPE_ATTRIBUTE:
	      /* Fix text of attribute */
	      object.data.attribute.text.text=(gchar *)attr_list->data;
	      attr_list=g_list_next(attr_list);
	      obj_transform_attribute( &object, &insert_transform);
	      obj_draw_attribute_block(doc, use_gc, &object, trans);
	      break;
	    }
	}

      list=g_list_next(list);
    }
}

void obj_draw_insert(DOCUMENT *doc, OBJECT *insert)
{
  obj_draw_insert_block(doc, misc_setup_gc_object(doc, insert), insert, NULL);
}

gint obj_select_insert(OBJECT *insert, gdouble x, gdouble y, gdouble d)
{
  LAYER *layer;
  TRANSFORM insert_transform;
  OBJECT object;
  BLOCK_OBJECT *block_object;
  GList *list;
  GList *attr_list;

  /* Make tranform matrix */
  insert_transform.type=OBJ_TRANSLATE;
  insert_transform.trans_x=insert->data.block.insert_point.x;
  insert_transform.trans_y=insert->data.block.insert_point.y;
  insert_transform.cen_x=0;
  insert_transform.cen_y=0;

  /* Check if need to scale block */
  if(ABS(insert->data.block.scale_x-1)>GFLOAT_SENSITIVE ||
     ABS(insert->data.block.scale_y-1)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_SCALE;
      insert_transform.scale_x=insert->data.block.scale_x;
      insert_transform.scale_y=insert->data.block.scale_y;
    }

  /* Check if need to rotate block */
  if(ABS(insert->data.block.rotation)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_ROTATE;
      insert_transform.rot_angle=-insert->data.block.rotation*M_PI/180.0;
    }

  /* Get list of attributes in this block */
  attr_list=g_list_first(insert->data.block.attr);

  /* Get first object in block */
  list=g_list_first(insert->data.block.block->object_list);

  while(list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)list->data;

      /* Check if we should use the layer of the block */
      if(!strcmp(block_object->name_of_layer,"0"))
	layer=insert->layer;
      else
	{
	  layer=layer_find(insert->layer->doc, block_object->name_of_layer);
	  if(layer==NULL)
	    {
	      g_warning( _("Layer %s not found. Creating it."), block_object->name_of_layer );
	      layer=layer_new(insert->layer->doc, -1, insert->layer->doc->current_linestyle, 0.0 , 0.0, 0.0, 0.0,  
			  block_object->name_of_layer, TRUE, TRUE);
	    }
	}

      /* Make nice object to draw */
      object.type=block_object->type;	 
      object.layer=layer;

      /* Make linewidth */
      object.type_linewidth=block_object->type_linewidth;
      switch(object.type_linewidth)
	{
	case BYLAYER:
	  object.linewidth=layer->linewidth;
	  break;
	case BYBLOCK:
	  object.linewidth=insert->linewidth;
	  break;
	case USERDEFINED:
	  object.linewidth=block_object->linewidth;
	  break;
	}
      
      memcpy(&object.data, &block_object->data, sizeof(OBJECT_UNION));
      make_object_selected(&object, OBJECT_SEL, FALSE);
	  
      switch(object.type)
	{
	case TYPE_LINE:
	  obj_transform_line(&object, &insert_transform);
	  if(obj_select_line(&object, x, y, d))
	     return TRUE;
	  break;
	case TYPE_CIRCLE:
	  obj_transform_circle(&object, &insert_transform);
	  if(obj_select_circle(&object, x, y, d))
	     return TRUE;
	  break;
	case TYPE_ARC:
	  obj_transform_arc(&object, &insert_transform);
	  if(obj_select_arc(&object, x, y, d))
	     return TRUE;
	  break;
	case TYPE_TEXT:
	  obj_transform_text(&object, &insert_transform);
	  if(obj_select_text(&object, x, y, d))
	     return TRUE;
	  break;
	case TYPE_ATTRIBUTE:
	  obj_transform_attribute(&object, &insert_transform);
	  /* Fix text of attribute */
	  object.data.attribute.text.text=(gchar *)attr_list->data;
	  attr_list=g_list_next(attr_list);
	  if(obj_select_attribute(&object, x, y, d))
	     return TRUE;
	  break;
	}
      
      list=g_list_next(list);
    }

  return FALSE;
}


