/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include <string.h>
#include "struct.h"
#include "global.h"
#include "calc.h"
#include "dia_comm.h"
#include "dia_message.h"
#include "dia_coords.h"
#include "draw.h"
#include "xor.h"
#include "misc.h"
#include "parse.h"
#include "select.h"
#include "calc_arc.h"
#include "translation.h"


/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static gdouble start_width=0.0, end_width=0.0;
static gint node_type;
static gdouble old_tangent=-M_PI/2;
static gdouble sx,sy;

void obj_draw_polyline(DOCUMENT *doc, OBJECT *pline);
void obj_draw_polyline_block(DOCUMENT *doc, GdkGC *gc, OBJECT *pline, TRANSFORM *trans);

/* --  F U N C T I O N S  -- */

/* Calculates the midpoint and the endpoint of the arc */
static void calc_midpoint_and_endpoint_of_arc(OBJECT_ARC *arc, gdouble *mx, gdouble *my,
					      gdouble *ex, gdouble *ey)
{
  gdouble sa,ea,ma;

  /* Calc start and end angle */
  sa=arc->start_angle;
  ea=arc->end_angle;

  /* Find midangle */
  if(sa<ea)
    ma=(ea-sa)/2+sa;
  else
    ma=(ea+360-sa)/2+sa;

  /* Fix nice values */
  while(ma>360) ma-=360;
  while(ma<0) ma+=360;

  /* Calculate midpoint coordinate */
  *mx=arc->centrum.x+arc->radius*cos(ma*M_PI/180);
  *my=arc->centrum.y-arc->radius*sin(ma*M_PI/180);

  /* Calculate endpoint coordinate */
  *ex=arc->centrum.x+arc->radius*cos(ea*M_PI/180);
  *ey=arc->centrum.y-arc->radius*sin(ea*M_PI/180);
}

gdouble calc_tangent(GList *pline, TRANSFORM *trans)
{
  GList *last_node;
  GList *slast_node;
  gdouble x,y;
  OBJECT_ARC arc;

  last_node=g_list_last(pline);
  slast_node=g_list_previous(last_node);

  if(!(last_node==NULL || slast_node==NULL))  /* First node. No calc! */
    {
      misc_transform_point(trans, ((NODE *)last_node->data)->x, 
			   ((NODE *)last_node->data)->y, &x,&y);
      
      if(((NODE *)last_node->data)->type==NODE_LINE && ((NODE *)slast_node->data)->type==NODE_ARC)
	{
	  gdouble Ax,Ay,Cx,Cy;
	  gdouble d;
	  
	  misc_transform_point(trans, ((NODE *)slast_node->prev->data)->x,
			       ((NODE *)slast_node->prev->data)->y, &Ax,&Ay);
	  misc_transform_point(trans, ((NODE *)slast_node->data)->x,
			       ((NODE *)slast_node->data)->y, &Cx,&Cy);

	  calc_arc_3p(Ax,Ay,Cx,Cy,x,y,&arc);

	  d=dist(Ax,Ay,x,y);
	  if(  ((Ay-Cy)*(x-Ax)-(Ax-Cx)*(y-Ay))/(d*d)<0)	
	    return (get_angle(arc.centrum.x, arc.centrum.y, x,y)*M_PI/180);
	  else
	    return (get_angle(x,y,arc.centrum.x, arc.centrum.y)*M_PI/180);	  
	}
      else
	{
	  gdouble Cx,Cy;

	  misc_transform_point(trans, ((NODE *)slast_node->data)->x,
			       ((NODE *)slast_node->data)->y, &Cx,&Cy);
	  return (get_angle(Cx,Cy,x,y)*M_PI/180)-M_PI/2;
	}
    }
  
  return -M_PI/2;
}


NODE *add_node(OBJECT *pline, gint type, gdouble x, gdouble y, gdouble sw, gdouble ew, gdouble w)
{
  GList *last_node;
  NODE *temp;

  last_node=g_list_last(pline->data.polyline.node_list);

  temp=(NODE *)g_malloc(sizeof(NODE));

  temp->selected=FALSE;
  temp->type=type;
  temp->x=x;
  temp->y=y;
  temp->sw=sw;
  temp->ew=ew;
  temp->weight=w;

  pline->data.polyline.node_list=g_list_append(pline->data.polyline.node_list, (gpointer)temp);

  return temp;
}

OBJECT *create_new_polyline(DOCUMENT *doc, LAYER *layer, gchar *name, 
			    gint type_colour, COLOUR *colour, 
			    gint type_linestyle, LINESTYLE *linestyle, 
			    gint type_linewidth, gdouble linewidth,
			    gint closed)
{
  OBJECT *obj;

  /* Make sure we have valid parameters */
  if(doc==NULL || layer==NULL)
    {
      g_warning( _("Unvalid parameters for function %s in file %s"), __FUNCTION__, __FILE__);
      return NULL;
    }
  
  /* Create object */
  obj=g_new(OBJECT, 1);
  
  /* Type of object */
  obj->type=TYPE_POLYLINE;
  
  /* Pointer to layer */
  obj->layer=layer;

  /* Set up attributes for this object */
  obj_set_attributes(doc, obj, name, type_colour, colour, type_linestyle, 
		     linestyle, type_linewidth, linewidth); 
  
  /* Set object specific attributes */  
  obj->data.polyline.type=0;
  obj->data.polyline.node_list=NULL;
  obj->data.polyline.closed=closed;

  /* Return new object */
  return obj;
}

void undo_polyline(DOCUMENT *doc, OBJECT *pline)
{
  GList *list;
  GList *plist;

  list=g_list_last(pline->data.polyline.node_list);

  if(g_list_previous(list)==NULL && g_list_next(list)==NULL)
    {
      print( _("\n  Cannot undo first point."));
    }
  else
    {
      remove_old();

      if(((NODE *)g_list_previous(list)->data)->type==NODE_LINE)
	{
	  gl_oldmx=((NODE *)g_list_previous(list)->data)->x;
	  gl_oldmy=((NODE *)g_list_previous(list)->data)->y;

	  pline->data.polyline.node_list=g_list_remove(pline->data.polyline.node_list, list->data);
	}
      else
	{
	  plist=g_list_previous(list);

	  gl_oldmx=((NODE *)g_list_previous(plist)->data)->x;
	  gl_oldmy=((NODE *)g_list_previous(plist)->data)->y;

	  pline->data.polyline.node_list=g_list_remove(pline->data.polyline.node_list, list->data);
	  pline->data.polyline.node_list=g_list_remove(pline->data.polyline.node_list, plist->data);
	}

      old_tangent=calc_tangent(pline->data.polyline.node_list, NULL);
    }
  set_prompt();
}

void obj_set_prompt_polyline(void)
{
  switch(gl_stage)
    {
    case 0:
      set_prompt_text( _("Start Point: "));
      break;
    case 1:
      set_prompt_text( _("Undo/Arc/Close/Length/<End point>: "));
      break;
    case 2:
      set_prompt_text( _("Starting width: "));
      break;
    case 3:
      set_prompt_text( _("Ending width: "));
      break;
    case 4:
      set_prompt_text( _("Undo/Angle/cEnter/Close/Direction/Line/Radius/Second point/<Endpoint>: "));
      break;
    case 5:
      set_prompt_text( _("Undo/<Second point>: "));
      break;
    case 6:
    case 13:
      set_prompt_text( _("Undo/<End point>: "));
      break;
    case 7:
      set_prompt_text( _("Undo/<Center point>: "));
      break;
    case 8:
      set_prompt_text( _("Undo/Angle/Length of coord/<End point>: "));
      break;
    case 9:
      set_prompt_text( _("Undo/<Included angle>: "));
      break;
    case 10:
      set_prompt_text( _("Undo/<Length of coord>: "));
      break;
    case 11:
      set_prompt_text( _("Undo/<Length of line>: "));
      break;
    case 12:
      set_prompt_text( _("Undo/<Direction from start point>: "));
      break;
    }
}

void obj_set_tool_polyline(GtkWidget *widget, gpointer data)
{
  remove_old();

  if(!command_from_keyboard)
    print( _("pline"));

  if(work_document!=NULL)
    deselect_all_objects(work_document);

  if(misc_object_get_document()==NULL)
    return;
  
  gl_tool=TOOL_POLYLINE;
  gl_stage=0;
  work_document=NULL;

  node_type=NODE_LINE;
  old_tangent=-M_PI/2;

  set_prompt();
  set_cursor(NULL);
  UpdateCoordsLabel();
}

void obj_bound_polyline(OBJECT *pline)
{
  gboolean first=TRUE;
  OBJECT_UNION s;
  gint type;
  gboolean ret;
  gboolean (*draw_vertex)(gint *type, OBJECT_UNION *s);
  
  draw_vertex=draw_polyline_set_polyline(pline, NULL);

  ret=draw_vertex(&type,&s);

  while(ret)
    {
      ret=draw_vertex(&type,&s);

      if(!ret)
	break;
      
      switch(type)
	{
	case NODE_LINE:
	  if(first)
	    {
	      pline->bounding_box[0]=MIN(s.line.p1.x, s.line.p2.x);
	      pline->bounding_box[1]=MIN(s.line.p1.y, s.line.p2.y);
	      pline->bounding_box[2]=MAX(s.line.p1.x, s.line.p2.x);
	      pline->bounding_box[3]=MAX(s.line.p1.y, s.line.p2.y);
	      first=FALSE;
	    }
	  else
	    {
	      pline->bounding_box[0]=MIN(pline->bounding_box[0], s.line.p1.x);
	      pline->bounding_box[0]=MIN(pline->bounding_box[0], s.line.p2.x);
	      pline->bounding_box[1]=MIN(pline->bounding_box[1], s.line.p1.y);
	      pline->bounding_box[1]=MIN(pline->bounding_box[1], s.line.p2.y);
	      pline->bounding_box[2]=MAX(pline->bounding_box[2], s.line.p1.x);
	      pline->bounding_box[2]=MAX(pline->bounding_box[2], s.line.p2.x);
	      pline->bounding_box[3]=MAX(pline->bounding_box[3], s.line.p1.y);
	      pline->bounding_box[3]=MAX(pline->bounding_box[3], s.line.p2.y);
	    }
	  break;
	case NODE_ARC:
	  if(first)
	    {
	      gdouble sx,sy,ex,ey;
	      gint i;
	      
	      sx=s.arc.centrum.x+s.arc.radius*cos(s.arc.start_angle*M_PI/180);
	      sy=s.arc.centrum.y-s.arc.radius*sin(s.arc.start_angle*M_PI/180);
	      
	      ex=s.arc.centrum.x+s.arc.radius*cos(s.arc.end_angle*M_PI/180);
	      ey=s.arc.centrum.y-s.arc.radius*sin(s.arc.end_angle*M_PI/180);
	      
	      pline->bounding_box[0]=MIN(sx,ex);
	      pline->bounding_box[1]=MIN(sy,ey);
	      pline->bounding_box[2]=MAX(sx,ex);
	      pline->bounding_box[3]=MAX(sy,ey);
	      
	      /* Check each quadrant */
	      for(i=90;i<=360;i+=90)
		{
		  
		  if(misc_within_angle(s.arc.start_angle, s.arc.end_angle, i))
		    {
		      sx=s.arc.centrum.x+s.arc.radius*cos(i*M_PI/180);
		      sy=s.arc.centrum.y-s.arc.radius*sin(i*M_PI/180);
		      
		      pline->bounding_box[0]=MIN(pline->bounding_box[0],sx);
		      pline->bounding_box[1]=MIN(pline->bounding_box[1],sy);
		      pline->bounding_box[2]=MAX(pline->bounding_box[2],sx);
		      pline->bounding_box[3]=MAX(pline->bounding_box[3],sy);
		    }
		}	      
	      first=FALSE;
	    }
	  else
	    {
	      gdouble sx,sy,ex,ey;
	      gint i;
	      
	      sx=s.arc.centrum.x+s.arc.radius*cos(s.arc.start_angle*M_PI/180);
	      sy=s.arc.centrum.y-s.arc.radius*sin(s.arc.start_angle*M_PI/180);
	      
	      ex=s.arc.centrum.x+s.arc.radius*cos(s.arc.end_angle*M_PI/180);
	      ey=s.arc.centrum.y-s.arc.radius*sin(s.arc.end_angle*M_PI/180);
	      
	      pline->bounding_box[0]=MIN(pline->bounding_box[0],sx);
	      pline->bounding_box[0]=MIN(pline->bounding_box[0],ex);
	      pline->bounding_box[1]=MIN(pline->bounding_box[1],sy);
	      pline->bounding_box[1]=MIN(pline->bounding_box[1],ey);
	      pline->bounding_box[2]=MAX(pline->bounding_box[2],sx);
	      pline->bounding_box[2]=MAX(pline->bounding_box[2],ex);
	      pline->bounding_box[3]=MAX(pline->bounding_box[3],sy);
	      pline->bounding_box[3]=MAX(pline->bounding_box[3],ey);
	      
	      /* Check each quadrant */
	      for(i=90;i<=360;i+=90)
		{
		  
		  if(misc_within_angle(s.arc.start_angle, s.arc.end_angle, i))
		    {
		      sx=s.arc.centrum.x+s.arc.radius*cos(i*M_PI/180);
		      sy=s.arc.centrum.y-s.arc.radius*sin(i*M_PI/180);
		      
		      pline->bounding_box[0]=MIN(pline->bounding_box[0],sx);
		      pline->bounding_box[0]=MIN(pline->bounding_box[0],ex);
		      pline->bounding_box[1]=MIN(pline->bounding_box[1],sy);
		      pline->bounding_box[1]=MIN(pline->bounding_box[1],ey);
		      pline->bounding_box[2]=MAX(pline->bounding_box[2],sx);
		      pline->bounding_box[2]=MAX(pline->bounding_box[2],ex);
		      pline->bounding_box[3]=MAX(pline->bounding_box[3],sy);
		      pline->bounding_box[3]=MAX(pline->bounding_box[3],ey);
		    }
		}	      
	    }

	  break;
	}
    }
}

void obj_button_polyline(DOCUMENT *doc, gint button, gdouble x, gdouble y)
{
  OBJECT_ARC arc;
  gdouble mx,my, ex,ey, ea, sa;

  /* Right mousebutton => done/cancel */
  if(button==3)
    {
      remove_old();
      if(work_document!=NULL && doc->working_polyline!=NULL)
	{
	  /* Make sure that the polyline contains more than one node */
	  if(g_list_length(doc->working_polyline->data.polyline.node_list)>1)
	    {
	      obj_bound_polyline(doc->working_polyline);
	      set_current_attributes(doc, doc->working_polyline);
	      obj_draw_polyline(doc, doc->working_polyline);
	      doc->current_layer->object_list=g_list_append(doc->current_layer->object_list,(gpointer)doc->working_polyline);
	      doc->working_polyline=NULL;      
	      doc->saved=FALSE;
	      make_window_title(doc);
	    }
	}
      tool_idle(doc);
      return;
    }

  switch(gl_stage)
    {
    case 0:
      /* Do we have an old polyline? If so erase it. */

      if(doc->working_polyline!=NULL)  /* <-- Memory leak!!! The nodes are not released */
	g_free(doc->working_polyline);
      
      if(doc->gl_snap)
	make_snap(doc, &x, &y);

      gl_oldmx=x;
      gl_oldmy=y;
      gl_stage=1;
      doc->working_polyline=create_new_polyline(doc, doc->current_layer, NULL, 
						doc->type_colour, doc->current_colour, 
						doc->type_linestyle, doc->current_linestyle,
						doc->type_linewidth, doc->current_linewidth,
						POLYLINE_OPEN);
      add_node(doc->working_polyline, NODE_LINE, x,y, start_width, end_width, 1);
      set_prompt();
      break;
    case 1:
      make_ortho_snap(doc,gl_oldmx, gl_oldmy, &x, &y);

      gl_oldmx=x;
      gl_oldmy=y;
      set_prompt();
      add_node(doc->working_polyline, NODE_LINE, x,y, start_width, end_width, 1);

      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);

      UpdateSingelCoord(0, gl_oldmx);
      UpdateSingelCoord(1, gl_oldmy);
      break;
    case 4:
      make_ortho_snap(doc,gl_oldmx, gl_oldmy, &x, &y);

      calc_arc_2p(old_tangent, gl_oldmx, gl_oldmy, x,y, &mx, &my, &arc);

      add_node(doc->working_polyline, NODE_ARC, mx,my, start_width, end_width, 1);
      add_node(doc->working_polyline, NODE_LINE, x,y, start_width, end_width, 1);

      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_oldmx=x;
      gl_oldmy=y;
      set_prompt();

      UpdateSingelCoord(0, gl_oldmx);
      UpdateSingelCoord(1, gl_oldmy);
      break;
    case 5:   /* Second point for arc */
      make_ortho_snap(doc,gl_oldmx, gl_oldmy, &x, &y);
      sx=x;
      sy=y;
      gl_stage++;
      set_prompt();
      break;
    case 6:  /* End point for arc */
      if(doc->gl_snap)
	make_snap(doc, &x, &y);

      remove_old();

      add_node(doc->working_polyline, NODE_ARC, sx, sy, start_width, end_width, 1);
      add_node(doc->working_polyline, NODE_LINE, x,y, start_width, end_width, 1);

      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_oldmx=x;
      gl_oldmy=y;
      gl_stage=4;
      set_prompt();
      break;
    case 7: /* Center point */
      make_ortho_snap(doc,gl_oldmx, gl_oldmy, &x, &y);
      sx=x;
      sy=y;
      gl_stage++;
      set_prompt();
      break;
    case 8:  /* End point 2 */
      if(doc->gl_snap)
	make_snap(doc, &x, &y);
      
      remove_old();

      calc_arc_CSE(sx, sy, gl_oldmx, gl_oldmy, x, y, &arc);
      calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);
      
      add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
      add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);
     
      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_oldmx=ex;
      gl_oldmy=ey;
      gl_stage=4;
      set_prompt();
      break;
    case 9:  /* Included angle */ 
      if(doc->gl_snap)
	make_snap(doc, &x, &y);
      
      remove_old();  

      sa=get_angle(sx, sy, gl_oldmx, gl_oldmy);
      ea=get_angle(sx, sy, x, y);
      
      if(sa<ea)
	ea-=sa;
      else
	ea=ea+(360.0-sa);
      
      if(calc_arc_CSA(sx, sy, gl_oldmx, gl_oldmy, ea, &arc))
	{
	  calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);	 
	  add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
	  add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);	 
	}

      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_oldmx=ex;
      gl_oldmy=ey;
      gl_stage=4;
      set_prompt();
      break;
    case 10: /* Length of coord */
      if(doc->gl_snap)
	make_snap(doc, &x, &y);
      
      remove_old();  

      if(calc_arc_CSL(sx, sy, gl_oldmx, gl_oldmy, dist(x,y,gl_oldmx, gl_oldmy), &arc))
	{
	  calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);	  
	  add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
	  add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);
	}
      
      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_oldmx=ex;
      gl_oldmy=ey;
      gl_stage=4;
      set_prompt();
      break;
    case 11:  /* Length of line */
      if(doc->gl_snap)
	make_snap(doc, &x, &y);
      
      remove_old();  

      ex=gl_oldmx+dist(gl_oldmx, gl_oldmy, x, y)*cos(old_tangent+M_PI/2);
      ey=gl_oldmy-dist(gl_oldmx, gl_oldmy, x, y)*sin(old_tangent+M_PI/2);

      add_node(doc->working_polyline, NODE_LINE, ex, ey, 1,1,1);
      
      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_oldmx=ex;
      gl_oldmy=ey;
      gl_stage=1;
      set_prompt();
      break;
    case 12:  /* Direction from start */
      make_ortho_snap(doc,gl_oldmx, gl_oldmy, &x, &y);      
      remove_old();
      old_tangent=get_angle(gl_oldmx, gl_oldmy, x,y)*M_PI/180.0-M_PI/2;
      gl_stage=13;
      set_prompt();
      break;
    case 13:  /* End point of direction */
      if(doc->gl_snap)
	make_snap(doc, &x, &y);
      
      remove_old();  

      calc_arc_2p(old_tangent, gl_oldmx, gl_oldmy, x,y, &mx, &my, &arc);

      add_node(doc->working_polyline, NODE_ARC, mx,my, start_width, end_width, 1);
      add_node(doc->working_polyline, NODE_LINE, x,y, start_width, end_width, 1);

      /* Figure out tangent for contiuning arcs */
      old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
      gl_stage=4;
      gl_oldmx=x;
      gl_oldmy=y;
      set_prompt();
      break;
    }
}

void obj_prompt_polyline(DOCUMENT *doc, gchar *buffer)
{
  OBJECT_ARC arc;
  gdouble tx,ty, value, ex, ey, mx, my;
  gint mn;

  switch(gl_stage)
    {
    case 0:  /* Start point */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  FALSE,NULL,0,0,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      }
      break;
    case 1:  /* End point */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/C/A/L/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	switch(mn)
	  {
	  case 0:   /* Undo */
	    undo_polyline(doc, doc->working_polyline);
	    break;
	  case 1:  /* Close */
	    remove_old();
	    obj_bound_polyline(doc->working_polyline);
	    doc->working_polyline->data.polyline.closed=POLYLINE_CLOSE_LINE;
	    set_current_attributes(doc, doc->working_polyline);
	    obj_draw_polyline(doc, doc->working_polyline);
	    doc->current_layer->object_list=g_list_append(doc->current_layer->object_list,(gpointer)doc->working_polyline);
	    doc->working_polyline=NULL;      
	    doc->saved=FALSE;
	    make_window_title(doc);
	    tool_idle(doc);
	    break;
	  case 2:   /* Arc */
	    gl_stage=4;
	    node_type=NODE_ARC;
	    set_prompt();
	    break;
	  case 3:   /* Length */
	    gl_stage=11;
	    set_prompt();
	    break;
	  }
	break;
      }
      break;
    case 4:   /* Arc prompt */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/A/E/C/D/L/R/S/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);	
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	remove_old();
	switch(mn)
	  {
	  case 0:  /* Undo */
	    undo_polyline(doc, doc->working_polyline);
	    break;
	  case 1:  /* Angle */
	    break;
	  case 2:  /* Center */
	    gl_stage=7;
	    set_prompt();
	    break;
	  case 3:  /* Close */
	    remove_old();
	    obj_bound_polyline(doc->working_polyline);
	    doc->working_polyline->data.polyline.closed=POLYLINE_CLOSE_ARC;
	    set_current_attributes(doc, doc->working_polyline);
	    obj_draw_polyline(doc, doc->working_polyline);
	    doc->current_layer->object_list=g_list_append(doc->current_layer->object_list,(gpointer)doc->working_polyline);
	    doc->working_polyline=NULL;      
	    doc->saved=FALSE;
	    make_window_title(doc);
	    tool_idle(doc);
	    break;
	  case 4:   /* Direction */
	    gl_stage=12;
	    set_prompt();
	    break;
	  case 5:   /* Line */
	    gl_stage=1;
	    node_type=NODE_LINE;
	    set_prompt();
	    break;
	  case 6:  /* Radius */
	    break;
	  case 7:  /* Second point */
	    gl_stage=5;
	    set_prompt();
	    break;
	  }
	break;
      }
      break;
    case 5:  /* Second point */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=4;
	set_prompt();
	break;
      }
      break;
    case 6:  /* End point */
     switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=5;
	set_prompt();
	break;
      }
      break;
    case 7: /* Center */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=4;
	set_prompt();
	break;
      }
      break;
    case 8: /* End point */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/A/L/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	remove_old();
	switch(mn)
	  {
	  case 0:   /* Undo */
	    gl_stage=7;
	    break;
	  case 1:  /* Angle */
	    gl_stage=9;
	    break;
	  case 2:  /* Length of coord */
	    gl_stage=10;
	    break;
	  }
	set_prompt();
	break;
      }
      break;
    case 9: /* Included angle */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1|PROMPT_MENU|PROMPT_RETURN,buffer,
			  FALSE,"U/",0,0,NULL,NULL,&value,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,0,0);
	break;
      case PROMPT_POINT1:
	work_document=doc;
	remove_old();

	while(value<0) value+=360;
	while(value>360) value-=360;	

	if(calc_arc_CSA(sx, sy, gl_oldmx, gl_oldmy, value, &arc))
	  {
	    calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);	 
	    add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
	    add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);	 
	  }
	
	/* Figure out tangent for contiuning arcs */
	old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
	gl_oldmx=ex;
	gl_oldmy=ey;
	gl_stage=4;
	set_prompt();
       	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=8;
	set_prompt();
	break;
      }
      break;
    case 10: /* Length of coord */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1|PROMPT_MENU|PROMPT_RETURN,buffer,
			  FALSE,"U/",0,0,NULL,NULL,&value,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,0,0);
	break;
      case PROMPT_POINT1:
	work_document=doc;
	remove_old();

	if(calc_arc_CSL(sx, sy, gl_oldmx, gl_oldmy, value, &arc))
	{
	  calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);	  
	  add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
	  add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);
	}
      	
	/* Figure out tangent for contiuning arcs */
	old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
	gl_oldmx=ex;
	gl_oldmy=ey;
	gl_stage=4;
	set_prompt();
       	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=8;
	set_prompt();
	break;
      }
      break;
    case 11:  /* Length of line */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1|PROMPT_MENU|PROMPT_RETURN,buffer,
			  FALSE,"U/",0,0,NULL,NULL,&value,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,0,0);
	break;
      case PROMPT_POINT1:
	work_document=doc;
	remove_old();

	ex=gl_oldmx+value*cos(old_tangent+M_PI/2);
	ey=gl_oldmy-value*sin(old_tangent+M_PI/2);
	
	add_node(doc->working_polyline, NODE_LINE, ex, ey, 1,1,1);
     
	/* Figure out tangent for contiuning arcs */
	old_tangent=calc_tangent(doc->working_polyline->data.polyline.node_list, NULL);
	gl_oldmx=ex;
	gl_oldmy=ey;
	gl_stage=1;
	set_prompt();
       	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=1;
	set_prompt();
	break;
      }
      break;
    case 12:  /* Direction from start */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT1|PROMPT_MENU|PROMPT_RETURN,buffer,
			  FALSE,"U/",0,0,NULL,NULL,&value,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,0,0);
	break;
      case PROMPT_POINT1:
	work_document=doc;
	remove_old();

	old_tangent=value*M_PI/180-M_PI/2;
	gl_stage=13;
	set_prompt();
       	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=4;
	set_prompt();
	break;
      }
      break;
    case 13:  /* End point for direction */
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_RETURN|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("U/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	obj_button_polyline(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	obj_button_polyline(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	remove_old();
	gl_stage=12;
	set_prompt();
	break;
      }
      break;
    }
}

void obj_xor_polyline(DOCUMENT *doc,gdouble x, gdouble y)
{
  NODE *data;

  data=add_node(doc->working_polyline, NODE_LINE, x,y, 1,1,1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data);
}

void obj_xor_polyline_length_of_line(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble length;
  NODE *data;

  misc_setup_black_line(doc);
  gdk_draw_line(doc->widgets.draw->window, 
		doc->xor_gc,
		WLDtoSCR_x(doc, gl_oldmx), WLDtoSCR_y(doc, gl_oldmy),
		WLDtoSCR_x(doc, x), WLDtoSCR_y(doc, y));

  length=dist(gl_oldmx, gl_oldmy, x, y);

  data=add_node(doc->working_polyline, NODE_LINE, 
		gl_oldmx+length*cos(old_tangent+M_PI/2), gl_oldmy-length*sin(old_tangent+M_PI/2), 
		1,1,1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data);  
}

void obj_xor_polyline_second_point(DOCUMENT *doc,gdouble x, gdouble y)
{
  NODE *data;

  data=add_node(doc->working_polyline, NODE_LINE, x,y, 1,1,1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data);
}

void obj_xor_polyline_end_point(DOCUMENT *doc,gdouble x, gdouble y)
{
  NODE *data1, *data2;
  
  data1=add_node(doc->working_polyline, NODE_ARC, sx, sy, 1,1,1);
  data2=add_node(doc->working_polyline, NODE_LINE, x, y, 1,1,1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data1);
  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data2);
}

void obj_xor_polyline_center_point(DOCUMENT *doc,gdouble x, gdouble y)
{
  NODE *data;

  data=add_node(doc->working_polyline, NODE_LINE, x,y, 1,1,1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data);
}

void obj_xor_polyline_end_point_2(DOCUMENT *doc,gdouble x, gdouble y)
{
  OBJECT_ARC arc;
  gdouble mx,my, ex,ey;
  NODE *data1, *data2;

  misc_setup_black_line(doc);
  gdk_draw_line(doc->widgets.draw->window, 
		doc->xor_gc,
		WLDtoSCR_x(doc,sx),WLDtoSCR_y(doc,sy),
		WLDtoSCR_x(doc,x),WLDtoSCR_y(doc,y));

  calc_arc_CSE(sx, sy, gl_oldmx, gl_oldmy, x, y, &arc);
  calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);
  
  data1=add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
  data2=add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data1);
  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data2);
}

void obj_xor_polyline_included_angle(DOCUMENT *doc,gdouble x, gdouble y)
{
  NODE *data1, *data2;
  gdouble eangle, sangle, mx, my, ex, ey;
  OBJECT_ARC arc;
  
  sangle=get_angle(sx, sy, gl_oldmx, gl_oldmy);
  eangle=get_angle(sx, sy, x, y);

  if(sangle<eangle)
    eangle-=sangle;
  else
    eangle=eangle+(360.0-sangle);

  /*  UpdateSingelCoord(0,eangle); */

  misc_setup_black_line(doc);
  gdk_draw_line(doc->widgets.draw->window, 
		doc->xor_gc,
		WLDtoSCR_x(doc, sx), WLDtoSCR_y(doc, sy),
		WLDtoSCR_x(doc, x), WLDtoSCR_y(doc, y));
  
  if(calc_arc_CSA(sx, sy, gl_oldmx, gl_oldmy, eangle, &arc))
    {
      calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);
      
      data1=add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
      data2=add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);
      
      obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);
      
      doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data1);
      doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data2);
    }
}

void obj_xor_polyline_length_of_coord(DOCUMENT *doc,gdouble x, gdouble y)
{
  gdouble e, ex, ey, mx, my;
  OBJECT_ARC arc;
  NODE *data1, *data2;

  e=dist(x,y,gl_oldmx, gl_oldmy);

  /*  UpdateSingelCoord(0,e); */

  misc_setup_black_line(doc);
  gdk_draw_line(doc->widgets.draw->window, 
		doc->xor_gc,
		WLDtoSCR_x(doc, gl_oldmx), WLDtoSCR_y(doc, gl_oldmy),
		WLDtoSCR_x(doc, x), WLDtoSCR_y(doc, y));
  
  if(calc_arc_CSL(sx, sy, gl_oldmx, gl_oldmy, e, &arc))
    {
      calc_midpoint_and_endpoint_of_arc(&arc, &mx, &my, &ex, &ey);
      
      data1=add_node(doc->working_polyline, NODE_ARC, mx, my, 1, 1, 1);
      data2=add_node(doc->working_polyline, NODE_LINE, ex, ey, 1, 1, 1);
      
      obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);
      
      doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data1);
      doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data2);
    }
 
}

void obj_xor_polyline_direction(DOCUMENT *doc, gdouble x, gdouble y)
{
  NODE *data;

  data=add_node(doc->working_polyline, NODE_LINE, x,y, 1,1,1);

  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data);
}

void obj_xor_polyline_arc(DOCUMENT *doc,gdouble x, gdouble y)
{
  OBJECT_ARC arc;
  NODE *data, *data2;
  gdouble mx,my;

  calc_arc_2p(old_tangent, gl_oldmx, gl_oldmy, x,y, &mx, &my, &arc);

  data=add_node(doc->working_polyline, NODE_ARC, mx,my, 1,1,1);
  data2=add_node(doc->working_polyline, NODE_LINE, x,y, 1,1,1);
  
  obj_draw_polyline_block(doc, misc_setup_gc(doc), doc->working_polyline, NULL);

  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data);
  doc->working_polyline->data.polyline.node_list=g_list_remove(doc->working_polyline->data.polyline.node_list,data2);
}

void obj_transform_polyline(OBJECT *pline, TRANSFORM *trans)
{
  POINT t;
  GList *node_list;
  gdouble tx,ty;

  node_list=g_list_first(pline->data.polyline.node_list);
  
  while(node_list!=NULL)
    {
      t.x=((NODE *)node_list->data)->x;
      t.y=((NODE *)node_list->data)->y;
      t.selected=((NODE *)node_list->data)->selected;
      misc_transform_point_select(trans, &t, &tx,&ty);
      ((NODE *)node_list->data)->x=tx;
      ((NODE *)node_list->data)->y=ty; 

      node_list=g_list_next(node_list);
    }

  obj_bound_polyline(pline);
}

void obj_draw_polyline_block(DOCUMENT *doc, GdkGC *gc, OBJECT *pline, TRANSFORM *trans)
{
  OBJECT_UNION s;
  gint type;
  gboolean ret;
  gboolean (*draw_vertex)(gint *type, OBJECT_UNION *s);
  
  draw_vertex=draw_polyline_set_polyline(pline, trans);

  ret=draw_vertex(&type,&s);

  while(ret)
    {
      ret=draw_vertex(&type,&s);

      if(!ret)
	break;
      
      switch(type)
	{
	case NODE_LINE:
	  gdk_draw_line(doc->widgets.draw->window, gc, 
			WLDtoSCR_x(doc,s.line.p1.x),WLDtoSCR_y(doc,s.line.p1.y),
			WLDtoSCR_x(doc,s.line.p2.x),WLDtoSCR_y(doc,s.line.p2.y));
	  break;
	case NODE_ARC:
	  draw_arc(doc->widgets.draw->window, gc, FALSE,
		   WLDtoSCR_x(doc, s.arc.centrum.x-s.arc.radius), WLDtoSCR_y(doc, s.arc.centrum.y-s.arc.radius),
		   WLDtoSCR(doc, s.arc.radius*2), WLDtoSCR(doc, s.arc.radius*2),
		   s.arc.start_angle, s.arc.end_angle);
	  break;
	}

    }

  return;
}

void obj_draw_polyline(DOCUMENT *doc, OBJECT *pline)
{
  obj_draw_polyline_block(doc,misc_setup_gc_object(doc, pline),pline,NULL);
}

gint obj_select_polyline(OBJECT *pline, gdouble x, gdouble y, gdouble d)
{
  OBJECT_UNION s;
  gint type;
  gboolean ret;
  gboolean (*draw_vertex)(gint *type, OBJECT_UNION *s);

  draw_vertex=draw_polyline_set_polyline(pline, NULL);

  ret=draw_vertex(&type,&s);

  while(ret)
    {
      ret=draw_vertex(&type,&s);

      if(!ret)
	break;
      
      switch(type)
	{
	case NODE_LINE:
	  if(misc_point_on_line(x,y,s.line.p1.x,s.line.p1.y,s.line.p2.x,s.line.p2.y,d))
	    return TRUE;
	  break;
	case NODE_ARC:
	  if(misc_point_on_arc(x,y,&s.arc,d))	    
	    return TRUE; 
	  break;
	}
    }
  
  return FALSE;
}

gint obj_select_polyline_object(OBJECT *pline, gdouble x, gdouble y, gdouble d, gint *type, OBJECT_UNION *s)
{
  gboolean ret;
  gboolean (*draw_vertex)(gint *type, OBJECT_UNION *s);

  draw_vertex=draw_polyline_set_polyline(pline, NULL);

  ret=draw_vertex(type,s);

  while(ret)
    {
      ret=draw_vertex(type,s);

      if(!ret)
	break;
      
      switch(*type)
	{
	case NODE_LINE:
	  if(misc_point_on_line(x,y,s->line.p1.x,s->line.p1.y,s->line.p2.x,s->line.p2.y,d))
	    return TRUE;
	  break;
	case NODE_ARC:
	  if(misc_point_on_arc(x,y,&(s->arc),d))	    
	    return TRUE; 
	  break;
	}
    }
  
  return FALSE;
}






