/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include "translation.h"
#include "struct.h"
#include "global.h"
#include "misc.h"
#include "xor.h"
#include "select.h"
#include "layer.h"
#include "math_tangent.h"
#include "dia_comm.h"
#include "dia_message.h"
#include "dia_toolbox.h"
#include "obj_line.h"
#include "obj_cline.h"
#include "obj_rline.h"
#include "obj_circle.h"
#include "obj_arc.h"
#include "obj_ellipse.h"
#include "obj_elarc.h"
#include "obj_insert.h"
#include "obj_point.h"
#include "obj_text.h"
#include "obj_polyline.h"
#include "obj_spline.h"
#include "obj_attribute.h"


/* Object:   Osnap that works  (Osnap that are not yet implemented)

  Line:      END, MID, NEA
  Circle:    CEN, NEA, QUA, TAN
  Arc:       END, MID, CEN, NEA, QUA, TAN
  Text:      INS
  Attribute: INS
  Insert:    INS
  Point:     NOD, NEA
  Ray:       END, NEA
  Cline:     NEA
  Polyline:  END, MID, CEN, NEA, QUA
  Spline:    END, CEN, NEA
  Ellipse:   CEN, (TAN)
  Elarc:     CEN, MID, END, (TAN)

 */


/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static gdouble sens = 5;

/* --  F U N C T I O N S  -- */

/* Return True if it is possible to use osnap TAN.
   False otherwise. */
static gboolean osnap_tan_possible(void)
{
  return( (gl_tool==TOOL_LINE && gl_stage==1) ||
	  (gl_tool==TOOL_RLINE && gl_stage==1) ||
	  (gl_tool==TOOL_CLINE && gl_stage==1) ||
	  (gl_tool==TOOL_POLYLINE && gl_stage==1)	  
	 );
}

static gboolean obj_osnap_line(gint osnap, OBJECT *line, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  gdouble tx,ty;

  switch(osnap)
    {
    case OSNAP_END:
      tx=line->data.line.p1.x;
      ty=line->data.line.p1.y;
      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }
      tx=line->data.line.p2.x;
      ty=line->data.line.p2.y;
      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }
      break;
    case OSNAP_MID:
      if(obj_select_line(line,mx,my,sens))
	{
	  *x=(line->data.line.p1.x+line->data.line.p2.x)/2;
	  *y=(line->data.line.p1.y+line->data.line.p2.y)/2;
	  return TRUE;
	}      
      break;
    case OSNAP_NEA:
      if(obj_select_line(line,mx,my,sens))
	{
	  misc_project_point_on_line(line->data.line.p1.x,line->data.line.p1.y, 
				     line->data.line.p2.x,line->data.line.p2.y, 
				     mx, my, x, y);
	  return TRUE;
	}
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_cline(gint osnap, OBJECT *cline, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  gdouble Ax, Ay, Bx, By;
  
  Ax=cline->data.line.p1.x;
  Ay=cline->data.line.p1.y;
  Bx=cline->data.line.p2.x;
  By=cline->data.line.p2.y;

  calc_endpoint_for_cline(cline->layer->doc, Ax, Ay, Bx, By, &Ax, &Ay, &Bx, &By);

  switch(osnap)
    {
    case OSNAP_NEA:
      if(obj_select_cline(cline,mx,my,sens))
	{
	  misc_project_point_on_line(Ax, Ay, Bx, By, mx, my, x, y);
	  return TRUE;
	}
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_rline(gint osnap, OBJECT *rline, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  gdouble Ax, Ay, Bx, By, tx, ty;
  
  Ax=rline->data.line.p1.x;
  Ay=rline->data.line.p1.y;
  Bx=rline->data.line.p2.x;
  By=rline->data.line.p2.y;

  calc_endpoint_for_ray(rline->layer->doc,Ax,Ay,Bx,By,&Bx,&By);

  switch(osnap)
    {
    case OSNAP_END:
      tx=rline->data.line.p1.x;
      ty=rline->data.line.p1.y;
      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }
      break;
    case OSNAP_NEA:
      if(obj_select_rline(rline,mx,my,sens))
	{
	  misc_project_point_on_line(Ax, Ay, Bx, By, mx, my, x, y);
	  return TRUE;
	}
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_circle(gint osnap, OBJECT *circle, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  POINT p1,p2,p3;
  gdouble cx,cy,cr;
  gint i;

  switch(osnap)
    {
    case OSNAP_CEN:
      if(obj_select_circle(circle,mx,my,sens))
	{
	  *x=circle->data.circle.centrum.x;
	  *y=circle->data.circle.centrum.y;
	  return TRUE;
	}      
      break;
    case OSNAP_NEA:
      if(obj_select_circle(circle,mx,my,sens))
	{	  
	  gdouble angle;

	  angle=get_angle(circle->data.circle.centrum.x, circle->data.circle.centrum.y, mx, my)*M_PI/180.0;
	  
	  *x=circle->data.circle.centrum.x+circle->data.circle.radius*cos(angle); 
	  *y=circle->data.circle.centrum.y-circle->data.circle.radius*sin(angle); 
	  return TRUE;
	}      
      break;
    case OSNAP_QUA:
      cr=circle->data.circle.radius;
      
      for(i=0;i<360;i+=90)
	{
	  cx=circle->data.circle.centrum.x+cr*cos(i*M_PI/180.0);
	  cy=circle->data.circle.centrum.y-cr*sin(i*M_PI/180.0);
	  
	  if( (ABS(mx-cx)<sens) &&
	      (ABS(my-cy)<sens))
	    {
	      *x=cx;
	      *y=cy;
	      return TRUE;
	    }
	}
      break;
    case OSNAP_TAN:
      if(osnap_tan_possible() && obj_select_circle(circle,mx,my,sens))
	{
	  p1.x=gl_oldmx;
	  p1.y=gl_oldmy;

	  if(math_tangent_CP_2D( &circle->data.circle, &p1, &p2, &p3))
	     {
	       /* Since we always find two points we should take the one closest
		  to the mouse cursor */
	       if(dist(p2.x, p2.y, mx, my)<dist(p3.x, p3.y, mx, my))
		 {
		   *x=p2.x;
		   *y=p2.y;
		 }
	       else
		 {
		   *x=p3.x;
		   *y=p3.y;
		 }

	       return TRUE;
	     }
	}      
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_arc(gint osnap, OBJECT *arc, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  POINT p1,p2,p3;
  gdouble tx,ty;
  gint i;

  switch(osnap)
    {
    case OSNAP_CEN:
      if(obj_select_arc(arc,mx,my,sens))
	{
	  *x=arc->data.arc.centrum.x;
	  *y=arc->data.arc.centrum.y;
	  return TRUE;
	}      
      break;
    case OSNAP_END:
      tx=arc->data.arc.centrum.x+arc->data.arc.radius*cos(arc->data.arc.start_angle*M_PI/180.0);
      ty=arc->data.arc.centrum.y-arc->data.arc.radius*sin(arc->data.arc.start_angle*M_PI/180.0);
      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }
      tx=arc->data.arc.centrum.x+arc->data.arc.radius*cos(arc->data.arc.end_angle*M_PI/180.0);
      ty=arc->data.arc.centrum.y-arc->data.arc.radius*sin(arc->data.arc.end_angle*M_PI/180.0);
      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }
      break;
    case OSNAP_MID:
      if(obj_select_arc(arc,mx,my,sens))
	{
	  if(arc->data.arc.end_angle>arc->data.arc.start_angle)
	    {
	      *x=arc->data.arc.centrum.x+arc->data.arc.radius*
		cos((arc->data.arc.start_angle+arc->data.arc.end_angle)/2*M_PI/180.0);
	      *y=arc->data.arc.centrum.y-arc->data.arc.radius*
		sin((arc->data.arc.start_angle+arc->data.arc.end_angle)/2*M_PI/180.0);
	    }
	else
	  {
	    gdouble angle;

	    angle=(360-arc->data.arc.start_angle+arc->data.arc.end_angle)/2+arc->data.arc.start_angle;
	    
	    while(angle<0) angle+=360;
	    while(angle>360) angle-=360;
	    
	    *x=arc->data.arc.centrum.x+arc->data.arc.radius*cos(angle*M_PI/180.0);
	    *y=arc->data.arc.centrum.y-arc->data.arc.radius*sin(angle*M_PI/180.0);
	  }
	  return TRUE;
	}      
      break;
    case OSNAP_NEA:
      if(obj_select_arc(arc,mx,my,sens))
	{
	  gdouble angle;

	  angle=get_angle(arc->data.arc.centrum.x, arc->data.arc.centrum.y, mx, my)*M_PI/180.0;
	  
	  /* If angle is not between the start and end angle of the arc, due to selection sensitity,
	     select the closest angle */
	  if(!misc_within_angle(arc->data.arc.start_angle, arc->data.arc.end_angle, angle))
	    {
	      if(ABS(arc->data.arc.start_angle-angle)<ABS(arc->data.arc.end_angle-angle))
		angle=arc->data.arc.start_angle;
	      else
		angle=arc->data.arc.end_angle;
	    }

	  *x=arc->data.arc.centrum.x+arc->data.arc.radius*cos(angle); 
	  *y=arc->data.arc.centrum.y-arc->data.arc.radius*sin(angle); 
	  return TRUE;
	}      
      break;
    case OSNAP_QUA:
      for(i=0;i<360;i+=90)
	{
	  tx=arc->data.arc.centrum.x+arc->data.arc.radius*cos(i*M_PI/180.0);
	  ty=arc->data.arc.centrum.y-arc->data.arc.radius*sin(i*M_PI/180.0);
	  
	  if( (ABS(mx-tx)<sens) && (ABS(my-ty)<sens) &&
	      misc_within_angle(arc->data.arc.start_angle,arc->data.arc.end_angle, 
				get_angle(arc->data.arc.centrum.x,arc->data.arc.centrum.y,mx,my)))
	    {
	      *x=tx;
	      *y=ty;
	      return TRUE;
	    }
	}
      break;
    case OSNAP_TAN:
      if(osnap_tan_possible() && obj_select_arc(arc,mx,my,sens))
	{      
	  OBJECT_CIRCLE cir;
	  
	  p1.x=gl_oldmx;
	  p1.y=gl_oldmy;

	  cir.centrum.x=arc->data.arc.centrum.x;
	  cir.centrum.y=arc->data.arc.centrum.y;
	  cir.radius=arc->data.arc.radius;

	  if(math_tangent_CP_2D( &cir, &p1, &p2, &p3))
	    {
	      /* Since we always find two points we should take the one closest
		 to the mouse cursor */
	       if(dist(p2.x, p2.y, mx, my)<dist(p3.x, p3.y, mx, my))
		 {
		   *x=p2.x;
		   *y=p2.y;
		 }
	       else
		 {
		   *x=p3.x;
		   *y=p3.y;
		 }
	       
	       return TRUE;
	    }  
	}
      break;
    }
  
  return FALSE;
}

static gboolean obj_osnap_text(gint osnap, OBJECT *text, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  switch(osnap)
    {
    case OSNAP_INS:
      if(obj_select_text(text,mx,my,sens))
	{
	  *x=text->data.text.insert.x;
	  *y=text->data.text.insert.y;
	  return TRUE;
	}      
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_attribute(gint osnap, OBJECT *attr, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  switch(osnap)
    {
    case OSNAP_INS:
      if(obj_select_attribute(attr,mx,my,sens))
	{
	  *x=attr->data.attribute.text.insert.x;
	  *y=attr->data.attribute.text.insert.y;
	  return TRUE;
	}      
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_point(gint osnap, OBJECT *point, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  switch(osnap)
    {
    case OSNAP_NEA:
    case OSNAP_NOD:
      if(obj_select_point(point,mx,my,sens))
	{
	  *x=point->data.point.centrum.x;
	  *y=point->data.point.centrum.y;
	  return TRUE;
	}      
      break;
    }
  return FALSE;
}

static gboolean obj_osnap_pline(gint osnap, OBJECT *pline, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  gdouble tx,ty, angle, tr;
  gint type, i;
  OBJECT_UNION s;

  switch(osnap)
    {
    case OSNAP_END:
      if(obj_select_polyline_object(pline, mx, my, gl_select_sensitivity/pline->layer->doc->zoom, &type, &s))
	{
	  if(type==NODE_ARC)
	    {
	      tx=s.arc.centrum.x+s.arc.radius*cos(s.arc.start_angle*M_PI/180.0);
	      ty=s.arc.centrum.y-s.arc.radius*sin(s.arc.start_angle*M_PI/180.0);
	      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
		*x=tx;
		*y=ty;
		return TRUE;
	      }
	      tx=s.arc.centrum.x+s.arc.radius*cos(s.arc.end_angle*M_PI/180.0);
	      ty=s.arc.centrum.y-s.arc.radius*sin(s.arc.end_angle*M_PI/180.0);
	      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
		*x=tx;
		*y=ty;
		return TRUE;
	      }	
	    }
	  else if(type==NODE_LINE)
	    {
	      tx=s.line.p1.x;
	      ty=s.line.p1.y;
	      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
		*x=tx;
		*y=ty;
		return TRUE;
	      }
	      tx=s.line.p2.x;
	      ty=s.line.p2.y;
	      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
		*x=tx;
		*y=ty;
		return TRUE;
	      }
	    }
	}
      break; 
    case OSNAP_CEN:
      if(obj_select_polyline_object(pline, mx, my, gl_select_sensitivity/pline->layer->doc->zoom, &type, &s))
	{
	  if(type==NODE_ARC)
	    {
	      *x=s.arc.centrum.x;
	      *y=s.arc.centrum.y;
	      return TRUE;
	    }	 
	}      
      break;
    case OSNAP_MID:
      if(obj_select_polyline_object(pline, mx, my, gl_select_sensitivity/pline->layer->doc->zoom, &type, &s))
	{
	  if(type==NODE_ARC)
	    {
	      if(s.arc.end_angle>s.arc.start_angle)
		{
		  *x=s.arc.centrum.x+s.arc.radius*
		    cos((s.arc.start_angle+s.arc.end_angle)/2*M_PI/180.0);
		  *y=s.arc.centrum.y-s.arc.radius*
		    sin((s.arc.start_angle+s.arc.end_angle)/2*M_PI/180.0);
		}
	      else
		{
		  angle=(360-s.arc.start_angle+s.arc.end_angle)/2+s.arc.start_angle;
		  
		  while(angle<0) angle+=360;
		  while(angle>360) angle-=360;
		  
		  *x=s.arc.centrum.x+s.arc.radius*cos(angle*M_PI/180.0);
		  *y=s.arc.centrum.y-s.arc.radius*sin(angle*M_PI/180.0);
		}
	      return TRUE;
	    }
	  else if(type==NODE_LINE)
	    {
	      *x=(s.line.p1.x+s.line.p2.x)/2;
	      *y=(s.line.p1.y+s.line.p2.y)/2;
	      return TRUE;
	    }
	}
      break;
    case OSNAP_NEA:
      if(obj_select_polyline(pline, mx, my, sens))
	{
	  if(obj_select_polyline_object(pline, mx, my, gl_select_sensitivity/pline->layer->doc->zoom, &type, &s))
	    {
	      if(type==NODE_ARC)
		{
		  angle=get_angle(s.arc.centrum.x, s.arc.centrum.y, mx, my)*M_PI/180.0;
		  
		  /* If angle is not between the start and end angle of the arc, due to selection sensitity,
		     select the closest angle */
		  if(!misc_within_angle(s.arc.start_angle, s.arc.end_angle, angle))
		    {
		      if(ABS(s.arc.start_angle-angle)<ABS(s.arc.end_angle-angle))
			angle=s.arc.start_angle;
		      else
			angle=s.arc.end_angle;
		    }
		  
		  *x=s.arc.centrum.x+s.arc.radius*cos(angle); 
		  *y=s.arc.centrum.y-s.arc.radius*sin(angle); 
		  return TRUE;
		}
	      else if(type==NODE_LINE)
		{
		  misc_project_point_on_line(s.line.p1.x, s.line.p1.y, 
					     s.line.p2.x, s.line.p2.y, 
					     mx, my, x, y);
		  return TRUE;
		}
	    }
	}
      break;
    case OSNAP_QUA:
      if(obj_select_polyline_object(pline, mx, my, gl_select_sensitivity/pline->layer->doc->zoom, &type, &s))
	{
	  if(type==NODE_ARC)
	    {
	      tr=s.arc.radius;
	      
	      for(i=0;i<360;i+=90)
		{
		  tx=s.arc.centrum.x+tr*cos(i*M_PI/180.0);
		  ty=s.arc.centrum.y-tr*sin(i*M_PI/180.0);
		  
		  if( (ABS(mx-tx)<sens) && (ABS(my-ty)<sens) &&
		      misc_within_angle(s.arc.start_angle, s.arc.end_angle, 
					get_angle(s.arc.centrum.x, s.arc.centrum.y,mx,my)))
		    {
		      *x=tx;
		      *y=ty;
		      return TRUE;
		    }
		}	       
	    }
	}
      break;
    }

  return FALSE;
}

static gboolean obj_osnap_spline(gint osnap, OBJECT *spline, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  gdouble tx,ty;
  GList *list;
  POINT *plist;
  gint i,num;
  gdouble Ax, Ay, Bx, By;
  
  switch(osnap)
    {
    case OSNAP_END:
      if(!spline->data.spline.closed)
	{
	  /* Check start node */
	  list=g_list_first(spline->data.spline.node_list);	  
	  tx=((NODE *)list->data)->x;
	  ty=((NODE *)list->data)->y;
	  if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	    *x=tx;
	    *y=ty;
	    return TRUE;
	  }
	  /* Check end node */
	  list=g_list_last(spline->data.spline.node_list);	  
	  tx=((NODE *)list->data)->x;
	  ty=((NODE *)list->data)->y;
	  if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	    *x=tx;
	    *y=ty;
	    return TRUE;
	  }  
	}
      break;
    case OSNAP_CEN:
      if(spline->data.spline.closed && obj_select_spline(spline, mx, my, sens))
	{
	  POINT p;
	  misc_calc_centroid(spline->data.spline.node_list, &p);
	  *x=p.x;
	  *y=p.y;
	  return TRUE;
	}      
      break;
    case OSNAP_NEA:
      /* If no cache list exists create one */
      if(spline->data.spline.cache_point==NULL)
	calc_bspline(spline);
      
      num=spline->data.spline.num;
      
      plist=spline->data.spline.cache_point;
      
      i=1;
      
      while(i<num)
	{
	  Ax=plist[i-1].x;
	  Ay=plist[i-1].y;
	  Bx=plist[i].x;
	  By=plist[i].y;
	  
	  if(misc_point_on_line(mx, my, Ax, Ay, Bx, By, gl_select_sensitivity/spline->layer->doc->zoom))
	    {
	      misc_project_point_on_line(Ax, Ay, Bx, By, mx, my, x, y);
	      return TRUE;
	    }
	  
	  i++;
	}
      break;
    }
  
  return FALSE;
}

static gboolean obj_osnap_ellipse(gint osnap, OBJECT *ellipse, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  POINT p1, p2, p3;

  switch(osnap)
    {
    case OSNAP_CEN:
      if(obj_select_ellipse(ellipse, mx, my, sens))
	{
	  *x=ellipse->data.ellipse.centrum.x;
	  *y=ellipse->data.ellipse.centrum.y;
	  return TRUE;
	} 
      break;
    case OSNAP_TAN:
      if(osnap_tan_possible() && obj_select_ellipse(ellipse,mx,my,sens))
	{
	  p1.x=gl_oldmx;
	  p1.y=gl_oldmy;

	  if(math_tangent_EP_2D( &ellipse->data.ellipse, &p1, &p2, &p3))
	     {
	       /* Since we always find two points we should take the one closest
		  to the mouse cursor */
	       if(dist(p2.x, p2.y, mx, my)<dist(p3.x, p3.y, mx, my))
		 {
		   *x=p2.x;
		   *y=p2.y;
		 }
	       else
		 {
		   *x=p3.x;
		   *y=p3.y;
		 }

	       return TRUE;
	     }
	}      
      break;
    }

  return FALSE;
}


static gboolean obj_osnap_elarc(gint osnap, OBJECT *elarc, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  gdouble angle;
  gdouble tx,ty;

  switch(osnap)
    {
    case OSNAP_CEN:
      if(obj_select_elarc(elarc, mx, my, sens))
	{
	  *x=elarc->data.elarc.centrum.x;
	  *y=elarc->data.elarc.centrum.y;
	  return TRUE;
	}      
      break;
    case OSNAP_MID:
      if(obj_select_elarc(elarc,mx,my,sens))
	{
	  if(elarc->data.elarc.end_angle>elarc->data.elarc.start_angle)
	    angle=(elarc->data.elarc.start_angle+elarc->data.elarc.end_angle)/2;
	  else
	    {
	      angle=(360-elarc->data.elarc.start_angle+elarc->data.elarc.end_angle)/2+elarc->data.elarc.start_angle;
	    
	      while(angle<0) angle+=360;
	      while(angle>360) angle-=360;
	    }
	  
	  *x=elarc->data.elarc.centrum.x+elarc->data.elarc.major*cos(angle*M_PI/180.0);
	  *y=elarc->data.elarc.centrum.y-elarc->data.elarc.minor*sin(angle*M_PI/180.0);
	  
	  misc_rotate_point(-elarc->data.elarc.angle*M_PI/180.0,elarc->data.elarc.centrum.x,elarc->data.elarc.centrum.y,
			    x,y);
	  return TRUE;
	}      
      break;
    case OSNAP_END:
      /* Check start angle */
      tx=elarc->data.elarc.major*cos(2*M_PI-elarc->data.elarc.start_angle*M_PI/180.0)+elarc->data.elarc.centrum.x;
      ty=elarc->data.elarc.minor*sin(2*M_PI-elarc->data.elarc.start_angle*M_PI/180.0)+elarc->data.elarc.centrum.y;
      
      misc_rotate_point(-elarc->data.elarc.angle*M_PI/180, 
			elarc->data.elarc.centrum.x, elarc->data.elarc.centrum.y, 
			&tx, &ty);

      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }  
      /* Check End angle */
      tx=elarc->data.elarc.major*cos(2*M_PI-elarc->data.elarc.end_angle*M_PI/180.0)+elarc->data.elarc.centrum.x;
      ty=elarc->data.elarc.minor*sin(2*M_PI-elarc->data.elarc.end_angle*M_PI/180.0)+elarc->data.elarc.centrum.y;
      
      misc_rotate_point(-elarc->data.elarc.angle*M_PI/180, 
			elarc->data.elarc.centrum.x, elarc->data.elarc.centrum.y, 
			&tx, &ty);
      
      if( (tx-sens<mx) && (tx+sens>mx) && (ty-sens<my) && (ty+sens>my)){
	*x=tx;
	*y=ty;
	return TRUE;
      }  
      break;
    }
  return FALSE;
}

static gboolean obj_osnap_block(gint osnap, OBJECT *insert, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  LAYER *layer;
  TRANSFORM insert_transform;
  OBJECT object; 
  BLOCK_OBJECT *block_object;
  GList *list;
  GList *attr_list;

  if(osnap==OSNAP_INS && obj_select_insert(insert, mx, my, sens))
    {
      *x=insert->data.block.insert_point.x;
      *y=insert->data.block.insert_point.y;
      return TRUE;
    }

  /* Check each object within block */

  /* Make tranform matrix */
  insert_transform.type=OBJ_TRANSLATE;
  insert_transform.trans_x=insert->data.block.insert_point.x;
  insert_transform.trans_y=insert->data.block.insert_point.y;
  insert_transform.cen_x=0;
  insert_transform.cen_y=0;

  /* Check if need to scale block */
  if(ABS(insert->data.block.scale_x-1)>GFLOAT_SENSITIVE ||
     ABS(insert->data.block.scale_y-1)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_SCALE;
      insert_transform.scale_x=insert->data.block.scale_x;
      insert_transform.scale_y=insert->data.block.scale_y;
    }

  /* Check if need to rotate block */
  if(ABS(insert->data.block.rotation)>GFLOAT_SENSITIVE)
    {
      insert_transform.type|=OBJ_ROTATE;
      insert_transform.rot_angle=-insert->data.block.rotation*M_PI/180.0;
    }

  /* Get list of attributes in this block */
  attr_list=g_list_first(insert->data.block.attr);

  /* Get first object in block */
  list=g_list_first(insert->data.block.block->object_list);

  while(list!=NULL)
    {
      block_object=(BLOCK_OBJECT *)list->data;

      /* Check if we should use the layer of the block */
      if(!strcmp(block_object->name_of_layer,"0"))
	layer=insert->layer;
      else
	{
	  layer=layer_find(insert->layer->doc, block_object->name_of_layer);
	  if(layer==NULL)
	    {
	      g_warning( _("Layer %s not found. Creating it."), block_object->name_of_layer );
	      layer=layer_new(insert->layer->doc, -1, insert->layer->doc->current_linestyle, 0.0 , 0.0, 0.0, 0.0,  
			  block_object->name_of_layer, TRUE, TRUE);
	    }
	}

      /* Make nice object to draw */
      object.type=block_object->type;	 
      object.layer=layer;
      
      /* Make linewidth */
      object.type_linewidth=block_object->type_linewidth;
      switch(object.type_linewidth)
	{
	case BYLAYER:
	  object.linewidth=layer->linewidth;
	  break;
	case BYBLOCK:
	  object.linewidth=insert->linewidth;
	  break;
	case USERDEFINED:
	  object.linewidth=block_object->linewidth;
	  break;
	}

      memcpy(&object.data, &block_object->data, sizeof(OBJECT_UNION));
      make_object_selected(&object, OBJECT_SEL, FALSE);
	  
      switch(object.type)
	{
	case TYPE_LINE:
	  obj_transform_line(&object, &insert_transform);
	  if(obj_osnap_line(osnap, &object, mx, my, x, y))
	     return TRUE;
	  break;
	case TYPE_CIRCLE:
	  obj_transform_circle(&object, &insert_transform);
	  if(obj_osnap_circle(osnap, &object, mx, my, x, y))
	     return TRUE;
	  break;
	case TYPE_ARC:
	  obj_transform_arc(&object, &insert_transform);
	  if(obj_osnap_arc(osnap, &object, mx, my, x, y))
	     return TRUE;
	  break;
	case TYPE_TEXT:
	  obj_transform_text(&object, &insert_transform);
	  if(obj_osnap_text(osnap, &object, mx, my, x, y))
	     return TRUE;
	  break;
	case TYPE_POINT:
	  obj_transform_point(&object, &insert_transform);
	  if(obj_osnap_point(osnap, &object, mx, my, x, y))
	     return TRUE;
	  break;
	case TYPE_ATTRIBUTE:
	  obj_transform_attribute(&object, &insert_transform);
	  /* Fix text of attribute */
	  object.data.attribute.text.text=(gchar *)attr_list->data;
	  attr_list=g_list_next(attr_list);
	  if(obj_osnap_text(osnap, &object, mx, my, x, y))
	     return TRUE;
	  break;
	}
      
      list=g_list_next(list);
    }

  return FALSE;
}

static gboolean test_layer(DOCUMENT *doc, LAYER *layer, gdouble mx, gdouble my,gdouble *x, gdouble *y)
{
  gint osnap;
  gint osnap_to_test;
  GList *list;
  OBJECT *obj;

  /* Layer must be visible */
  if(!layer->visible)
    return FALSE;

  /* Extract each object in layer and test it */
  list=g_list_first(layer->object_list);

  while(list!=NULL)
    {
      obj=(OBJECT *)list->data;

      osnap=gl_osnap;

      /* We can have many osnap active at once. Therefor test them one by one */
      while(osnap!=0)
	{
	  if(osnap & OSNAP_CEN)
	    osnap_to_test=OSNAP_CEN;
	  else if(osnap & OSNAP_END)
	    osnap_to_test=OSNAP_END;
	  else if(osnap & OSNAP_INS)
	    osnap_to_test=OSNAP_INS;
	  else if(osnap & OSNAP_INT)
	    osnap_to_test=OSNAP_INT;
	  else if(osnap & OSNAP_MID)
	    osnap_to_test=OSNAP_MID;
	  else if(osnap & OSNAP_NEA)
	    osnap_to_test=OSNAP_NEA;
	  else if(osnap & OSNAP_NOD)
	    osnap_to_test=OSNAP_NOD;
	  else if(osnap & OSNAP_QUA)
	    osnap_to_test=OSNAP_QUA;
	  else if(osnap & OSNAP_TAN)
	    osnap_to_test=OSNAP_TAN;
	  else
	    break;

	  switch(obj->type)
	    {
	    case TYPE_LINE:
	      if(obj_osnap_line(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_CLINE:
	      if(obj_osnap_cline(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_RLINE:
	      if(obj_osnap_rline(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_CIRCLE:
	      if(obj_osnap_circle(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_ARC:
	      if(obj_osnap_arc(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_TEXT:
	      if(obj_osnap_text(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_ATTRIBUTE:
	      if(obj_osnap_attribute(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_POINT:
	      if(obj_osnap_point(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_ELLIPSE:
	      if(obj_osnap_ellipse(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_ELARC:
	      if(obj_osnap_elarc(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_POLYLINE:
	      if(obj_osnap_pline(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_SPLINE:
	      if(obj_osnap_spline(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    case TYPE_BLOCK:
	      if(obj_osnap_block(osnap_to_test, obj, mx, my, x, y))
		return TRUE;
	      break;
	    }

	  /* Since this osnap is tested remove it */
	  osnap-=osnap_to_test;
	}

      list=g_list_next(list);
    }
  
  return FALSE;
}


gboolean osnap(DOCUMENT *doc, gdouble mx, gdouble my, gdouble *x, gdouble *y)
{
  GList *list;

  /* Scale sensitivity to world scale */
  sens=gl_osnap_sensitivity/doc->zoom;

  /* First test current layer */
  if(gl_osnap_apply_to!=APPLY_ALL_LAYERS_BUT_CURRENT)
    {
      if(test_layer(doc, doc->current_layer, mx,my,x,y))
	return TRUE;
    }

  /* Dont check the following layers */
  if(gl_osnap_apply_to==APPLY_CURRENT_LAYER)
    return FALSE;

  /* Then all the others */
  list=g_list_first(doc->layer_list);

  while(list!=NULL)
    {
      if((LAYER *)list->data!=doc->current_layer)   /* We have already checked current layer, no need to do it again */
	if(test_layer(doc, (LAYER *)list->data, mx,my,x,y))
	  return TRUE;
      list=g_list_next(list);
    }

  return FALSE;
}

void set_snap_text(gint type)
{
  switch(type)
    {
    case OSNAP_CEN:
      add_prompt_text( _("Centerpoint of "));
      break;
    case OSNAP_END:
      add_prompt_text( _("Endpoint of "));
      break;
    case OSNAP_INS:
      add_prompt_text( _("Insertpoint of "));
      break;
    case OSNAP_INT:
      add_prompt_text( _("Intersectionpoint of "));
      break;
    case OSNAP_MID:
      add_prompt_text( _("Midpoint of "));
      break;
    case OSNAP_NEA:
      add_prompt_text( _("Nearestpoint of "));
      break;
    case OSNAP_NOD:
      add_prompt_text( _("Nodepoint of "));
      break;
    case OSNAP_PER:
      add_prompt_text( _("Perpendicularpoint of "));
      break;
    case OSNAP_QUA:
      add_prompt_text( _("Quadrantpoint of "));
      break;
    case OSNAP_TAN:
      add_prompt_text( _("Tangentpoint of "));
      break;
    case OSNAP_NONE:
      break;
    default:
      add_prompt_text( _("Snappoint of "));
      break;
  }
}

void set_snap(GtkWidget *widget, gpointer type)
{
  gint t;

  t=GPOINTER_TO_INT(type);

  if(!command_from_keyboard)
    {
      switch(t)
	{
	case OSNAP_CEN:
	  print( _("cen"));
	  break;
	case OSNAP_END:
	  print( _("end"));
	  break;
	case OSNAP_INS:
	  print( _("ins"));
	  break;
	case OSNAP_INT:
	  print( _("int"));
	  break;
	case OSNAP_MID:
	  print( _("mid"));
	  break;
	case OSNAP_NEA:
	  print( _("nea"));
	  break;
	case OSNAP_NOD:
	  print( _("nod"));
	  break;
	case OSNAP_PER:
	  print( _("per"));
	  break;
	case OSNAP_QUA:
	  print( _("qua"));
	  break;
	case OSNAP_TAN:
	  print( _("tan"));
	  break;
	case OSNAP_NONE:
	  print( _("non"));
	  break;
	}
    }

  /* Set osnap mode */
  gl_osnap=t;

  set_prompt();
  set_cursor(NULL);
  toolbox_set_osnap(gl_osnap);
}


gboolean osnap_button(DOCUMENT *doc, gint button, gdouble *x, gdouble *y)
{
  switch(button)
    {
    case 1:   /* Left button */
      if(!osnap(doc, *x,*y, x,y))
	{
	  print( _("\n  No valid snappoint found."));
	  set_prompt();
	  //	  set_snap_text(gl_osnap);
	}
      else
	{
	  if(gl_osnap_is==OSNAP_ONCE)
	    gl_osnap=OSNAP_NONE;
	  
	  remove_old();
	  set_cursor(doc);
	  toolbox_set_osnap(gl_osnap);
	  return TRUE;
	}
      break;
    case 3: /* Right button */
      print("*Cancel*");
      gl_osnap=OSNAP_NONE;
      set_prompt();
      set_cursor(doc);
      toolbox_set_osnap(gl_osnap);
      break;
    }

  return FALSE;
}



