/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <string.h>
#include <math.h>
#include "struct.h"
#include "struct/grdraw.h"
#include "global.h"
#include "initmenu.h"
#include "calc.h"
#include "dia_comm.h"
#include "dia_coords.h"
#include "xor.h"
#include "misc.h"
#include "parse.h"
#include "redraw.h"
#include "obj_line.h"
#include "obj_circle.h"
#include "obj_arc.h"
#include "obj_attribute.h"
#include "obj_polyline.h"
#include "obj_point.h"
#include "obj_text.h"
#include "obj_cline.h"
#include "obj_rline.h"
#include "obj_ellipse.h"
#include "obj_elarc.h"
#include "obj_insert.h"
#include "obj_spline.h"
#include "dim_select.h"
#include "tool_erase.h"
#include "tool_list.h"
#include "dia_toolbox.h"
#include "edit_prop.h"
#include "translation.h"

/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static gint o_gl_tool, o_gl_stage;
static gdouble o_gl_oldmx, o_gl_oldmy;
static gint add_object;

static gchar select_prompt[50];

static gint number_of_objects;

static gdouble sens;

static gboolean select_first_object; /* This holds if this is the first object user select 
					for this selection session. It is used to know when
					to free the selection list. */

/* --  P R O T O T Y P E S  -- */
static OBJECT *select_object(DOCUMENT *doc, gdouble x, gdouble y, gboolean check);

/* --  F U N C T I O N S  -- */
static void fix_window(gdouble *x1,gdouble *y1, gdouble *x2, gdouble *y2)
     /* Fix a window so that it returns a correct window
	with a left-top coordinate and a right-down coordinate */
{
  gdouble tx,ty;

  tx=*x1;
  ty=*y1;

  *x1=MIN( (*x1), (*x2) );
  *y1=MIN( (*y1), (*y2) );
  *x2=MAX( (tx), (*x2) );
  *y2=MAX( (ty), (*y2) );
}

void make_object_selected(OBJECT *object, gint sel, gboolean fix_select_list)
     /* Selects/deselects a object and all
	its nodes. Also updates the
	selection list if boolean fix_select_list is true. */
{
  gint node;
  GList *list;

  node=(sel==OBJECT_SEL);

  switch(object->type)
    {
    case TYPE_LINE:
    case TYPE_RLINE:
    case TYPE_CLINE:
      object->data.line.p1.selected=node;
      object->data.line.p2.selected=node;
      break;
    case TYPE_CIRCLE:
      object->data.circle.centrum.selected=node;
      break;
    case TYPE_POINT:
      object->data.point.centrum.selected=node;
      break;
    case TYPE_ELLIPSE:
      object->data.circle.centrum.selected=node;
      break;
    case TYPE_TEXT:
      object->data.text.insert.selected=node;
      break;
    case TYPE_ATTRIBUTE:
      object->data.attribute.text.insert.selected=node;
      break;
    case TYPE_ARC:
      object->data.arc.selected_startpoint=node;
      object->data.arc.selected_endpoint=node;
      break;
    case TYPE_ELARC:
      object->data.elarc.selected_startpoint=node;
      object->data.elarc.selected_endpoint=node;
      break;
    case TYPE_POLYLINE:
      list=object->data.polyline.node_list;

      while(list!=NULL)
	{
	  ((NODE *)list->data)->selected=node;
	  list=g_list_next(list);
	}

      break;
    case TYPE_SPLINE:
      list=object->data.spline.node_list;

      while(list!=NULL)
	{
	  ((NODE *)list->data)->selected=node;
	  list=g_list_next(list);
	}
      break;
    case TYPE_BLOCK:
      object->data.block.insert_point.selected=node;
      break;
    }

  object->selected=sel;

  if(fix_select_list)
    {
      /* Fix selection list */
      if(node)
	object->layer->doc->selection_list=g_list_append(object->layer->doc->selection_list, (gpointer)object);
      else
	object->layer->doc->selection_list=g_list_remove(object->layer->doc->selection_list, (gpointer)object);
    }
}

void select_set_radiobuttons(gboolean value)
{
  if(value)
    {
      menus_set_sensitive("/Assist/Select/Add", TRUE);
      menus_set_sensitive("/Assist/Select/Remove", FALSE);
      toolbox_set_button("Select Add", 0);
    }
  else
    {
      menus_set_sensitive("/Assist/Select/Add", FALSE);
      menus_set_sensitive("/Assist/Select/Remove", TRUE);
      toolbox_set_button("Select Sub", 0);
    }
}

void select_submenu(gboolean status)
{
  menus_set_sensitive("/Assist/Select/Add", status);
  menus_set_sensitive("/Assist/Select/Remove", status);
  menus_set_sensitive("/Assist/Select/All", status);
  menus_set_sensitive("/Assist/Select/Invert", status);
  menus_set_sensitive("/Assist/Select/Point", status);
  menus_set_sensitive("/Assist/Select/Window", status);
  menus_set_sensitive("/Assist/Select/Crossing", status);
}

void deselect_all_objects(DOCUMENT *doc)
     /* Function to deselect all selected objects
	in document doc */
{
  OBJECT *object;
  GList *list;

  select_first_object=FALSE;

  if(doc==NULL || doc->selection_list==NULL)
    return;

  /* Goes through the selection list and 
     deselect all objects */
  list=g_list_first(doc->selection_list);
    
  while(list!=NULL)
    {
      object=(OBJECT *)list->data;

      make_object_selected(object, OBJECT_NOT_SEL, FALSE);
      redraw_object(doc, object);

      list=g_list_next(list);
    }

  if(doc->selection_list!=NULL)
    g_list_free(doc->selection_list);
  doc->selection_list=NULL;
}

static OBJECT *test_layer(LAYER *layer ,gdouble x, gdouble y, gboolean check)
{
  GList *object_list;
  OBJECT *object;

  /* Is this layer visible and editable ? */
  if( (!layer->visible) || (!layer->editable) )
    return NULL;

  object_list=g_list_first(layer->object_list);

  while(object_list!=NULL)
    {
      object=(OBJECT *)object_list->data;
	 
      if( !check || ((object->selected==OBJECT_NOT_SEL && add_object) || (object->selected==OBJECT_SEL && !add_object)))
	{
	  if(object->type==TYPE_LINE && (gl_select_mask & TYPE_LINE) && obj_select_line(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_POLYLINE && (gl_select_mask & TYPE_POLYLINE) && obj_select_polyline(object, x, y, sens))
	    return object;
	  
	  if(object->type==TYPE_CIRCLE && (gl_select_mask & TYPE_CIRCLE) && obj_select_circle(object, x, y, sens))
	    return object;
	  
	  if(object->type==TYPE_ARC && (gl_select_mask & TYPE_ARC) && obj_select_arc(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_POINT && (gl_select_mask & TYPE_POINT) && obj_select_point(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_TEXT && (gl_select_mask & TYPE_TEXT) && obj_select_text(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_ATTRIBUTE && (gl_select_mask & TYPE_ATTRIBUTE) && obj_select_attribute(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_RLINE && (gl_select_mask & TYPE_RLINE) && obj_select_rline(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_CLINE && (gl_select_mask & TYPE_CLINE) && obj_select_cline(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_SPLINE && (gl_select_mask & TYPE_SPLINE) && obj_select_spline(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_ELLIPSE && (gl_select_mask & TYPE_ELLIPSE) && obj_select_ellipse(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_BLOCK && (gl_select_mask & TYPE_BLOCK) && obj_select_insert(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_ELARC && (gl_select_mask & TYPE_ELARC) && obj_select_elarc(object, x, y, sens))
	    return object;

	  if(object->type==TYPE_DIMENSION && (gl_select_mask & TYPE_DIMENSION) && 
	     dim_select_dimension(layer->doc, object, x, y, sens))
	     return object;
	}
      
      object_list=g_list_next(object_list);
    }

  return NULL;
}

static OBJECT *select_object(DOCUMENT *doc, gdouble x, gdouble y, gboolean check)
{
  GList *layer_list;
  OBJECT *object;

  layer_list=g_list_first(doc->layer_list);

  /* check first on current layer */
  if(gl_select_apply_to!=APPLY_ALL_LAYERS_BUT_CURRENT)
    {
      if((object=test_layer(doc->current_layer, x, y, check))!=NULL)
	return object;
    }

  /* Dont check the following layers */
  if(gl_select_apply_to==APPLY_CURRENT_LAYER)
    return NULL;

  while(layer_list!=NULL)
    {
      if(layer_list->data!=doc->current_layer)
	{
	  if((object=test_layer((LAYER *)layer_list->data, x, y, check))!=NULL)
	    return object;
	}
      layer_list=g_list_next(layer_list);
    }
  
  return NULL; /* No objects found! */
}

OBJECT *select_singel_object(DOCUMENT *doc, gdouble x, gdouble y)
{
  OBJECT *obj;

  add_object=TRUE;

  /* Scale sensitivity to world scale */
  sens=gl_select_sensitivity/doc->zoom;

  obj=select_object(doc,x,y, TRUE);

  if(obj!=NULL)
    make_object_selected(obj, OBJECT_SEL, TRUE);

  return obj;
}

OBJECT *select_singel_object_without_check(DOCUMENT *doc, gdouble x, gdouble y)
{
  OBJECT *obj;

  add_object=TRUE;

  /* Scale sensitivity to world scale */
  sens=gl_select_sensitivity/doc->zoom;

  obj=select_object(doc,x,y, FALSE);

  if(obj!=NULL)
    make_object_selected(obj, OBJECT_SEL, TRUE);

  return obj;
}

OBJECT *select_singel_object_not_sel(DOCUMENT *doc, gdouble x, gdouble y)
{
  OBJECT *obj;

  add_object=TRUE;

  /* Scale sensitivity to world scale */
  sens=gl_select_sensitivity/doc->zoom;

  obj=select_object(doc,x,y, TRUE);

  return obj;
}

static void select_all_object(DOCUMENT *doc, LAYER *layer, OBJECT *object)
{
  if(add_object && object->selected!=OBJECT_SEL)
    {
      make_object_selected(object, OBJECT_SEL, TRUE);
      redraw_object(doc,object);
      number_of_objects++;
    }
  else if(!add_object && object->selected!=OBJECT_NOT_SEL)
    {
      make_object_selected(object, OBJECT_NOT_SEL, TRUE);
      redraw_object(doc,object);
      number_of_objects++;
    }
}

void select_all(GtkWidget *widget, gint data)
{
  DOCUMENT *doc;

  number_of_objects=0;

  doc=get_document();
  if(doc==NULL)
    return;

  if(select_first_object)
    {
      if(doc->selection_list!=NULL)
	g_list_free(doc->selection_list);

      doc->selection_list=NULL;
      select_first_object=FALSE;
    }

  foreach_object(doc, select_all_object);

  if(add_object)
    print(_("\n  %d objects added."), number_of_objects);
  else
    print(_("\n  %d objects removed."), number_of_objects);

  set_prompt();
}

void select_invert_object(DOCUMENT *doc, LAYER *layer, OBJECT *object)
{
  if(object->selected==OBJECT_SEL)
    make_object_selected(object, OBJECT_NOT_SEL, TRUE);
  else if(object->selected==OBJECT_NOT_SEL)
    make_object_selected(object, OBJECT_SEL, TRUE);

  redraw_object(doc, object);
  number_of_objects++;
}

void select_invert(GtkWidget *widget, gint data)
{
  DOCUMENT *doc;

  number_of_objects=0;

  doc=get_document();
  if(doc==NULL)
    return;

  foreach_object(doc, select_invert_object);

  if(select_first_object)
    {
      if(doc->selection_list!=NULL)
	g_list_free(doc->selection_list);
      
      doc->selection_list=NULL;
      select_first_object=FALSE;
    }

  print( _("\n  %d objects inverted."), number_of_objects);

  set_prompt();
}

void select_window(GtkWidget *widget, gint data)
{
  gl_stage=1;
  set_prompt();
}

static gdouble win_x1, win_y1, win_x2, win_y2;

void calc_select_window_object(DOCUMENT *doc, LAYER *layer, OBJECT *object)
{
  gdouble tx,ty;
  gint found;

  if(add_object && object->selected==OBJECT_SEL)
    return;
  if(!add_object && object->selected==OBJECT_NOT_SEL)
    return;

  found=TRUE;

  if(object->type==TYPE_LINE && (gl_select_mask & TYPE_LINE))
    {
      if( !((object->data.line.p1.x>win_x1) &&
	    (object->data.line.p1.x<win_x2) &&
	    (object->data.line.p1.y>win_y1) &&
	    (object->data.line.p1.y<win_y2)))
	found=FALSE;  
      
      if( !((object->data.line.p2.x>win_x1) &&
	    (object->data.line.p2.x<win_x2) &&
	    (object->data.line.p2.y>win_y1) &&
	    (object->data.line.p2.y<win_y2)))
	found=FALSE;
    }

  if(object->type==TYPE_CIRCLE && (gl_select_mask & TYPE_CIRCLE))
    {
      if( !((object->data.circle.centrum.x>win_x1) &&
	    (object->data.circle.centrum.x<win_x2) &&
	    (object->data.circle.centrum.y>win_y1) &&
	    (object->data.circle.centrum.y<win_y2)))
	found=FALSE;
    }

  if(object->type==TYPE_ELLIPSE && (gl_select_mask & TYPE_ELLIPSE))
    {
      if( !((object->data.ellipse.centrum.x>win_x1) &&
	    (object->data.ellipse.centrum.x<win_x2) &&
	    (object->data.ellipse.centrum.y>win_y1) &&
	    (object->data.ellipse.centrum.y<win_y2)))
	found=FALSE;
    }

  if(object->type==TYPE_TEXT && (gl_select_mask & TYPE_TEXT))
    {
      if( !((object->data.text.insert.x>win_x1) &&
	    (object->data.text.insert.x<win_x2) &&
	    (object->data.text.insert.y>win_y1) &&
	    (object->data.text.insert.y<win_y2)))
	found=FALSE;
    }

  if(object->type==TYPE_ATTRIBUTE && (gl_select_mask & TYPE_ATTRIBUTE))
    {
      if( !((object->data.attribute.text.insert.x>win_x1) &&
	    (object->data.attribute.text.insert.x<win_x2) &&
	    (object->data.attribute.text.insert.y>win_y1) &&
	    (object->data.attribute.text.insert.y<win_y2)))
	found=FALSE;
    }


  if(object->type==TYPE_ARC && (gl_select_mask & TYPE_ARC))
    {
      tx=object->data.arc.centrum.x+object->data.arc.radius*cos(object->data.arc.start_angle*M_PI/180.0);
      ty=object->data.arc.centrum.y-object->data.arc.radius*sin(object->data.arc.start_angle*M_PI/180.0);
      if( !((tx>win_x1) && (tx<win_x2) && (ty>win_y1) && (ty<win_y2)))
	found=FALSE;  

      tx=object->data.arc.centrum.x+object->data.arc.radius*cos(object->data.arc.end_angle*M_PI/180.0);
      ty=object->data.arc.centrum.y-object->data.arc.radius*sin(object->data.arc.end_angle*M_PI/180.0);
      if( !((tx>win_x1) && (tx<win_x2) && (ty>win_y1) && (ty<win_y2)))
	found=FALSE;  
    }

  if(object->type==TYPE_POLYLINE && (gl_select_mask & TYPE_POLYLINE))
    {
      GList *list;

      list=object->data.polyline.node_list;
      
      while(list!=NULL)
	{
	  tx=((NODE *)list->data)->x;
	  ty=((NODE *)list->data)->y;	 
	  if( !((tx>win_x1) && (tx<win_x2) && (ty>win_y1) && (ty<win_y2)))
	    {
	      found=FALSE;
	      break;
	    }
	  list=g_list_next(list);
	}
    }

  if(object->type==TYPE_SPLINE && (gl_select_mask & TYPE_SPLINE))
    {
      POINT *list;
      gint i;

      if(object->data.spline.cache_point==NULL)
	calc_bspline(object);

      list=object->data.spline.cache_point;
       
      for(i=0;i<object->data.spline.num;i++)
	{
	  tx=list[i].x;
	  ty=list[i].y;
	 
	  if( !((tx>win_x1) && (tx<win_x2) && (ty>win_y1) && (ty<win_y2)))
	    {
	      found=FALSE;
	      break;
	    }
	}
    }

  if(object->type==TYPE_POINT && (gl_select_mask & TYPE_POINT))
    {
      if( !((object->data.point.centrum.x>win_x1) &&
	    (object->data.point.centrum.x<win_x2) &&
	    (object->data.point.centrum.y>win_y1) &&
	    (object->data.point.centrum.y<win_y2)))
	found=FALSE;
    }

  if(object->type==TYPE_DIMENSION && (gl_select_mask & TYPE_DIMENSION))
    {      
      found=FALSE;
    }

  if(found==TRUE && add_object && object->selected==OBJECT_NOT_SEL)
    {
      make_object_selected(object,OBJECT_SEL, TRUE);
      redraw_object(doc, object);
      number_of_objects++;
    }
  else if(found==TRUE && !add_object && object->selected==OBJECT_SEL)
    {
      make_object_selected(object,OBJECT_NOT_SEL, TRUE);
      redraw_object(doc, object);
      number_of_objects++;
    }
}

void calc_select_window(DOCUMENT *doc, gdouble x1, gdouble y1, gdouble x2, gdouble y2)
{
  fix_window(&x1,&y1,&x2,&y2);
  win_x1=x1;
  win_y1=y1;
  win_x2=x2;
  win_y2=y2;

  number_of_objects=0;
  foreach_object(doc, calc_select_window_object);
  if(add_object)
    print( _("\n  %d objects added."), number_of_objects);
  else
    print( _("\n  %d objects removed."), number_of_objects);
}

void calc_select_crossing_object(DOCUMENT *doc, LAYER *layer, OBJECT *object)
{
  gint found;
  gdouble tx,ty;

  found=FALSE;

  if( (object->type==TYPE_LINE && (gl_select_mask & TYPE_LINE)))
    {
      if( ((object->data.line.p1.x>win_x1) &&
	   (object->data.line.p1.x<win_x2) &&
	   (object->data.line.p1.y>win_y1) &&
	   (object->data.line.p1.y<win_y2)))
	{
	  object->data.line.p1.selected=TRUE;
	  found=TRUE;  
	}

      if( ((object->data.line.p2.x>win_x1) &&
	   (object->data.line.p2.x<win_x2) &&
	   (object->data.line.p2.y>win_y1) &&
	   (object->data.line.p2.y<win_y2)))
	{
	  object->data.line.p2.selected=TRUE;
	  found=TRUE;
	}
    }

  if((object->type==TYPE_RLINE && (gl_select_mask & TYPE_RLINE)))
    {
      if( ((object->data.line.p1.x>win_x1) &&
	   (object->data.line.p1.x<win_x2) &&
	   (object->data.line.p1.y>win_y1) &&
	   (object->data.line.p1.y<win_y2)))
	{
	  object->data.line.p1.selected=TRUE;
	  found=TRUE;  
	}
    }

  if(object->type==TYPE_CIRCLE && (gl_select_mask & TYPE_CIRCLE))
    {
      if( ((object->data.circle.centrum.x>win_x1) &&
	   (object->data.circle.centrum.x<win_x2) &&
	   (object->data.circle.centrum.y>win_y1) &&
	   (object->data.circle.centrum.y<win_y2)))
	{
	  object->data.circle.centrum.selected=TRUE;
	  found=TRUE;
	}
    }

  if(object->type==TYPE_ELLIPSE && (gl_select_mask & TYPE_ELLIPSE))
    {
      if( ((object->data.ellipse.centrum.x>win_x1) &&
	   (object->data.ellipse.centrum.x<win_x2) &&
	   (object->data.ellipse.centrum.y>win_y1) &&
	   (object->data.ellipse.centrum.y<win_y2)))
	{
	  object->data.ellipse.centrum.selected=TRUE;
	  found=TRUE;
	}
    }

  if(object->type==TYPE_TEXT && (gl_select_mask & TYPE_TEXT))
    {
      if( ((object->data.text.insert.x>win_x1) &&
	   (object->data.text.insert.x<win_x2) &&
	   (object->data.text.insert.y>win_y1) &&
	   (object->data.text.insert.y<win_y2)))
	{
	  object->data.text.insert.selected=TRUE;
	  found=TRUE;
	}
    }

  if(object->type==TYPE_ATTRIBUTE && (gl_select_mask & TYPE_ATTRIBUTE))
    {
      if( ((object->data.attribute.text.insert.x>win_x1) &&
	   (object->data.attribute.text.insert.x<win_x2) &&
	   (object->data.attribute.text.insert.y>win_y1) &&
	   (object->data.attribute.text.insert.y<win_y2)))
	{
	  object->data.text.insert.selected=TRUE;
	  found=TRUE;
	}
    }

  if(object->type==TYPE_ARC && (gl_select_mask & TYPE_ARC))
    {
      tx=object->data.arc.centrum.x+object->data.arc.radius*cos(object->data.arc.start_angle*M_PI/180.0);
      ty=object->data.arc.centrum.y-object->data.arc.radius*sin(object->data.arc.start_angle*M_PI/180.0);

      if( ((tx>win_x1) && (tx<win_x2) &&
	   (ty>win_y1) && (ty<win_y2)))
	{
	  object->data.arc.selected_startpoint=TRUE;
	  found=TRUE;
	}

      tx=object->data.arc.centrum.x+object->data.arc.radius*cos(object->data.arc.end_angle*M_PI/180.0);
      ty=object->data.arc.centrum.y-object->data.arc.radius*sin(object->data.arc.end_angle*M_PI/180.0);

      if( ((tx>win_x1) && (tx<win_x2) &&
	   (ty>win_y1) && (ty<win_y2)))
	{
	  object->data.arc.selected_endpoint=TRUE;
	  found=TRUE;
	}
    }

  if(object->type==TYPE_POLYLINE && (gl_select_mask & TYPE_POLYLINE))
    {
      GList *list;

      list=object->data.polyline.node_list;

      while(list!=NULL)
	{
	  tx=((NODE *)list->data)->x;
	  ty=((NODE *)list->data)->y;

	  if( ((tx>win_x1) && (tx<win_x2) &&
	       (ty>win_y1) && (ty<win_y2)))
	    {
	      ((NODE *)list->data)->selected=TRUE;
	      found=TRUE;
	    }
	  
	  list=g_list_next(list);
	}
    }

  if(object->type==TYPE_SPLINE && (gl_select_mask & TYPE_SPLINE))
    {
      GList *list;

      list=object->data.spline.node_list;

      while(list!=NULL)
	{
	  tx=((NODE *)list->data)->x;
	  ty=((NODE *)list->data)->y;

	  if( ((tx>win_x1) && (tx<win_x2) &&
	       (ty>win_y1) && (ty<win_y2)))
	    {
	      ((NODE *)list->data)->selected=TRUE;
	      found=TRUE;
	    }
	  
	  list=g_list_next(list);
	}
    }

  if(object->type==TYPE_POINT && (gl_select_mask & TYPE_POINT))
    {
      if( ((object->data.point.centrum.x>win_x1) &&
	   (object->data.point.centrum.x<win_x2) &&
	   (object->data.point.centrum.y>win_y1) &&
	   (object->data.point.centrum.y<win_y2)))
	{
	  object->data.point.centrum.selected=TRUE;
	  found=TRUE;
	}
    }

  if(found==TRUE && add_object && object->selected==OBJECT_NOT_SEL)
    {
      object->selected=OBJECT_SEL;
      object->layer->doc->selection_list=g_list_append(object->layer->doc->selection_list, (gpointer)object);
      redraw_object(doc, object);
      number_of_objects++;
    }
  else if(found==TRUE && !add_object && object->selected==OBJECT_SEL)
    {
      object->selected=OBJECT_NOT_SEL;
      object->layer->doc->selection_list=g_list_remove(object->layer->doc->selection_list, (gpointer)object);
      redraw_object(doc, object);
      number_of_objects++;
    }
}

void calc_select_crossing(DOCUMENT *doc, gdouble x1, gdouble y1, gdouble x2, gdouble y2)
{
  fix_window(&x1,&y1,&x2,&y2);
  win_x1=x1;
  win_y1=y1;
  win_x2=x2;
  win_y2=y2;
  number_of_objects=0;

  foreach_object(doc, calc_select_crossing_object);
  if(add_object)
    print( _("\n  %d objects added."), number_of_objects);
  else
    print( _("\n  %d objects removed."), number_of_objects);
}

void select_crossing(GtkWidget *widget, gint data)
{
  gl_stage=3;
  set_prompt();
}

void select_point(GtkWidget *widget, gint data)
{
  gl_stage=0;
  set_prompt();
}

void select_add_remove_object(GtkWidget *widget, gpointer data)
{
  add_object=GPOINTER_TO_INT(data);

  select_set_radiobuttons(add_object);
  set_prompt();
}

void spec_set_select(gchar *new_prompt, gchar *new_tool)
{
  o_gl_tool=gl_tool;
  o_gl_stage=gl_stage;
  o_gl_oldmx=gl_oldmx;
  o_gl_oldmy=gl_oldmy;

  gl_tool=TOOL_SELECT;
  gl_stage=0;

  if(new_prompt!=NULL)
    strcpy(select_prompt, new_prompt);
  else
    strcpy(select_prompt, _("Select object(s): "));

  UpdateCoordsLabel();

  if(new_tool!=NULL)
    UpdateCoordsLabelTool(new_tool);
  else
    UpdateCoordsLabelTool( _("Selection"));

  add_object=TRUE;

  select_first_object=TRUE;

  select_submenu(TRUE);
  select_set_radiobuttons(add_object);
}

void spec_set_prompt_select(void)
{
  switch(gl_stage)
    {
    case 0: 
      if(add_object)
	set_prompt_text(select_prompt);
      else
	set_prompt_text( _("Remove object(s): "));
      break;
    case 3:
    case 1: 
      set_prompt_text( _("First corner: "));
      break;
    case 4:
    case 2: 
      set_prompt_text( _("Second corner: "));
      break;
    }
}

void spec_button_select(DOCUMENT *doc, gint button, gdouble x, gdouble y)
{
  OBJECT *object;

  /* Scale sensitivity to world scale */
  sens=gl_select_sensitivity/doc->zoom;

  /* Right mousebutton => cancel/break */
  if(button==3)
    {
      remove_old();
      gl_tool=o_gl_tool;
      gl_stage=o_gl_stage;
      gl_oldmx=o_gl_oldmx;
      gl_oldmy=o_gl_oldmy;
      set_prompt();
      set_cursor(doc);

      if(gl_tool==TOOL_ERASE)
	erase_select_object(doc);

      if(gl_tool==TOOL_LIST)
	list_select_object(doc);

      if(gl_tool==TOOL_EPROP)
	eprop_select_object(doc);

      select_submenu(FALSE);
      return;
    }

  /* If this is the first selection, free
     the selection list */
  if(select_first_object)
    {
      if(doc->selection_list!=NULL)
	g_list_free(doc->selection_list);

      doc->selection_list=NULL;
      select_first_object=FALSE;
    }

  switch(gl_stage)
    {
    case 0:
      object=select_object(doc,x,y, TRUE);
      if(object!=NULL)
	{
	  if(add_object && object->selected==OBJECT_NOT_SEL)
	    {
	      make_object_selected(object,OBJECT_SEL, TRUE);
	      print( _("\n  1 object selected."));
	    }
	  else if(!add_object && object->selected==OBJECT_SEL)
	    {
	      make_object_selected(object,OBJECT_NOT_SEL, TRUE);
	      print( _("\n  1 object deselected."));
	    }
	  redraw_object(doc,object);
	}
      else
	print( _("\n  No object(s) found!"));
      break;
    case 3:
    case 1:
      gl_stage++;
      gl_oldmx=x;
      gl_oldmy=y;
      break;
    case 2:
      calc_select_window(doc, gl_oldmx, gl_oldmy, x,y);
      remove_old();
      gl_stage=0;
      break;
    case 4:
      calc_select_crossing(doc, gl_oldmx, gl_oldmy, x,y);
      remove_old();
      gl_stage=0;
      break;
    }

  set_prompt();
}

void spec_menu_select(gint menusel)
{
  switch(menusel)
    {
    case 0:  /* Add */
      select_add_remove_object(NULL, GINT_TO_POINTER(TRUE));
      break;
    case 1:  /* Remove */
      select_add_remove_object(NULL, GINT_TO_POINTER(FALSE));
      break;
    case 2:  /* All */
      select_all(NULL,0);
      break;
    case 3:  /* Invert */
      select_invert(NULL,0);
      break;
    case 4:  /* Point */
      select_point(NULL,0);
      break;
    case 5:  /* Window */
      select_window(NULL,0);
      break;
    case 6:  /* Crossing */
      select_crossing(NULL,0);
      break;
    }
}

gint search_object_by_name(DOCUMENT *doc, gchar *buffer)
{
  GList *layer;
  GList *obj_list;
  OBJECT *object;

  layer=doc->layer_list;

  while(layer!=NULL){
    obj_list=((LAYER *)layer->data)->object_list;
    while(obj_list!=NULL){

      object=(OBJECT *)obj_list->data;

      if(!strcmp(object->name,buffer))
	{
	  if(add_object && object->selected==OBJECT_NOT_SEL)
	    {
	      object->selected=OBJECT_SEL;
	      redraw_object(doc, object);
	      print( _("\n  1 object selected."));
	    }
	  else if(!add_object && object->selected==OBJECT_SEL)
	    {
	      object->selected=OBJECT_NOT_SEL;
	      redraw_object(doc, object);
	      print( _("\n  1 object deselected."));
	    }
	  return TRUE;
	}
      obj_list=g_list_next(obj_list);
    }

    layer=g_list_next(layer);
  }

  return FALSE;
}


void spec_prompt_select(DOCUMENT *doc, gchar *buffer)
{
  gdouble tx,ty;
  gint mn;

  switch(gl_stage)
    {
    case 1:
    case 0:  /* Simple pointer */
      switch(parse_prompt(PROMPT_RETURN|PROMPT_OSNAP|PROMPT_POINT2|PROMPT_MENU|PROMPT_ERROR,buffer,
			  FALSE, _("A/R/ALL/I/P/W/C/"),0,0,&tx,&ty,NULL,&mn)){
      case PROMPT_RETURN:
	work_document=doc;
	remove_old();
	spec_button_select(doc,3,tx,ty);
	break;
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	spec_button_select(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	spec_menu_select(mn);
	break;
      case PROMPT_ERROR:
	if(!search_object_by_name(doc, buffer))	  
	  print( _("\n  No object with name \"%s\" found."),buffer);
	set_prompt();
	break;
      }
      break;
    case 2:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2|PROMPT_MENU,buffer,
			  TRUE, _("A/R/ALL/I/P/W/C/"),gl_oldmx,gl_oldmy,&tx,&ty,NULL,&mn)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	spec_button_select(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	spec_menu_select(mn);
	break;
      }
      break;
    }

}








