/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include "struct.h"
#include "global.h"
#include "misc.h"
#include "dia_coords.h"
#include "dia_comm.h"
#include "dia_message.h"
#include "xor.h"
#include "select.h"
#include "obj_line.h"
#include "obj_circle.h"
#include "obj_arc.h"
#include "redraw.h"
#include "parse.h"
#include "calc.h"
#include "intersect.h"
#include "translation.h"


/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static gdouble dist1=10, dist2=10;
static OBJECT *object_first, *object_second;
static gint trim_objects=TRUE;

/* --  F U N C T I O N S  -- */
static void calc_polychamfer(OBJECT *polyline, NODE *n1, NODE *n2, NODE *n3)
{
  gdouble Ax1,Ay1,Bx1,By1;
  gdouble Ax2,Ay2,Bx2,By2;
  gdouble r;
  gdouble aangle;
  NODE *temp;

  /* Extract endpoints */
  Ax1=n1->x;
  Ay1=n1->y;
  Bx1=Ax2=n2->x;
  By1=Ay2=n2->y;
  Bx2=n3->x;
  By2=n3->y;

  /* Add a node */
  aangle=get_angle(Ax1,Ay1,Bx1,By1)*M_PI/180;
  r=dist(Ax1,Ay1,Bx1,By1);
  r=r-dist1;
  temp=g_new(NODE,1);
  temp->selected=n2->selected;
  temp->type=n2->type;
  temp->x=r*cos(aangle)+Ax1;;
  temp->y=-r*sin(aangle)+Ay1;
  temp->sw=n2->sw;
  temp->ew=n2->ew;
  temp->weight=n2->weight;
  polyline->data.polyline.node_list=g_list_insert(polyline->data.polyline.node_list,
						  (gpointer)temp,g_list_index(polyline->data.polyline.node_list,
									      (gpointer)n1)+1);

  /* Modify  second node */
  r=dist(Bx2,By2,Ax2,Ay2);
  r=r-dist2;   
  aangle=get_angle(Bx2,By2,Ax2,Ay2)*M_PI/180;
  n2->x=r*cos(aangle)+Bx2;
  n2->y=-r*sin(aangle)+By2;
}

static void calculate_chamfer_polyline(DOCUMENT *doc, OBJECT *pline)
{
  GList *node_list;
  NODE *n1,*n2,*n3;

  pline->selected=OBJECT_ERASE;
  redraw_object(doc,pline);

  node_list=pline->data.polyline.node_list;

  while(node_list!=NULL)
    {
      n1=NULL; n2=NULL; n3=NULL;

      /* Try to extract three nodes */
      n1=(NODE *)node_list->data;

      if(node_list->next!=NULL)
	n2=(NODE *)node_list->next->data;
      if(node_list->next->next!=NULL)
	n3=(NODE *)node_list->next->next->data;

      /* Check that all nodes exists */
      if(n1==NULL || n2==NULL || n3==NULL)
      	break;

      /* All the nodes must be lines */
      if(!(n1->type!=NODE_LINE || n2->type!=NODE_LINE || n3->type!=NODE_LINE))
	{
	  calc_polychamfer(pline,n1,n2,n3);
	  /* We dont to calc on the one we just calculated */
	  node_list=g_list_next(node_list);
	}
      node_list=g_list_next(node_list);
    }

  /* Check if this is a closed polyline */
  if(pline->data.polyline.closed==POLYLINE_CLOSE_LINE)
    {
      node_list=pline->data.polyline.node_list;

      n1=(NODE *)g_list_last(node_list)->prev->data;
      n2=(NODE *)g_list_last(node_list)->data;
      n3=(NODE *)g_list_first(node_list)->data;      

      if(!(n1->type!=NODE_LINE || n2->type!=NODE_LINE || n3->type!=NODE_LINE))
	calc_polychamfer(pline,n1,n2,n3);

      n1=(NODE *)g_list_last(node_list)->data;
      n2=(NODE *)g_list_first(node_list)->data;
      n3=(NODE *)g_list_first(node_list)->next->data;      

      if(!(n1->type!=NODE_LINE || n2->type!=NODE_LINE || n3->type!=NODE_LINE))
	calc_polychamfer(pline,n1,n2,n3);
    }

  pline->selected=OBJECT_NOT_SEL;
  redraw_object(doc,pline);
}

static void calculate_chamfer(DOCUMENT *doc, OBJECT *line1, OBJECT *line2)
{
  gdouble int_x, int_y;
  gdouble Ax1,Ay1,Bx1,By1;
  gdouble Ax2,Ay2,Bx2,By2;
  gdouble r;
  gdouble aangle;
  gdouble x1,y1,x2,y2;
  gdouble k1,k2,m1,m2;

  /* Extract endpoints */
  Ax1=line1->data.line.p1.x;
  Ay1=line1->data.line.p1.y;
  Bx1=line1->data.line.p2.x;
  By1=line1->data.line.p2.y;
  Ax2=line2->data.line.p1.x;
  Ay2=line2->data.line.p1.y;
  Bx2=line2->data.line.p2.x;
  By2=line2->data.line.p2.y;

  /* Calculate intersectionpoint */
  if( ABS(Bx1-Ax1)<GFLOAT_SENSITIVE)
    k1=(By1-Ay1)/GFLOAT_SENSITIVE;
  else
    k1=(By1-Ay1)/(Bx1-Ax1);

  if( ABS(Bx2-Ax2)<GFLOAT_SENSITIVE)
    k2=(By2-Ay2)/GFLOAT_SENSITIVE;
  else
    k2=(By2-Ay2)/(Bx2-Ax2);

  if(ABS(k1-k2)<GFLOAT_SENSITIVE)
    {
      print( _("\n  Lines are parallell."));
      return;
    }

  m1=Ay1-k1*Ax1;
  m2=Ay2-k2*Ax2;

  int_x=(m2-m1)/(k1-k2);
  int_y=m1+k1*int_x;

  /* Calculate for line 1 */
  line1->selected=OBJECT_ERASE;
  redraw_object(doc,line1);

  if(dist(Ax1,Ay1,int_x,int_y)<dist(Bx1,By1,int_x,int_y))
    {	  
      Ax1=int_x;
      Ay1=int_y;
      
      r=dist(Ax1,Ay1,Bx1,By1);
      r=r-dist1;
      
      aangle=get_angle(Bx1,By1,Ax1,Ay1)*M_PI/180;
      x1=r*cos(aangle)+Bx1;
      y1=-r*sin(aangle)+By1;
      
      if(trim_objects)
	{
	  line1->data.line.p1.x=x1; 
	  line1->data.line.p1.y=y1; 
	}
    }
  else
    {
      Bx1=int_x;
      By1=int_y;
      
      r=dist(Ax1,Ay1,Bx1,By1);
      r=r-dist1;
      
      aangle=get_angle(Ax1,Ay1,Bx1,By1)*(gdouble)M_PI/180;
      x1=r*cos(aangle)+Ax1;
      y1=-r*sin(aangle)+Ay1;
      
      if(trim_objects)
	{
	  line1->data.line.p2.x=x1; 
	  line1->data.line.p2.y=y1; 
	}
      
    }
  
  line1->selected=OBJECT_SEL;
  redraw_object(doc,line1);
  
  /* Calculate for line 2 */
  
  line2->selected=OBJECT_ERASE;
  redraw_object(doc,line2);
    
  if(dist(Ax2,Ay2,int_x,int_y)<dist(Bx2,By2,int_x,int_y))
    {	  
      Ax2=int_x;
      Ay2=int_y;
      
      r=dist(Ax2,Ay2,Bx2,By2);
      r=r-dist2;
      
      aangle=get_angle(Bx2,By2,Ax2,Ay2)*M_PI/180;
      x2=r*cos(aangle)+Bx2;
      y2=-r*sin(aangle)+By2;
      
      if(trim_objects)
	{
	  line2->data.line.p1.x=x2; 
	  line2->data.line.p1.y=y2; 
	}
    }
  else
    {
      Bx2=int_x;
      By2=int_y;
      
      r=dist(Ax2,Ay2,Bx2,By2);
      r=r-dist2;
      
      aangle=get_angle(Ax2,Ay2,Bx2,By2)*(gdouble)M_PI/180;
      x2=r*cos(aangle)+Ax2;
      y2=-r*sin(aangle)+Ay2;
      
      if(trim_objects)
	{
	  line2->data.line.p2.x=x2; 
	  line2->data.line.p2.y=y2; 
	}
    }
  
  line2->selected=OBJECT_SEL;
  redraw_object(doc,line2);
  
  /* Make new line between the old onces */
  redraw_object(doc, obj_add_line(doc,doc->current_layer, NULL, doc->type_colour, doc->current_colour, 
				  doc->type_linestyle, doc->current_linestyle,
				  doc->type_linewidth, doc->current_linewidth, x1, y1 ,x2, y2));

  obj_bound_line(line1);
  obj_bound_line(line2);
}

void tool_set_prompt_chamfer(void)
{
  switch(gl_stage)
    {
    case 0: 
      set_prompt_text( _("Polyline/Distance/Trim/<Select first line>: "));
      break;
    case 1: 
      set_prompt_text( _("Select second line: "));
      break;
    case 2: 
      set_prompt_text( _("Set Distance 1: "));
      break;
    case 3: 
      set_prompt_text( _("Set Distance 2: "));
      break;
    case 4: 
      set_prompt_text( _("trim True/False: "));
      break;
    case 5:
      set_prompt_text( _("Select polyline to chamfer: "));
      break;
    }
}

void tool_set_tool_chamfer(GtkWidget *widget, gpointer data)
{
  remove_old();
 
  if(!command_from_keyboard)
    print( _("chamfer"));

  if(work_document!=NULL)
    deselect_all_objects(work_document);
  
  if(misc_tool_get_document()==NULL)
    return;

  gl_tool=TOOL_CHAMFER;
  gl_stage=0;
  work_document=NULL;

  UpdateCoordsLabel();

  print( _("\n  Current distance 1: %f, distance 2: %f"), dist1,dist2);
  if(trim_objects)
    print( _(", Trim objects."));
  else
    print( _(", Don't trim objects."));

  UpdateSingelCoord(2,dist1);
  UpdateSingelCoord(3,dist2);

  set_prompt();

}

void tool_button_chamfer(DOCUMENT *doc, gint button, gdouble x, gdouble y)
{
  /* Right mousebutton => cancel/break */
  if(button==3)
    {
      remove_old();
      print( _("*Cancel*"));
      deselect_all_objects(doc);
      tool_idle(doc);
      return;
    }
  
  switch(gl_stage)
    {
    case 0:
      object_first=select_singel_object(doc,x,y);
      if(object_first!=NULL)
	{
	  if(object_first->type==TYPE_LINE)
	    {
	      object_first->selected=OBJECT_SEL;
	      redraw_object(doc, object_first);
	      gl_stage++;
	      gl_oldmx=x;
	      gl_oldmy=y;
	    }
	  else
	    print( _("\n  Object must be a line."));
	}
      else
	print( _("\n  No object found!"));
      set_prompt();
      break;
    case 1:
      object_second=select_singel_object(doc,x,y);
      if(object_second!=NULL)
	{
	  if(object_second->type==TYPE_LINE)
	    {
	      object_second->selected=OBJECT_SEL;
	      redraw_object(doc, object_second);
	      remove_old();
	      calculate_chamfer(doc, object_first, object_second);
	      deselect_all_objects(doc);
	      tool_idle(doc);
	      return;
	    }
	  else
	    print( _("\n Object must be a line."));
	}
      else
	print( _("\n  No object found!"));
      set_prompt();
      break;
    case 5:
      object_first=select_singel_object(doc,x,y);
      if(object_first!=NULL)
	{
	  if(object_first->type==TYPE_POLYLINE)
	    {
	      remove_old();
	      calculate_chamfer_polyline(doc, object_first);
	      deselect_all_objects(doc);
	      tool_idle(doc);
	      return;
	    }
	  else
	    print( _("\n Object must be a polyline."));
	}
      else
	print( _("\n  No object found."));
      set_prompt();
      break;
    }
}

void tool_prompt_chamfer(DOCUMENT *doc, gchar *buffer)
{
  gdouble tx,ty, value;
  gint mn;

  switch(gl_stage)
    {
    case 0:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2|PROMPT_MENU,buffer,
			  FALSE, _("P/D/T/"),0,0,&tx,&ty,NULL,&mn)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	tool_button_chamfer(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	switch(mn)
	  {
	  case 0:  /* Select polyline */
	    gl_stage=5;
	    set_prompt();
	    break;
	  case 1:  /* Enter distances */
	    gl_stage=2;
	    set_prompt();
	    break;
	  case 2:  /* Angle */
	    break;
	  case 3:  /* Trim */
	    gl_stage=4;
	    set_prompt();
	    break;
	  case 4:  /* Method */
	    break;
	  }
      }
      break;
    case 1:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  TRUE,NULL,gl_oldmx,gl_oldmy,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	tool_button_chamfer(doc,1,tx,ty);
	break;
      }
      break;
    case 2:   /* Set distance 1 */
      switch(parse_prompt(PROMPT_RETURN|PROMPT_OSNAP|PROMPT_POINT1,buffer,
			  FALSE,NULL,0,0,NULL,NULL,&value,NULL)){
      case PROMPT_POINT1:
	dist1=value;
      case PROMPT_RETURN:
	gl_stage=3;
	UpdateSingelCoord(2,dist1);
	set_prompt();
	break;
      }
      break;
    case 3:   /* Set distance 2 */
      switch(parse_prompt(PROMPT_RETURN|PROMPT_OSNAP|PROMPT_POINT1,buffer,
			  FALSE,NULL,0,0,NULL,NULL,&value,NULL)){
      case PROMPT_POINT1:
	dist2=value;
      case PROMPT_RETURN:
	gl_stage=0;
	print( _("\n  New distance 1: %f, distance 2: %f"), dist1,dist2);
	UpdateSingelCoord(3,dist2);
	set_prompt();
	break;
      }
      break;
    case 4: /* Trimming */
      switch(parse_prompt(PROMPT_RETURN|PROMPT_OSNAP|PROMPT_MENU,buffer,
			  FALSE, _("T/F/"),0,0,NULL,NULL,NULL,&mn)){
      case PROMPT_MENU:
	trim_objects=(mn==0);
      case PROMPT_RETURN:
	gl_stage=0;
	print( _("\n  Trimming is set to "));
	if(trim_objects)
	  print( _("ON."));
	else
	  print( _("OFF."));
	set_prompt();
	break;
      }
      break;
    case 5:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  FALSE,NULL,0,0,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	tool_button_chamfer(doc,1,tx,ty);
	break;
      }
      break;
    }
}








