/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include <math.h>
#include "struct.h"
#include "global.h"
#include "misc.h"
#include "dia_coords.h"
#include "dia_comm.h"
#include "dia_message.h"
#include "xor.h"
#include "select.h"
#include "obj_line.h"
#include "obj_circle.h"
#include "obj_arc.h"
#include "redraw.h"
#include "parse.h"
#include "calc.h"
#include "translation.h"


/* --  G L O B A L   V A R I A B L E S  -- */

/* --  L O C A L   V A R I A B L E S  -- */
static gdouble fillet_radius=10;
static gint trim_objects=TRUE;

/* --  F U N C T I O N S  -- */
static gboolean calc_fillet(gdouble *Ax1, gdouble *Ay1, gdouble *Bx1, gdouble *By1, 
			    gdouble *Ax2, gdouble *Ay2, gdouble *Bx2, gdouble *By2,
			    OBJECT_ARC *arc)
{
  gdouble sangle,eangle;
  gint w;
  gdouble int_x, int_y;
  gdouble fill_x, fill_y;
  gdouble d;
  gdouble k1,k2,m1,m2;
  gdouble angle1,angle2,angle3;
  
  /* Calculate intersectionpoint */
  if( ABS(*Bx1-*Ax1)<GFLOAT_SENSITIVE)
    k1=(*By1-*Ay1)/GFLOAT_SENSITIVE;
  else
    k1=(*By1-*Ay1)/(*Bx1-*Ax1);

  if( ABS(*Bx2-*Ax2)<GFLOAT_SENSITIVE)
    k2=(*By2-*Ay2)/GFLOAT_SENSITIVE;
  else
    k2=(*By2-*Ay2)/(*Bx2-*Ax2);

  /* No intersection => No arc possible */
  if(ABS(ABS(k1)-ABS(k2))<GFLOAT_SENSITIVE)
    return FALSE;

  m1=(*Ay1)-k1*(*Ax1);
  m2=(*Ay2)-k2*(*Ax2);

  int_x=(m2-m1)/(k1-k2);
  int_y=m1+k1*int_x;

  if(dist(int_x, int_y, *Ax1, *Ay1)<dist(int_x, int_y, *Bx1, *By1))
    angle1=get_angle(int_x, int_y, *Bx1, *By1);
  else
    angle1=get_angle(int_x, int_y, *Ax1, *Ay1);

  if(dist(int_x, int_y, *Ax2, *Ay2)<dist(int_x, int_y, *Bx2, *By2))
    angle2=get_angle(int_x, int_y, *Bx2, *By2);
  else
    angle2=get_angle(int_x, int_y, *Ax2, *Ay2);

  if(angle2>angle1)
    angle3=(angle2-angle1);
  else
    angle3=(angle2+(360.0-angle1));
    
  w=(angle3>180);

  angle3*=0.5;

  d=fillet_radius/sin((angle3)*M_PI/180.0);

  angle3=angle3+angle1;
  if(angle3>360.0)
    angle3=angle3-360.0;

  if(w)
    angle3=angle3-180.0;

  fill_x=int_x+d*cos(angle3*M_PI/180.0);
  fill_y=int_y-d*sin(angle3*M_PI/180.0);

  /* Calculate start and end angles for arc */
  if(w)
    {
      sangle=180+angle1-90;
      eangle=180+angle2+90;
    }
  else
    {
      eangle=angle1-90;
      sangle=angle2+90;
    }

  while(sangle>360) sangle-=360;
  while(sangle<0) sangle+=360;
  while(eangle>360) eangle-=360;
  while(eangle<0) eangle+=360;

  /* Arc data */
  arc->centrum.x=fill_x;
  arc->centrum.y=fill_y;
  arc->radius=fillet_radius;
  arc->start_angle=sangle;
  arc->end_angle=eangle;

  /* New endpoints for lines */

  /* First line  */
  if(w)
    {
      k1=fill_x+fillet_radius*cos(sangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(sangle*M_PI/180.0);
    }
  else
    {
      k1=fill_x+fillet_radius*cos(eangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(eangle*M_PI/180.0);
    }

  if(dist(int_x, int_y, *Ax1, *Ay1)<dist(int_x, int_y, *Bx1, *By1))
    {
      *Ax1=k1;
      *Ay1=k2;      
    }
  else
    {
      *Bx1=k1;
      *By1=k2;      
    }
  
  /* Second line  */
  if(w)
    {
      k1=fill_x+fillet_radius*cos(eangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(eangle*M_PI/180.0);
    }
  else
    {
      k1=fill_x+fillet_radius*cos(sangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(sangle*M_PI/180.0);
    }

  if(dist(int_x, int_y, *Ax2, *Ay2)<dist(int_x, int_y, *Bx2, *By2))
    {
      *Ax2=k1;
      *Ay2=k2;      
    }
  else
    {
      *Bx2=k1;
      *By2=k2;      
    }

  return TRUE;
}


static void calc_polyfillet(DOCUMENT *doc,OBJECT *polyline, NODE *n1, NODE *n2, NODE *n3)
{
  gdouble Ax1,Ay1,Bx1,By1;
  gdouble Ax2,Ay2,Bx2,By2;
  gdouble k1,k2;
  NODE *temp;
  gdouble sangle,eangle;
  gdouble fill_x,fill_y;
  gdouble angle1,angle2,angle3;
  gdouble d;
  gint w;

  /* Extract endpoints */
  Ax1=n1->x;
  Ay1=n1->y;
  Bx1=Ax2=n2->x;
  By1=Ay2=n2->y;
  Bx2=n3->x;
  By2=n3->y;

  angle1=get_angle(n2->x, n2->y, Ax1,Ay1);
  angle2=get_angle(n2->x, n2->y, Bx2,By2);

  if(angle2>angle1)
    angle3=(angle2-angle1);
  else
    angle3=(angle2+(360.0-angle1));

  w=(angle3>180);
    
  angle3*=0.5;

  d=fillet_radius/sin((angle3)*M_PI/180.0);

  angle3=angle3+angle1;
  if(angle3>360.0)
    angle3=angle3-360.0;

  if(w)
    angle3=angle3-180.0;

  fill_x=n2->x+d*cos(angle3*M_PI/180.0);
  fill_y=n2->y-d*sin(angle3*M_PI/180.0);
  
  /* Calculate start and end angles for arc */
  if(w)
    {
      sangle=180+angle1-90;
      eangle=180+angle2+90;
    }
  else
    {
      eangle=angle1-90;
      sangle=angle2+90;
    }

  if(sangle>360)
    sangle-=360;
  if(eangle>360)
    eangle-=360;
  if(sangle<0)
    sangle+=360;
  if(eangle<0)
    eangle+=360;

  /* Find end point of arc */
  if(w)
    {
      k1=fill_x+fillet_radius*cos(sangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(sangle*M_PI/180.0);
    }
  else
    {
      k1=fill_x+fillet_radius*cos(eangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(eangle*M_PI/180.0);
    }

  /* Add a node */
  temp=g_new(NODE,1);
  temp->selected=n2->selected;
  temp->type=n2->type;
  temp->x=k1;
  temp->y=k2;
  temp->sw=n2->sw;
  temp->ew=n2->ew;
  temp->weight=n2->weight;

  polyline->data.polyline.node_list=g_list_insert(polyline->data.polyline.node_list,
  						  (gpointer)temp,g_list_index(polyline->data.polyline.node_list,
  								      (gpointer)n1)+1);  
  if(w)
    {
      k1=fill_x+fillet_radius*cos(eangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(eangle*M_PI/180.0);
    }
  else
    {
      k1=fill_x+fillet_radius*cos(sangle*M_PI/180.0);
      k2=fill_y-fillet_radius*sin(sangle*M_PI/180.0);
    }

  /* Add a node */
  temp=g_new(NODE,1);
  temp->selected=n2->selected;
  temp->type=n2->type;
  temp->x=k1;
  temp->y=k2;
  temp->sw=n2->sw;
  temp->ew=n2->ew;
  temp->weight=n2->weight;
  polyline->data.polyline.node_list=g_list_insert(polyline->data.polyline.node_list,
						  (gpointer)temp,g_list_index(polyline->data.polyline.node_list,
									      (gpointer)n2)+1);  
  /* Calc mid point */
  n2->type=NODE_ARC;

  n2->x=fill_x+fillet_radius*cos(angle3*M_PI/180.0+M_PI);
  n2->y=fill_y-fillet_radius*sin(angle3*M_PI/180.0+M_PI);  
}

static void calculate_fillet_polyline(DOCUMENT *doc, OBJECT *pline)
{
  GList *node_list;
  NODE *n1,*n2,*n3;

  pline->selected=OBJECT_ERASE;
  redraw_object(doc,pline);

  node_list=g_list_first(pline->data.polyline.node_list);

  while(node_list!=NULL)
    {
      n1=n2=n3=NULL;

      /* Try to extract three nodes */
      n1=(NODE *)node_list->data;

      if(node_list->next!=NULL)
	n2=(NODE *)node_list->next->data;
      if(node_list->next->next!=NULL)
	n3=(NODE *)node_list->next->next->data;

      /* Check that all nodes exists */
      if(n1==NULL || n2==NULL || n3==NULL)
      	break;

      /* All the nodes must be lines */
      if( n1->type==NODE_LINE && n2->type==NODE_LINE && n3->type==NODE_LINE)
	{
	  calc_polyfillet(doc, pline, n1, n2, n3);
	  /* We dont to calc on the one we just calculated */
	  node_list=g_list_next(node_list);
	  node_list=g_list_next(node_list);
	}
      node_list=g_list_next(node_list);
    }

  /* Check if this is a closed polyline */
  if(pline->data.polyline.closed==POLYLINE_CLOSE_LINE)
    {
      node_list=pline->data.polyline.node_list;

      n1=(NODE *)g_list_last(node_list)->prev->data;
      n2=(NODE *)g_list_last(node_list)->data;
      n3=(NODE *)g_list_first(node_list)->data;      

      if(n1->type==NODE_LINE && n2->type==NODE_LINE && n3->type==NODE_LINE)
	calc_polyfillet(doc, pline,n1,n2,n3);

      n1=(NODE *)g_list_last(node_list)->data;
      n2=(NODE *)g_list_first(node_list)->data;
      n3=(NODE *)g_list_first(node_list)->next->data;      

      if( n1->type==NODE_LINE && n2->type==NODE_LINE && n3->type==NODE_LINE )
	calc_polyfillet(doc, pline,n1,n2,n3);
    }

  pline->selected=OBJECT_NOT_SEL;
  redraw_object(doc,pline);
}

static void calculate_fillet_lines(DOCUMENT *doc, OBJECT *obj1, OBJECT *obj2)
{
  OBJECT_ARC arc;
  gdouble Ax1, Ay1, Bx1, By1;
  gdouble Ax2, Ay2, Bx2, By2;

  /* Erase old objects */
  obj1->selected=OBJECT_ERASE;
  redraw_object(doc,obj1);
  obj2->selected=OBJECT_ERASE;
  redraw_object(doc,obj2);

  /* Extract endpoints */
  Ax1=obj1->data.line.p1.x;
  Ay1=obj1->data.line.p1.y;
  Bx1=obj1->data.line.p2.x;
  By1=obj1->data.line.p2.y;
  Ax2=obj2->data.line.p1.x;
  Ay2=obj2->data.line.p1.y;
  Bx2=obj2->data.line.p2.x;
  By2=obj2->data.line.p2.y;

  /* Calculate the fillet */
  if(calc_fillet(&Ax1, &Ay1, &Bx1, &By1, 
	      &Ax2, &Ay2, &Bx2, &By2,
	      &arc))
    {
      
      /* Create arc */
      redraw_object(doc,obj_add_arc(doc, doc->current_layer, NULL, doc->type_colour, doc->current_colour, 
				    doc->type_linestyle, doc->current_linestyle,
				    doc->type_linewidth, doc->current_linewidth,
				    arc.centrum.x, arc.centrum.y, arc.radius, 
				    arc.start_angle, arc.end_angle));
      

      /*  Set the new endpoints if trimming is TRUE */
      if(trim_objects)
	{
	  obj1->data.line.p1.x=Ax1;
	  obj1->data.line.p1.y=Ay1;
	  obj1->data.line.p2.x=Bx1;
	  obj1->data.line.p2.y=By1;
	  obj2->data.line.p1.x=Ax2;
	  obj2->data.line.p1.y=Ay2;
	  obj2->data.line.p2.x=Bx2;
	  obj2->data.line.p2.y=By2;
	}
    }
  else
    print( _("\n  Lines are parallell. Fillet not possible."));

  obj_bound_line(obj1);
  obj_bound_line(obj2);
}

void tool_set_prompt_fillet(void)
{
  switch(gl_stage)
    {
    case 0: 
      set_prompt_text( _("Radius/Trim/<Select polyline or first line>: "));
      break;
    case 1: 
      set_prompt_text( _("Select second line: "));
      break;
    case 3: 
      set_prompt_text( _("Enter fillet radius <%f>: "), fillet_radius);
      break;
    case 4: 
      set_prompt_text( _("Trim objects? Yes/No: "));
      break;
    }
}

void tool_set_tool_fillet(GtkWidget *widget, gpointer data)
{
  remove_old();
 
  if(!command_from_keyboard)
    print( _("fillet"));

  if(work_document!=NULL)
    deselect_all_objects(work_document);
  
  if(misc_tool_get_document()==NULL)
    return;

  gl_tool=TOOL_FILLET;
  gl_stage=0;
  work_document=NULL;

  UpdateCoordsLabel();

  print( _("\n  Current fillet radius: %f"),fillet_radius );
  if(trim_objects)
    print( _(", Trim objects."));
  else
    print( _(", Don't trim objects.")); 

  set_prompt();

  work_document=NULL;
}

void tool_button_fillet(DOCUMENT *doc, gint button, gdouble x, gdouble y)
{
  static OBJECT *object_first, *object_second;

  /* Right mousebutton => cancel/break */
  if(button==3)
    {
      remove_old();
      deselect_all_objects(doc);
      tool_idle(doc);
      return;
    }
  
  switch(gl_stage)
    {
    case 0:
      object_first=select_singel_object(doc,x,y);
      if(object_first!=NULL)
	{
	  if(object_first->type==TYPE_LINE)
	    {
	      object_first->selected=OBJECT_SEL;
	      redraw_object(doc, object_first);
	      gl_stage++;
	    }
	  else if(object_first->type==TYPE_POLYLINE)
	    {
	      remove_old();
	      calculate_fillet_polyline(doc, object_first);
	      deselect_all_objects(doc);
	      tool_idle(doc);
	      return;
	    }
	  else
	    print( _("\n  Object must be a polyline or a line."));
	}	  
      else
	print( _("\n  No object found!"));
      set_prompt();
      break;
    case 1:
      object_second=select_singel_object(doc,x,y);
      if(object_second!=NULL)
	{
	  if(object_second->type==TYPE_LINE)
	    {
	      calculate_fillet_lines(doc, object_first, object_second);
	      deselect_all_objects(doc);
	      tool_idle(doc);
	      return;
	    }
	  else
	    print( _("\n  Object must be a line."));
	}
      else
	print( _("\n  No object found!"));
      set_prompt();
      break;
    }

}

void tool_prompt_fillet(DOCUMENT *doc, gchar *buffer)
{
  gdouble tx,ty,value;
  gint mn;

  switch(gl_stage)
    {
    case 0:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2|PROMPT_MENU,buffer,
			  FALSE, _("R/T/"),0,0,&tx,&ty,NULL,&mn)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	tool_button_fillet(doc,1,tx,ty);
	break;
      case PROMPT_MENU:
	switch(mn)
	  {
	  case 0:  /* Enter radius */
	    gl_stage=3;
	    set_prompt();
	    break;
	  case 1:  /* Trim */
	    gl_stage=4;
	    set_prompt();
	    break;
	  }
      }
      break;
    case 1:
      switch(parse_prompt(PROMPT_OSNAP|PROMPT_POINT2,buffer,
			  TRUE,NULL,gl_oldmx,gl_oldmy,&tx,&ty,NULL,NULL)){
      case PROMPT_POINT2:
	work_document=doc;
	remove_old();
	tool_button_fillet(doc,1,tx,ty);
	break;
      }
      break;
    case 3:   /* radius */
      switch(parse_prompt(PROMPT_RETURN|PROMPT_OSNAP|PROMPT_POINT1,buffer,
			  FALSE,NULL,0,0,NULL,NULL,&value,NULL)){
      case PROMPT_POINT1:
	fillet_radius=value;
      case PROMPT_RETURN:
	gl_stage=0;
	//	UpdateSingelCoord(2,dist1);
	set_prompt();
	break;
      }
      break;
    case 4:  /* Trim */
      switch(parse_prompt(PROMPT_RETURN|PROMPT_OSNAP|PROMPT_MENU,buffer,
			  FALSE, _("Y/N/"),0,0,NULL,NULL,NULL,&mn)){
      case PROMPT_MENU:
	trim_objects=(mn==0);
      case PROMPT_RETURN:
	gl_stage=0;
	set_prompt();
	break;
      }
      break;
    }
}


