/* oCADis
 * Copyright (C) 2001 Jonas Jarvoll
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* --  I N C L U D E S  -- */
#include <gtk/gtk.h>
#include "struct.h"
#include "global.h"
#include "misc.h"
#include "dia_comm.h"
#include "parse.h"
#include "translation.h"


/*
  V a r i a b l e s
  -------------------------------------------------------------
  oCADis is capable of handling variables. You can create a variable
  using the command dim. A variable must be created by the command 
  dim before you use or else a error will be shown. After the variable
  has been created you can set its value by using the command let.
  
  If you want to use the variable in for example a line prompt you can
  do like this:

   > dim a as float
   > dim b as float
   > let a=100
   > let b=100
   > line
   Start Point: a,b

   This is the same as writing:

   > line
   Start Point: 100,100
   
   Perhaps a bit longer but more powerful. You can reuse the variables
   as often as you want but these are not saved between oCADis sessions.

   The variables are local and can only be used in the drawing that they
   where created.

   A shorter version of the above example is using the variable-type point

   > dim a as point
   > let a=100,100
   > line
   Start Point: a

   In future version of oCADis you should be able to make calculation with
   variables like:

   > let a=100
   > let b=a+10
   > line
   Start Point: a+120,b*sin(170)
*/

/* --  L O C A L   V A R I A B L E S  -- */
#define VARTYPE_FLOAT  0
#define VARTYPE_STRING 1
#define VARTYPE_POINT  2

typedef union u_vartype{
  gdouble float_value;
  gchar *string_value;
  gdouble point_value[2];
} VARTYPE;

typedef struct s_variable
{
  gchar *name;     /* Name of variable          */
  gboolean empty;  /* True means that this variable contains no value */
  gint type;       /* float, point, string */
  VARTYPE value;
} VARIABLE;

/* --  F U N C T I O N S  -- */
gboolean var_get_variable_float(DOCUMENT *doc, gchar *name, gdouble *ret)
{
  VARIABLE *var;

  if(doc->variable_list==NULL)
    return FALSE;

  var= (VARIABLE *) g_hash_table_lookup(doc->variable_list, name);

  if(var==NULL)
    return FALSE;

  if(var->type!=VARTYPE_FLOAT)
    return FALSE;

  *ret=var->value.float_value;
  return TRUE;
}
 
static void print_value(gpointer key, gpointer value, gpointer user_data)
{
  VARIABLE *temp;

  temp=(VARIABLE *)value;

  switch(temp->type)
    {
    case VARTYPE_FLOAT:
      if(!temp->empty)
	print( _("\n  %s as float = %f"), temp->name, temp->value.float_value);
      else
	print( _("\n  %s as float = Empty"), temp->name);
      break;
    case VARTYPE_STRING:
      if(!temp->empty)
	print( _("\n  %s as string = \"%s\""), temp->name, temp->value.string_value);
      else
	print( _("\n  %s as string = Empty"), temp->name);
      break;
    case VARTYPE_POINT:
      if(!temp->empty)     
	print( _("\n  %s as point = %f,%f"), temp->name, temp->value.point_value[0], temp->value.point_value[1]);
      else
	print( _("\n  %s as point = Empty"), temp->name);
      break;
    }
}

static void set_variable(DOCUMENT *doc, gchar *name, void *value)
{
  VARIABLE *var;

  var= (VARIABLE *) g_hash_table_lookup(doc->variable_list, name);

  if(var==NULL)
    {
      print( _("\n  Undefined variabel \"%s\""), name);
      return;
    }

  switch(var->type)
    {
    case VARTYPE_FLOAT:
      var->value.float_value=*((gdouble *)value);
      var->empty=FALSE;
      break;
    case VARTYPE_STRING:
      var->value.string_value=g_strdup((gchar *)value);
      var->empty=FALSE;
      break;
    case VARTYPE_POINT:
      var->value.point_value[0]=((gdouble *)value)[0];
      var->value.point_value[1]=((gdouble *)value)[1];
      var->empty=FALSE;
      break;
    }
}

static void create_variable(DOCUMENT *doc, gchar *name, gint type)
{
  VARIABLE *temp;

  /* Make sure this variable doesnt already exists */
  if(doc->variable_list!=NULL && g_hash_table_lookup(doc->variable_list, name)!=NULL)
    {
      print( _("\n  The variable \"%s\" is already defined."), name);
      return;
    }

  temp=g_new(VARIABLE,1);

  temp->name=g_strdup(name);
  temp->type=type;
  temp->empty=TRUE;

  if(doc->variable_list==NULL)
    doc->variable_list=g_hash_table_new(g_str_hash, g_str_equal);

  g_hash_table_insert(doc->variable_list, temp->name, (gpointer)temp);
}

void tool_var_list(GtkWidget *widget, gpointer data)
{
  DOCUMENT *doc;

  doc=get_document();

  if(doc==NULL)
    {
      print( _("\n  No document found"));
      set_prompt();
      return;
    }

  if(doc->variable_list==NULL)
    print( _("\n  No variables defined."));
  else
    {
      print("\n");
      g_hash_table_foreach(doc->variable_list, print_value, NULL);
      print( _("\n\n  Number of defined variables: %d\n"),  
	    g_hash_table_size(doc->variable_list) );
    }

  set_prompt();
}

void tool_var_dim(DOCUMENT *doc, gchar *buffer)
{
  gint type=-1;
  gchar *varname=NULL;
  gchar *vartype=NULL;
  gint i;
  gchar **token;

  if(doc==NULL)
    {
      print( _("\n  No drawing found."));
      return;
    }

  token=g_strsplit(buffer," ", 255);

  for(i=0;token[i]!=NULL;i++)
    {
      /* Remove all leading and trailing spaces */
      g_strstrip(token[i]);
      
      if(i==1)
	varname=token[i];

      if(i==3)
	vartype=token[i];
    }

  if(i==4 && vartype!=NULL && varname!=NULL)
    {
      if(!strcmp(vartype, _("float")))
	type=VARTYPE_FLOAT;
      else if(!strcmp(vartype, _("string")))
	type=VARTYPE_STRING;
      else if(!strcmp(vartype, _("point")))
	type=VARTYPE_POINT;
      else
	print( _("\n  Illegal variable type."));

      if(type!=-1)
	create_variable(doc, varname, type);
    }
  else
    print( _("\n  Syntax error."));
  
  g_strfreev(token);      

}

void tool_var_let(DOCUMENT *doc, gchar *buffer)
{
  VARIABLE *var=NULL;
  gchar varname[32];
  gchar vartype[255];
  gint i,j,k;

  if(doc==NULL)
    {
      print( _("\n  No drawing found."));
      return;
    }

  for(i=0;i<strlen(buffer);i++)
    if(buffer[i]=='=')
      break;

  if(i>=strlen(buffer))
     {
       print( _("\n  Syntax error."));
       return;
     }

  /* Extract the variable type */
  for(j=i-1,k=0;j>3;j--,k++)
    varname[k]=buffer[j];

  varname[k]='\0';

  g_strreverse(varname);
  g_strstrip(varname);

  /* Extract the variable value */
  for(j=i+1,k=0;j<strlen(buffer);j++,k++)
    vartype[k]=buffer[j];

  vartype[k]='\0';

  g_strstrip(vartype);

  /* Check if the variabel exists */
  if(doc->variable_list!=NULL)
    var= (VARIABLE *) g_hash_table_lookup(doc->variable_list, varname);

  if(var==NULL || doc->variable_list==NULL)
    {
      print( _("\n  Undefined variable."));
      return;
    }

  /* Check for type of variabel */
  if(vartype[0]=='"' && var->type==VARTYPE_STRING)
    {
      vartype[strlen(vartype)-1]='\0';
      set_variable(doc, varname, (void *)&vartype[1]);
    }
  else if(var->type==VARTYPE_FLOAT)
    {
      gdouble t;

      if(parse_single_value(doc, vartype, &t))
	set_variable(doc, varname, (void *)&t);
    }
  else if(var->type==VARTYPE_POINT)
    {
      gdouble t[2];

      if(get_coordinate(doc, varname, FALSE, 0, 0, &t[0], &t[1]))
	set_variable(doc, varname, (void *)&t);
    }
  else
    print( _("\n  Illegal value for variabel."));
}







